#!/usr/bin/env python

# Copyright (C) 2016-2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from collections import OrderedDict
from cStringIO import StringIO
from grp import getgrnam
from network_docopt import sort_for_humans
from pprint import pformat
from pwd import getpwnam
from socket import gethostbyaddr
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate  # This is an add-on from Pip.
import errno
import filecmp                 # Using filecmp.cmp.  Note that cmp() is a built-in function.
import json
import logging
import logging.handlers
import os
import re
import shutil
import sys

log = logging.getLogger('netd')

PENDING_COMMANDS_FILE = '/var/run/nclu/pending_commands.txt'
NCLU_VERSION = 1.0

ifupdown2_glob_regexs = [re.compile(r"([A-Za-z0-9\-]+)\[(\d+)\-(\d+)\]([A-Za-z0-9\-]+)\[(\d+)\-(\d+)\](.*)"),
                         re.compile(r"([A-Za-z0-9\-]+[A-Za-z])(\d+)\-(\d+)(.*)"),
                         re.compile(r"([A-Za-z0-9\-]+)\[(\d+)\-(\d+)\](.*)")]

glob_msg = """
ERROR: 'INTERFACE' looks sort of like a glob but does not use valid glob syntax.
Globs use '-' to specify a range of interfaces.

Examples:
    swp1-4
    swp1-4,swp8,swp9
    swp1-4,8,9
    swp1-4,swp8-9
    swp1,swp8,swp9
    swp1,8,9
    swp1,swp8-9

"""


# Each show_pending needs to return the correct state
# since we could have one wrapper with an error.
# In that case we should not commit any changes at all.
PENDING_COMMIT_NOCHANGE = 0
PENDING_COMMIT_CHANGE   = 1
PENDING_COMMIT_ERROR    = 2

class Capturing(list):

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        sys.stdout = self._stdout


def files_match(fileA, fileB):
    # TODO - Depending on how and where this is used downstream, I might be able to replace this
    # function with filecmp.cmp, which I'm calling here.
    if not os.path.isfile(fileA) and not os.path.isfile(fileB):
        return True   # Neither file exists.  TODO - Should this raise an exception instead?

    if not os.path.isfile(fileA) or not os.path.isfile(fileB):
        # TODO - Should this raise an exception instead?
        return False  # Only one file exists, so they cannot be equal.

    return filecmp.cmp(fileA, fileB)


def print_file(filename):
    with open(filename, 'r') as f:
        print f.read()


def print_diff(fileA, fileB, color_diffs):
    """
    Print the diff between fileA vs. fileB with colored output, and
    return True if we printed anything.
    """
    # TODO - Can I delete code by using the difflib standard library package?  This would remove
    # the subprocess call, but difflib.Differ().compare(...) has a different return value syntax
    # than diff.  The former does not include lines beginning with --- or +++.

    if not os.path.isfile(fileA) or not os.path.isfile(fileB):
        return False  # One or both files do not exist.

    try:
        check_output(['/usr/bin/diff',
                      '-U', '10',
                      '--ignore-blank-lines',
                      '--ignore-space-change',
                      fileA,
                      fileB])
        return False
    except CalledProcessError as e:

        # Print the deletes in red and the adds in green
        for line in e.output.splitlines():
            if color_diffs:
                if line.startswith('---') or line.startswith('+++'):
                    print "\033[94m%s\033[0m" % line
                elif line.startswith('-'):
                    print "\033[91m%s\033[0m" % line
                elif line.startswith('+'):
                    print "\033[92m%s\033[0m" % line
                else:
                    print line
            else:
                print line
        return True


def chown_chgroup(filename, user, group):
    uid = getpwnam(user).pw_uid
    gid = getgrnam(group).gr_gid
    os.chown(filename, uid, gid)


def file_contains(filename, target_line):
    """
    Return True if the given file contains the given line.  Otherwise, return False.
    """
    if os.path.isfile(filename):
        with open(filename) as f:
            line = f.readline()
            while line:
                if line.strip() == target_line.strip():
                    return True
                line = f.readline()
    return False


def ifname_is_valid(ifname):
    """
    Enforce the rules for interface names defined in the kernel's dev_valid_name().
    Valid interface names can have up to 15 visible characters plus a null terminator.
        http://stackoverflow.com/questions/24932172/what-length-can-a-network-interface-name-have#29398765
    """
    IF_NAMESIZE = 15
    ILLEGAL_CHARACTERS = ('/', ' ', '%', '*', '?', '^', '$', '+', '[', ']', '{', '}', '(', ')', '\\')
    ILLEGAL_NAMES = ('.', ':')  # Illegal names comprised of legal characters

    ifnames = ifname if isinstance(ifname, list) else [ifname]

    for ifacename in ifnames:
        if ifacename in ILLEGAL_NAMES:
            print "ERROR: '{0}' is an invalid interface name (cannot use '.' or ':').".format(ifacename)
            return False

        for character in ILLEGAL_CHARACTERS:
            if character in ifacename:
                print "ERROR: '{0}' is an invalid interface name (cannot use '{1}').".format(ifacename, "', '".join(ILLEGAL_CHARACTERS))
                return False

        if len(ifacename) > IF_NAMESIZE:
            print "ERROR: '{0}' is an invalid interface name (more than {1} characters).".format(ifacename, IF_NAMESIZE)
            return False

    return True

def ifname_is_globish(ifname, print_msg=True):
    """
    This assumes that ifname is not a glob, as verified by calling ifname_is_glob().

    Return True if the ifname contains characters that one might use when
    attempting to enter a glob.
    """

    assert isinstance(ifname, str) or isinstance(ifname, unicode)

    GLOBISH_CHARACTERS = (',', '[', ']', '(', ')', '%', '*')

    if re.match(r"""^\w+0\d+(?:-0\d+)?$""", ifname) is not None:
        return False

    if re.match(r"""^swp\d+-\w+\d+$""", ifname) is not None:
        if print_msg:
            print glob_msg.replace('INTERFACE', ifname)
        return True

    for character in GLOBISH_CHARACTERS:
        if character in ifname:
            if print_msg:
                print glob_msg.replace('INTERFACE', ifname)
            return True

    return False


def ifupdown2_ifname_is_glob(ifname):
    """
    From ifupdown2's ifupdownaddons/modulebase.py

    >>> ifupdown2_ifname_is_glob('swp1-5,8')
    True
    """
    if (ifupdown2_glob_regexs[0].match(ifname) or
        ifupdown2_glob_regexs[1].match(ifname) or
        ifupdown2_glob_regexs[2].match(ifname)):
        return True
    return False


def ifupdown2_parse_glob(ifacename, expr):
    """
    From ifupdown2's ifupdownaddons/modulebase.py
    """
    errmsg = ('error parsing glob expression \'%s\'' %expr +
                ' (supported glob syntax: swp1-10.300 or swp[1-10].300' +
                '  or swp[1-10]sub[0-4].300')
    regexs = ifupdown2_glob_regexs

    if regexs[0].match(expr):
        # the first regex checks for exactly two levels of ranges defined only with square brackets
        # (e.g. swpxyz[10-23]subqwe[0-4].100) to handle naming with two levels of port names.
        m = regexs[0].match(expr)
        mlist = m.groups()
        if len(mlist) < 7:
            # we have problems and should not continue
            raise Exception('%s: error: unhandled glob expression %s\n%s' % (ifacename, expr,errmsg))

        prefix = mlist[0]
        suffix = mlist[6]
        start_index = int(mlist[1])
        end_index = int(mlist[2])
        sub_string = mlist[3]
        start_sub = int(mlist[4])
        end_sub = int(mlist[5])
        for i in range(start_index, end_index + 1):
            for j in range(start_sub, end_sub + 1):
                yield prefix + '%d%s%d' % (i,sub_string,j) + suffix

    elif regexs[1].match(expr) or regexs[2].match(expr):
        # the second regex for 1 level with a range (e.g. swp10-14.100
        # the third regex checks for 1 level with [] (e.g. swp[10-14].100)
        start_index = 0
        end_index = 0
        if regexs[1].match(expr):
            m = regexs[1].match(expr)
        else:
            m = regexs[2].match(expr)
        mlist = m.groups()
        if len(mlist) != 4:
            raise Exception('%s: ' %ifacename + errmsg + '(unexpected len)')
        prefix = mlist[0]
        suffix = mlist[3]
        start_index = int(mlist[1])
        end_index = int(mlist[2])
        for i in range(start_index, end_index + 1):
            yield prefix + '%d' %i + suffix

    else:
        # Could not match anything.
        log.warning('%s: %s' %(ifacename, errmsg))
        yield expr


def ifname_is_glob(ifname):
    """
    Return True if the given name refers to multiple interfaces (a.k.a. ifname is a glob).
    Otherwise, return False.

    >>> ifname_is_glob('swp1-5')
    True

    >>> ifname_is_glob('swp1-5,8')
    True

    >>> ifname_is_glob('swp1-5,swp8')
    True

    >>> ifname_is_glob('swp1-5,swp7')
    True

    >>> ifname_is_glob('swp1-5,swp7-8')
    True

    >>> ifname_is_glob('swp7')
    False

    >>> ifname_is_glob('server01')
    False
    """

    assert isinstance(ifname, str) or isinstance(ifname, unicode), "ifname={0}".format(ifname)

    if not ifname or ' ' in ifname:
        return False

    # "server01", for example, is not a glob.
    if re.match(r"""^\w+0\d+(?:-0\d+)?$""", ifname) is not None:
        return False

    if ',' in ifname:
        glob_parts = ifname.split(',')
        base_swp = glob_parts[0]

        re_swp = re.compile(r"""[a-zA-Z]+\d+""")
        re_number_number = re.compile(r"""\d+-\d+""")

        if ifupdown2_ifname_is_glob(base_swp):
            for part in glob_parts[1:]:
                if not (
                           part.isdigit() or
                           ifupdown2_ifname_is_glob(part) or
                           (re_swp.match(part) is not None) or
                           (re_number_number.match(part) is not None)
                       ):
                    return False

            return True

        if ifname_is_valid(base_swp):
            for part in glob_parts[1:]:
                if not (
                           part.isdigit() or
                           ifupdown2_ifname_is_glob(part) or
                           (re_swp.match(part) is not None) or
                           (re_number_number.match(part) is not None)
                       ):
                    return False

            return True

    # swp1-5
    return ifupdown2_ifname_is_glob(ifname)


def ifname_expand_glob(ifname):
    """
    Given a glob of interface names, return a list of strings containing each interface.
    If the input is a simple interface name, return it as a list.

    >>> ifname_expand_glob('swp7')
    ['swp7']

    >>> ifname_expand_glob('swp7,eth0')
    ['swp7', 'eth0']

    >>> ifname_expand_glob('swp7,swp8')
    ['swp7', 'swp8']

    >>> ifname_expand_glob('swp7,8')
    ['swp7', 'swp8']

    >>> ifname_expand_glob('swp7-8')
    ['swp7', 'swp8']

    >>> ifname_expand_glob('swp1,8,9')
    ['swp1', 'swp8', 'swp9']

    >>> ifname_expand_glob('swp1,swp8-9')
    ['swp1', 'swp8', 'swp9']
    """

    # This prevents undesirable downstream assertion failures.
    if isinstance(ifname, unicode):
        ifname = str(ifname)

    assert isinstance(ifname, str), "ifname={0}; type(ifname)={1}".format(ifname, type(ifname))

    # Early out if the input is not a glob.
    if not ifname_is_glob(ifname):
        return [ifname]

    ifnames = []

    # ex.: swp1,8,9
    if ',' in ifname:
        glob_parts = ifname.split(',')
        base_swp = glob_parts[0]

        if ifupdown2_ifname_is_glob(base_swp):

            # base_swp is a glob.  Expand it, and take the first interface as our new base_swp.
            modbase_expansion = []
            for name in ifupdown2_parse_glob('foobarblah', base_swp):  # ifupdown2_parse_glob returns a generator
                modbase_expansion.append(name.strip())
            base_swp = modbase_expansion[0]
            ifnames.extend(modbase_expansion)

        elif ifname_is_valid(base_swp):
            ifnames.append(base_swp)
        else:
            base_swp = None

        if base_swp:
            base_swp = base_swp.rstrip('1234567890')

            for part in glob_parts[1:]:
                re_swp = re.search(r"""^[a-zA-Z]+\d+(?:[a-zA-Z]+\d+)?$""", part)
                re_number_number = re.search(r"""(\d+)-(\d+)""", part)

                if ifupdown2_ifname_is_glob(part):
                    ifnames.extend(ifupdown2_parse_glob('foobarblah', part))

                elif part.isdigit():
                    ifnames.append(base_swp + part)

                elif re_swp:
                    ifnames.append(part)

                elif re_number_number:
                    min_x = int(re_number_number.group(1))
                    max_x = int(re_number_number.group(2)) + 1

                    for y in xrange(min_x, max_x):
                        ifnames.append(base_swp + str(y))
                else:
                    log.info("unknown glob component '{0}' in ifname_expand_glob()".format(part))

    # swp1-5
    elif ifupdown2_ifname_is_glob(ifname):
        for name in ifupdown2_parse_glob('foobarblah', ifname):
            ifnames.append(name)

    return ifnames


def ifnames_to_glob(interfaces, do_sort=True):
    """
    Given an iterable of strings containing interface names and globs, return
    a single glob.  Numbers in interface names must not begin with zero, be
    globbish, nor be subinterfaces.

    This function is recursive.

    >>> ifnames_to_glob([])
    ''

    >>> ifnames_to_glob(['swp1', 'swp2', 'swp3'])
    'swp1-3'

    >>> ifnames_to_glob(['swp1-3,5'])
    'swp1-3,5'

    TODO - The next test fails because the sorting of interface names is inadequate.
    The current output is 'swp1,swp1s0-1,swp2-3'.

    >>> ifnames_to_glob(['swp1', 'swp1s1', 'swp2,3', 'swp1s0'])
    'swp1-3,swp1s0-1'

    >>> ifnames_to_glob(['vlan100', 'vlan1'])
    'vlan1,100'

    >>> ifnames_to_glob(['vlan99', 'vlan9'])
    'vlan9,99'

    >>> ifnames_to_glob(['vlan'])
    'vlan'

    >>> ifnames_to_glob(['vlan99', 'vlan'])
    'vlan,vlan99'

    >>> ifnames_to_glob(['vlan', 'vlan99'])
    'vlan,vlan99'

    >>> ifnames_to_glob(['vlan2000', 'vlan20'])
    'vlan20,2000'

    >>> ifnames_to_glob(['vlan100', 'swp1', 'swp2', 'vlan1', 'swp3', 'eth0'])
    'eth0,swp1-3,vlan1,100'

    >>> ifnames_to_glob(['vlan100', 'swp1-3', 'vlan1', 'eth0'])
    'eth0,swp1-3,vlan1,100'

    >>> ifnames_to_glob(['server01'])
    Traceback (most recent call last):
        ...
    ValueError: 'server01' cannot have leading zeros.

    >>> ifnames_to_glob(['swp1.1'])
    Traceback (most recent call last):
        ...
    ValueError: 'swp1.1' is a subinterface, which is not permitted.

    >>> ifnames_to_glob(['swp1a'])
    Traceback (most recent call last):
        ...
    ValueError: For interface 'swp1a', suffix '1a' must be a digit.

    >>> ifnames_to_glob(['vlan100', 'swp1', 'vlan', 'swp2', 'vlan1', 'swp3', 'eth0'])
    'eth0,swp1-3,vlan,vlan1,100'
    """

    # Sort and deduplicate the interfaces.  This groups interface names with the same prefix,
    # and prefixes without suffixes come before the same prefix with a suffix.  This is only
    # necessary for the first, non-resursive call.
    if do_sort:
        interfaces = sorted(set(interfaces))  # A list

    output = ''
    ifnumbers = []    # Ex: ['swp1-3', 'swp20'] --> [1, 2, 3, 20]
    prev_prefix = ''

    # Define a regex to partition an interface name into a prefix and suffix.
    ifname_re = re.compile(r"""([^\d\s]+(?:\d+s)?)(\d.*)?""")

    # Define a regex to detect interface names in which the numerical suffix begins with a zero.
    lead_zeros_re = re.compile(r"""[a-zA-Z]+0[1-9]""")

    for index, ifname in enumerate(interfaces):
        if lead_zeros_re.match(ifname) is not None:
            raise ValueError("'{0}' cannot have leading zeros.".format(ifname))

        if '.' in ifname:
            raise ValueError("'{0}' is a subinterface, which is not permitted.".format(ifname))

        is_glob = ifname_is_glob(ifname)

        if not is_glob and ifname_is_globish(ifname, False):
            raise ValueError("'{0}' is globish but not a glob.".format(ifname))

        ifname_match = ifname_re.match(ifname)
        if ifname_match is None:
            # This should never happen because the regex matches anything.
            raise RuntimeError

        prefix = ifname_match.group(1)  # Ex: 'swp'
        suffix = ifname_match.group(2)  # Ex: '1' or None

        if suffix is None:
            # Append any previously parsed data to the output.
            if prev_prefix and ifnumbers:
                if output:
                    output += ','
                output += prev_prefix + numbers_to_glob(ifnumbers)
                ifnumbers = []  # Necessary because this block does not "return"

            # Append the interface as-is to the output.
            if output:
                output += ','
            output += prefix
            continue

        # The interface contains a suffix, which should be a digit or range.

        if not prev_prefix or prefix == prev_prefix:
            if not prev_prefix:
                # This is the first interface with this prefix that has a suffix.
                prev_prefix = prefix
            # Else, this is another interface with the current prefix.  The action is the same.

            # Add the number(s) in the suffix to ifnumbers.
            if is_glob:
                # The suffix looks like "1-3,5".  Note that glob_to_numbers raises an exception if the input contains non-digits.
                ifnumbers.extend(glob_to_numbers(suffix))
            else:
                # The suffix should be a string representation of a digit, like "31".
                if not suffix.isdigit():
                    raise ValueError("For interface '{0}', suffix '{1}' must be a digit.".format(ifname, suffix))
                ifnumbers.append(int(suffix))

        elif prefix != prev_prefix:
            # The prefix name has changed.
            assert prev_prefix

            # Append previously parsed data to the output.
            if ifnumbers:
                if output:
                    output += ','
                output += prev_prefix + numbers_to_glob(ifnumbers)

            # Make a recursive call, but don't resort the interfaces (False parameter).
            return output + ',' + ifnames_to_glob(interfaces[index:], False)

        else:
            # Shouldn't get here.
            raise RuntimeError("Programming error.")

    # Finished iterating through interface names.  Append data, if any, to the output.
    if ifnumbers:
        if output:
            output += ','
        output += prev_prefix + numbers_to_glob(ifnumbers)

    return output


def numbers_to_glob(ids):
    """
    For example, given an iterable:
     (3,4,7,10,11,12,17,22,4001,4002,4003,7777,8000,8001,8002,8003,8004)

    Return a string:
     3-4,7,10-12,17,22,4001-4003,7777,8000-8004

    The parameter, ids, is an iterable that might contain integers, integers as strings, or
    integer ranges as strings.  This method handles all these types.

    >>> numbers_to_glob((100, 1))
    '1,100'

    >>> numbers_to_glob((99, 9))
    '9,99'

    >>> numbers_to_glob((2000, 20))
    '20,2000'

    >>> numbers_to_glob((3,4,7,10,11,12,17,22,4001,4002,4003,7777,8000,8001,8002,8003,8004))
    '3-4,7,10-12,17,22,4001-4003,7777,8000-8004'
    """
    ids_final = []

    # Convert the input, which can contain integers and strings, to a sorted list of integers.
    for _id in ids:  # Note that "id" is a Python built-in function.
        if isinstance(_id, int):
            ids_final.append(_id)
        elif _id.isdigit():
            ids_final.append(int(_id))
        else:
            re_number_range = re.match(r"""(\d+)-(\d+)""", _id)
            if re_number_range is not None:
                start = int(re_number_range.group(1))
                end = int(re_number_range.group(2)) + 1
                if end < start:
                    print "ERROR: invalid numeric range"
                    return []
                ids_final.extend(range(start, end))

    if not ids_final:
        print "ERROR: numbers_to_glob() could not extract any IDs from %s" % pformat(ids)
        return ''

    ids = sorted(set(ids_final))  # A list
    min_id = ids[0]
    max_id = ids[-1]
    limiters = []

    # If the range of integers is continuous (i.e. 1,2,3) . . .
    if len(ids) == (max_id - min_id + 1):
        if min_id == max_id:
            limiters.append(str(min_id))
        else:
            limiters.append("{0}-{1}".format(min_id, max_id))
    # Otherwise, some integers get skipped (i.e. 1,2,5).
    else:
        streak_start_x = 0
        streak_end_x = 0
        len_ids = len(ids)

        for x in range(len_ids):

            # The first number in the list
            if x == 0:
                current = ids[x]
                next_ids = ids[x + 1]

                if current + 1 == next_ids:
                    streak_start_x = current
                else:
                    limiters.append(str(current))

            # The last number in the list
            elif x == len_ids - 1:
                prev_ids = ids[x - 1]
                current = ids[x]

                if prev_ids + 1 == current:
                    streak_end_x = current
                    limiters.append("{0}-{1}".format(streak_start_x, streak_end_x))
                else:
                    limiters.append(str(current))

            # Any number in the middle of the list
            else:
                prev_ids = ids[x - 1]
                current = ids[x]
                next_ids = ids[x + 1]

                # In the middle of a streak...do nothing
                if prev_ids + 1 == current and current + 1 == next_ids:
                    pass

                # The end of a streak
                elif prev_ids + 1 == current:
                    streak_end_x = current
                    limiters.append("{0}-{1}".format(streak_start_x, streak_end_x))
                    streak_start_x = 0
                    streak_end_x = 0

                # The beginning of a streak
                elif current + 1 == next_ids:
                    streak_start_x = current

                # A loner
                else:
                    limiters.append(str(current))

    return ','.join(limiters)


def glob_to_numbers(glob):
    """
    Given a string containing single numbers and ranges, return a sorted
    list of deduplicated integers.

    glob - A string of digits and ranges

    >>> glob_to_numbers('3-4,7,10-12,17,22,4001-4003,7777,8000-8004')
    [3, 4, 7, 10, 11, 12, 17, 22, 4001, 4002, 4003, 7777, 8000, 8001, 8002, 8003, 8004]
    """

    assert isinstance(glob, str), "glob={0}".format(glob)

    # Using split(',') instead of the replacement could yield empty strings in the result.
    glob_list = glob.replace(',', ' ').split()

    numbers = set()
    range_re = re.compile(r"""^(\d+)-(\d+)$""")  # ex. 4-6

    for x in glob_list:
        if x.isdigit():
            numbers.add(int(x))
        else:
            range_match = range_re.match(x)

            if range_match is None:
                # The substring is neither a digit nor a range.
                print "ERROR: Globs must consist of numbers or ranges, but {0} is neither. We were given glob '{1}'.".format(x, glob)
                return []
            else:
                min_range = int(range_match.group(1))
                max_range = int(range_match.group(2))

                if max_range >= min_range:
                    numbers.update(range(min_range, max_range + 1))
                else:
                    print "ERROR: Glob '{0}' contains the invalid range '{1}'.".format(glob, x)  # ex. 6-4
                    return []

    return sorted(numbers)  # A sorted list


def is_number_glob(glob):
    """
    Return True if the input is a numeric glob.  The input must be a string, such as '1-5,10,13-15,20'.
    """

    assert isinstance(glob, str), "glob={0}".format(glob)

    # Using split(',') instead of the replacement could yield empty strings in the result.
    glob = glob.replace(',', ' ').split()  # A list of strings

    # All elements must be string digits or ranges.
    glob_re = re.compile(r"""^\d+(?:-\d+)?$""")
    for x in glob:
        if glob_re.match(x) is None:
            return False

    return True


def get_lldp():
    """
    Return the Link Layer Discovery Protocall.
    """
    lldp = {}

    if os.path.exists('/usr/sbin/lldpctl'):
        try:
            output = check_output(['/usr/sbin/lldpctl', '-f', 'json'])
            lldp_original = json.loads(output).get('lldp')

            if lldp_original and lldp_original[0]:
                for value in lldp_original[0].get('interface'):
                    lldp[value['name']] = value['chassis'][0]['name'][0]['value']

        except Exception as e:
            log.exception(e)

        for ifname in os.listdir('/sys/class/net/'):
            bonding_slaves_filename = "/sys/class/net/%s/bonding/slaves" % ifname

            if os.path.exists(bonding_slaves_filename):

                with open(bonding_slaves_filename) as f:
                    lines = list(f)

                lldp_peers = []

                if lines:
                    firstline = lines[0]

                    if firstline:
                        for bond_slave in firstline.split():
                            if bond_slave in lldp:
                                lldp_peers.append(lldp[bond_slave])

                if lldp_peers:
                    lldp_peers = sorted(list(set(lldp_peers)))
                    lldp[ifname] = ', '.join(lldp_peers)

    return lldp


def tabulate_pending_add_del_commands():
    header = ['User', 'Timestamp', 'Command']
    data = []

    if os.path.exists(PENDING_COMMANDS_FILE):
        with open(PENDING_COMMANDS_FILE) as f:
            for line in f:
                (user, timestamp, command) = line.strip().split('\t')
                data.append((user, timestamp, command))

    return tabulate(tabulate_remove_unicode(data), header)


def print_pending_add_del_commands():
    print "Please run 'net abort' and use net to re-issue your changes."

    if os.path.exists(PENDING_COMMANDS_FILE):
        print "\nThe following is a list of all add/del"
        print "commands since the last 'net commit'."
        print "======================================"

        with open(PENDING_COMMANDS_FILE) as f:
            for line in f:
                (user, timestamp, command) = line.strip().split('\t')
                print command
        print ''


def globify_interface_commands(init_commands):
    """
    >>> globify_interface_commands(['net add int swp1 mtu 1500', 'net add int swp2 mtu 1500', 'net add int swp3 mtu 1500'])
    ['net add int swp1-3 mtu 1500']

    >>> globify_interface_commands(['net add int swp3 mtu 1500', 'net add int Vlan100 mtu 1500', 'net add int swp1 mtu 1500'])
    ['net add int swp1,3,Vlan100 mtu 1500']

    >>> globify_interface_commands(['net add int vlan2', 'net add int vlan1'])
    ['net add int vlan1-2']

    >>> globify_interface_commands(['net add vlan 1 vlan-raw-device VlanA-1'])
    ['net add vlan 1 vlan-raw-device VlanA-1']

    >>> globify_interface_commands(['net add vlan 2 vlan-raw-device VlanA-1', 'net add vlan 1 vlan-raw-device VlanA-1'])
    ['net add vlan 2 vlan-raw-device VlanA-1', 'net add vlan 1 vlan-raw-device VlanA-1']
    """

    final_commands = []              # Return value
    globbed_commands = []
    non_globbed_commands = []
    commands_by_id = OrderedDict()

    for command in init_commands:
        # Ex. 'net add int swp1 mtu 1500' --> ['net', 'add', 'int', 'swp1', 'mtu', '1500']
        command_list = command.split()

        if len(command_list) <= 3:
            # The command does not specify interfaces.  Add it as-is.
            # Ex: net add bridge
            final_commands.append(command)
            continue

        iface_type = command_list[2]  # bond, interface, loopback, vlan, vxlan, etc.

        if iface_type == 'vlan':
            # Add a vlan command as-is.
            # Ex: net add vlan 10-20
            # Ex: net add vlan 1 vlan-raw-device VlanA-1
            final_commands.append(command)
            continue

        add_del = command_list[1]
        ifname = command_list[3]
        command_id_part3 = ' '.join(command_list[4:]) if (len(command_list) > 4) else ''
        command_id = (add_del, iface_type, command_id_part3)

        if command_id not in commands_by_id:
            commands_by_id[command_id] = []
        commands_by_id[command_id].append(ifname)

    # We must first configure the bond-slaves, vlan-interface, and other statements.
    for bond_slaves_etc_first in (True, False):
        for (command_id, ifnames) in commands_by_id.iteritems():
            assert isinstance(ifnames, list)
            (add_del, iface_type, keyword_value) = command_id

            if (keyword_value.startswith('bond-slaves') or
                keyword_value.startswith('vxlan')):
                if not bond_slaves_etc_first:
                    continue
            elif bond_slaves_etc_first:
                continue

            if len(ifnames) > 1:
                try:
                    ifnames_glob = ifnames_to_glob(ifnames)
                except (RuntimeError, ValueError) as e:
                    # Could not globify the interfaces.
                    log.warning("(globify_interface_commands) '{0}' raised an exception in ifnames_to_glob: {1}.".format(str(ifnames), e))
                    ifnames_glob = None
            else:
                ifnames_glob = None

            if ifnames_glob:
                globbed_commands.append("net {0} {1} {2} {3}".format(add_del, iface_type, ifnames_glob, keyword_value).strip())
            else:
                for ifname in ifnames:
                    non_globbed_commands.append("net {0} {1} {2} {3}".format(add_del, iface_type, ifname, keyword_value).strip())

    final_commands.extend(globbed_commands)
    final_commands.extend(non_globbed_commands)
    return final_commands


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def seconds_to_hhmmss(secs):
    # TODO - Delete code by using the Python standard library "time" package?
    if not secs:
        return 'never'

    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)
    days, hours = divmod(hours, 24)

    if days:
        return '{0} day{1}, {2:0>2}:{3:0>2}:{4:0>2}'.format(days, 's' if days > 1 else '',  hours, mins, secs)
    else:
        return '{0:0>2}:{1:0>2}:{2:0>2}'.format(hours, mins, secs)


def get_hostname(ip):
    try:
        hostname = gethostbyaddr(ip)
        return " ({0})".format(hostname[0])
    except Exception:
        return ""


def file_set_key_value(filename, keyword, new_value, comment_out=False):
    """
    For config files that follow a key=value notation (i.e. ip = 10.1.1.1), this function
    provides a way to alter the value for a key.
    """

    if not os.path.exists(filename):
        log.info("Could not set {0} to {1} in {2}.  The file does not exist.".format(keyword, new_value, filename))
        return

    new_content = []

    with open(filename) as f:
        for line in f:
            # This regex matches the case when a key is present but not assigned a value (i.e. key=<blank> # Comment).
            re_key_no_value = re.search(r'''^(?:\s*#)?(\s*{0}\s*)=[\t ]*(#[\S\t ]*)?$'''.format(keyword), line, re.IGNORECASE)

            if re_key_no_value:
                before_equals = re_key_no_value.group(1)
                comment = re_key_no_value.group(2)
                line = ('#{0}= {1}' if comment_out else '{0}= {1}').format(before_equals, new_value)
                if comment:
                    line += ' ' + comment
                line += '\n'
            else:
                # This regex matches the case when a key is present and assigned a value (i.e. key=value # Comment).
                re_key_value = re.search(r'''^(?:\s*#)?(\s*{0}\s*=\s*)(?:".*"|\S[^#\s]*)([\t ]*#[\S\t ]*)?$'''.format(keyword), line, re.IGNORECASE)

                if re_key_value:
                    before_value = re_key_value.group(1)
                    comment = re_key_value.group(2)
                    line = ('#{0}{1}' if comment_out else '{0}{1}').format(before_value, new_value)
                    if comment:
                        line += comment
                    line += '\n'

            new_content.append(line)

    with open(filename, 'w') as f:
        f.write(''.join(new_content))


def is_mac_address(word):
    return (re.match('(?:[0-9A-Fa-f]{2}[:-]){5}[0-9A-Fa-f]{2}', word) or
            re.match('[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}\.[0-9A-Fa-f]{4}', word))


class ConfigFile(object):

    def __init__(self, parent, filename, cmds):
        """
        'parent' is the ConfigWrapper class that owns us.

        TODO - Passing the parent class is not necessary.
        http://stackoverflow.com/questions/10091957/get-parent-class-name#10092179

        'cmds' is a list of commands to run if the configuration file is modified.
        """

        self.parent = parent
        self.filename = filename.strip()

        filename_underscore = self.filename.replace('/', '_').replace('.', '_').strip('_')
        self.scratchpad = os.path.join(self.parent.var_run_nclu, "%s_scratchpad" % filename_underscore)
        self.backup = os.path.join(self.parent.var_run_nclu, "%s_backup" % filename_underscore)
        self.cmds = cmds

    def __str__(self):
        return self.filename

    def remove_backup(self):
        if os.path.exists(self.backup):
            os.remove(self.backup)

    def remove_scratchpad(self):
        if os.path.exists(self.scratchpad):
            os.remove(self.scratchpad)


class ConfigFileSingleLine(ConfigFile):
    """
    For config files with a single line
    """

    def edit(self, value):
        """
        Return True if the file was modified
        """
        current_value = self.get_value(True)
        if current_value == value:
            print "{0} already has {1}".format(self.scratchpad, value)
            return False
        else:
            with open(self.scratchpad, 'w') as f:
                f.write(value + '\n')
            return True

    def get_value(self, read_scratchpad=False):
        """
        This file should only contain one line. If it happens to contain more
        than one return the first valid (not blank and not a comment) and log
        a warning.
        """
        result = None
        filename = self.scratchpad if read_scratchpad else self.filename

        if os.path.exists(filename):
            with open(filename) as f:
                for line in f.readlines():
                    line = line.strip()

                    if line and not line.startswith('#'):
                        if result is None:
                            result = line
                        else:
                            log.warning("{0} contains multiple values.  It should only contain one value.".format(filename))

        return result


class ConfigFileKeyValue(ConfigFile):

    def edit(self, key, value, comment_out=False):
        """
        Return True if the file was modified.
        """
        current_value = self.get_value(True)  # A dictionary

        if current_value.get(key) == value:
            print "{0} already has '{1} = {2}'".format(self.scratchpad, key, value)
            return False
        else:
            file_set_key_value(self.scratchpad, key, value, comment_out)
            return True

    def get_value(self, read_scratchpad=False):
        filename = self.scratchpad if read_scratchpad else self.filename
        results = {}

        if os.path.exists(filename):
            with open(filename) as f:
                for line in f:
                    re_key_value = re.search(r'''^\s*(\w+)\s*=\s*"?\s*(.*?)\s*"?\s*$''', line, re.IGNORECASE)

                    if re_key_value:
                        key = re_key_value.group(1)
                        value = re_key_value.group(2)
                        results[key] = value
        return results


class ConfigWrapperBase(object):

    def __init__(self, name, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs
        self.name = name
        self.var_run_nclu = "/var/run/nclu/{0}".format(self.name)

        if not os.path.exists(self.var_run_nclu):
            mkdir_p(self.var_run_nclu)

        self.files = {}  # The keys are file names; the values are ConfigFile instances (including subclasses).

    def __str__(self):
        return self.name

    def add_config_file(self, filename, cmd):
        self.files[filename] = ConfigFile(self, filename, cmd)

    def get_managed_files(self):
        return [x.filename for x in self.files.itervalues()]

    def get_service_cmd_for_file(self, filename):
        cfg = self.files.get(filename)
        return cfg.cmds if cfg else []

    def abort_pending(self):
        if os.path.exists(self.var_run_nclu):
            for filename in os.listdir(self.var_run_nclu):
                full_filename = os.path.join(self.var_run_nclu, filename)

                if os.path.isdir(full_filename):
                    shutil.rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def del_all(self, cli):
        raise NotImplementedError("Subclasses should override the del_all method.")

    def commit_pending(self, verbose):
        files_modified = []
        commit_ok = True
        cmds = []

        # Copy scratchpad files over the real config files
        for cfg in self.files.itervalues():
            if os.path.exists(cfg.scratchpad) and not files_match(cfg.filename, cfg.scratchpad):
                shutil.copy(cfg.filename, cfg.backup)
                shutil.copy(cfg.scratchpad, cfg.filename)
                files_modified.append(cfg)
                cmds.extend(cfg.cmds)

        # Run all commands that are needed to apply the changes in the new config files
        for (cmd_index, cmd) in enumerate(cmds):
            try:
                output = check_output(cmd, stderr=STDOUT)

                if verbose:
                    print output

            except Exception as e:
                print "\nERROR: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)
                commit_ok = False
                break

        if commit_ok:
            # Remove scratchpad and backup files
            for cfg in files_modified:
                cfg.remove_scratchpad()
                cfg.remove_backup()

            return (True, [cfg.filename for cfg in files_modified])

        else:
            # Restore the config file from our backup
            for cfg in files_modified:
                shutil.move(cfg.backup, cfg.filename)

            # Rerun the commands we ran to apply config changes
            for cmd in cmds[cmd_index]:
                try:
                    output = check_output(cmd, stderr=STDOUT)

                    if verbose:
                        print output

                except Exception as e:
                    print "\nERROR: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)

            return (False, [])

    def eval_command_line_args(self, cli):
        # If we are doing an add or del, we must create the scratchpad file.
        if cli.args.get('add') or cli.args.get('del'):
            for cfg in self.files.itervalues():
                if not os.path.exists(cfg.scratchpad):
                    shutil.copyfile(cfg.filename,  cfg.scratchpad)

        return True
        # The child class will need to do additional work in its eval_command_line_args().
        # The overrided function should return True to inform the calling block that the
        # command was successfully executed.  Among other consequences of not returning True,
        # the command will not show up among the pending commands in the output of "net pending".

    def show_pending(self, print_errors=True):
        diff_pending = False

        for cfg in self.files.itervalues():
            if os.path.exists(cfg.scratchpad):
                if print_diff(cfg.filename, cfg.scratchpad, self.color_diffs):
                    diff_pending = True

        return diff_pending

    def show_config_files(self, user_may_edit):

        for cfg in self.files.itervalues():
            if os.path.exists(cfg.filename) and os.path.getsize(cfg.filename) > 0:
                print cfg.filename
                print '=' * len(cfg.filename)
                print_file(cfg.filename)

    def show_config_commands(self, user_may_edit):
        raise NotImplementedError("Subclasses should override the show_config_commands method.")

    def show_config_summary(self, user_may_edit, summary):
        raise NotImplementedError("Subclasses should override the show_config_summary method.")


def get_interfaces():
    """
    Get a full list by combining the interfaces in /etc/network/interfaces
    along with any that are pending.
    """
    ifnames = set()

    if os.path.exists('/var/run/nclu/iface/'):
        for filename in os.listdir('/var/run/nclu/iface/'):
            if not filename.endswith('.original') and not filename.endswith('.del'):
                ifnames.add(filename)

    if os.path.exists('/etc/network/interfaces'):
        with open('/etc/network/interfaces') as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith('iface '):
                    ifname = line.split()[1]
                    ifnames.add(ifname)

    if os.path.exists('/sys/class/net/'):
        for ifname in os.listdir('/sys/class/net/'):
            if ifname not in ('swid0_eth', 'bonding_masters') and not ifname.endswith('-v0'):
                ifnames.add(ifname)

    return sort_for_humans(list(ifnames))


def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, unicode):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = decode_dict(item)
        rv.append(item)
    return rv


def decode_dict(data):
    rv = {}
    for key, value in data.iteritems():
        if isinstance(key, unicode):
            key = key.encode('utf-8')
        if isinstance(value, unicode):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = decode_dict(value)
        rv[key] = value
    return rv


def service_is_active(name):
    try:
        output = check_output(['/bin/systemctl', 'is-active', name]).splitlines()
        if output and output[0] == 'active':
            return True
        return False
    except CalledProcessError:
        return False


def tabulate_remove_unicode_helper(datum):
    """
    >>> tabulate_remove_unicode_helper('Go Torros!')
    'Go Torros!'

    >>> tabulate_remove_unicode_helper(u'Jos\xe9')
    'Jos?'

    This test case works manually in the interpreter but not here.  ???
#    >>> tabulate_remove_unicode_helper('Jos\xe9')
#    'Jos\xe9'
    """

    # "tabulate" can take input that is not an iterable-of-iterables, but we're not
    # using the other options, such as iterables of dictionaries.  Don't permit them.
    #     https://pypi.python.org/pypi/tabulate
    assert not isinstance(datum, dict)

    return datum.encode('ascii', 'replace') if isinstance(datum, unicode) else datum


def tabulate_remove_unicode(data):
    """
    CM-16979 demonstrated that unicode input in the first parameter to the
    third party "tabulate" function raises UnicodeEncodeError.  This
    function takes an iterable of iterables, congruent with the first
    "tabulate" parameter, and returns the input with any unicode removed.

    >>> tabulate_remove_unicode([('yo', 'hey', u'Jos\xe9')])
    [['yo', 'hey', 'Jos?']]
    """

    # Return a list of lists.
    return [map(tabulate_remove_unicode_helper, datum) for datum in data]


if __name__ == '__main__':
    import doctest
    doctest.testmod()
