# Copyright (C) 2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import logging
import re
import os

from collections import OrderedDict
from difflib import Differ
from nclu import ConfigFile, ConfigWrapperBase
from os.path import exists
from shutil import copy
from stat import S_IXUSR, S_IXGRP, S_IXOTH  # Flags used to set file permissions for executibility
from subprocess import check_output

log = logging.getLogger('netd')


def build_doc_string():
    global __doc__

    __doc__ = """
Usage:
    dns add dns nameserver (ipv4 <ipv4>|ipv6 <ipv6>) [vrf <interface>]
    dns del dns nameserver (ipv4 <ipv4>|ipv6 <ipv6>)
    net show configuration dns

Options:
    nameserver    : An IP address of a remote DNS server
    vrf           : Virtual routing and forwarding
"""


class ConfigWrapper(ConfigWrapperBase):
    """
    Use this wrapper to specify DNS parameters.

    http://man7.org/linux/man-pages/man5/resolv.conf.5.html
    """
    RESOLV = '/etc/resolv.conf'
    NO_DNS_UPDATE = '/etc/dhcp/dhclient-enter-hooks.d/nodnsupdate'

    def __init__(self, show_linux_command, color_diffs):
        super(ConfigWrapper, self).__init__('dns', show_linux_command, color_diffs)

        # Create the DNS configuration file, if necessary.
        if not exists(ConfigWrapper.RESOLV):
            # Create an empty file.
            open(ConfigWrapper.RESOLV, 'w').close()

        self.files[ConfigWrapper.RESOLV] = ConfigFile(self, ConfigWrapper.RESOLV, [])

    def create_no_dns_update_file(self):
        # Create /etc/dhcp/dhclient-enter-hooks.d/nodnsupdate so DHCP doesn't overwrite
        # users' changes when the device is rebooted.
        #     https://www.vultr.com/docs/stop-dhcp-from-changing-resolve-conf
        if not exists(ConfigWrapper.NO_DNS_UPDATE):
            with open(ConfigWrapper.NO_DNS_UPDATE, 'w') as f:
                f.write('#!/bin/sh\n')
                f.write('make_resolv_conf(){\n')
                f.write('    :\n')
                f.write('}\n')

            # Make the file executable.
            existing_permissions = os.stat(ConfigWrapper.NO_DNS_UPDATE).st_mode
            os.chmod(ConfigWrapper.NO_DNS_UPDATE, existing_permissions | S_IXUSR | S_IXGRP | S_IXOTH)

    def eval_command_line_args(self, cli):
        super(ConfigWrapper, self).eval_command_line_args(cli)  # Creates the scratchpad, if necessary.

        if cli.args.get('add') and cli.args.get('dns') and cli.args.get('nameserver'):
            self.add_dns_nameserver(cli)
        elif cli.args.get('del') and cli.args.get('dns') and cli.args.get('nameserver'):
            self.del_dns_nameserver(cli)
        else:
            raise RuntimeError("Unrecognized DNS command.")

        return True  # Returning True indicates that the command was successfully processed.  This has consequences downstream.

    def add_dns_nameserver(self, cli):
        """
        dns add dns nameserver (ipv4 <ipv4>|ipv6 <ipv6>) [vrf <interface>]

        Add a nameserver IP address to the configuration file.
        """
        ip = cli.args.get('<ipv4>') if (cli.args.get('ipv4') is not None) else cli.args.get('<ipv6>')

        # Make sure the IP isn't already in the configuration file.
        with open(self.files[ConfigWrapper.RESOLV].scratchpad) as f:
            if re.search('^\s*nameserver\s+{0}'.format(ip), f.read(), re.IGNORECASE | re.MULTILINE) is not None:
                return  # Do nothing because the IP is already present.

        # Append a new nameserver line to the configuration file.
        with open(self.files[ConfigWrapper.RESOLV].scratchpad, 'a') as f:
            if cli.args.get('vrf'):
                vrf = cli.args.get('<interface>')
                f.write('nameserver {0} # vrf {1}\n'.format(ip, vrf))
            else:
                f.write('nameserver {0}\n'.format(ip))

    def del_dns_nameserver(self, cli):
        """
        dns del dns nameserver (ipv4 <ipv4>|ipv6 <ipv6>)

        Remove a "nameserver" line containing the given IP address.
        """
        # Input the configuration scratchpad file.
        with open(self.files[ConfigWrapper.RESOLV].scratchpad) as f:
            config_file = f.read()

        # Remove the given IP from the configuration file text.
        ip = cli.args.get('<ipv4>') if (cli.args.get('ipv4') is not None) else cli.args.get('<ipv6>')
        config_file = re.sub('^\s*nameserver\s+{0}.*?$'.format(ip), '', config_file, 0, re.IGNORECASE | re.MULTILINE)

        # Output the modified configuration text.
        with open(self.files[ConfigWrapper.RESOLV].scratchpad, 'w') as f:
            f.write(config_file)

    def del_all(self, cli):
        """
        net del all

        Remove all "nameserver" entries from the DNS configuration file other than the first entry,
        and remove the NO_DNS_UPDATE file.
        """
        if not exists(self.files[ConfigWrapper.RESOLV].scratchpad):
            copy(self.files[ConfigWrapper.RESOLV].filename, self.files[ConfigWrapper.RESOLV].scratchpad)

        with open(self.files[ConfigWrapper.RESOLV].scratchpad) as f:
            config_file = f.read()
        nameservers = re.findall(r'''^\s*nameserver\s+[\d.:a-f]+.*?$''', config_file, re.IGNORECASE | re.MULTILINE)
        if len(nameservers) > 1:
            # Delete every nameserver entry after the first.
            for nameserver in nameservers[1:]:
                config_file = re.sub(nameserver, '', config_file, 0, re.MULTILINE)

            # Output the modified file.
            with open(self.files[ConfigWrapper.RESOLV].scratchpad, 'w') as f:
                f.write(config_file)

        if exists(ConfigWrapper.NO_DNS_UPDATE):
            os.remove(ConfigWrapper.NO_DNS_UPDATE)

    def show_config_commands(self, user_may_edit):
        """
        net show configuration commands
        """
        if not exists(ConfigWrapper.RESOLV):
            return ([], [])  # Nothing to do.

        # Read the "nameserver" parameter lines from the configuration file.
        with open(ConfigWrapper.RESOLV) as f:
            servers = re.findall(r'''^\s*nameserver\s+([.:\da-f]+)(?:\s+#\s+vrf\s+(\w+))?''', f.read(), re.IGNORECASE | re.MULTILINE)

        commands = []

        for ip, vrf in servers:
            if '.' in ip:
                # IPv4
                if vrf:
                    commands.append('net add dns nameserver ipv4 {0} vrf {1}'.format(ip, vrf))
                else:
                    commands.append('net add dns nameserver ipv4 {0}'.format(ip))
            else:
                # IPv6
                if vrf:
                    commands.append('net add dns nameserver ipv6 {0} vrf {1}'.format(ip, vrf))
                else:
                    commands.append('net add dns nameserver ipv6 {0}'.format(ip))

        return (commands, [])

    def show_config_summary(self, user_may_edit, summary):
        """
        net show configuration dns
        """
        if not exists(ConfigWrapper.RESOLV):
            return  # Nothing to do.

        with open(ConfigWrapper.RESOLV) as f:
            servers = re.findall(r'''^\s*nameserver\s+([.:\da-f]+(?:\s+#\s+vrf\s+\w+)?)''', f.read(), re.IGNORECASE | re.MULTILINE)

        if servers:
            summary['dns'] = OrderedDict()
            summary['dns']['nameserver'] = servers

    def commit_pending(self, verbose):
        """
        In addition to copying a scratchpad to a configuration file, committing changes to
        DNS might necessitate running an executable that ensures VRF rules (FIBs) are updated.
        This is the case when the scratchpad adds or removes nameserver lines with vrf information.
        """
        if os.path.exists(self.files[ConfigWrapper.RESOLV].scratchpad):
            with open(self.files[ConfigWrapper.RESOLV].scratchpad) as f:
                scratchpad_text = f.read()
        else:
            return (True, [])  # There was nothing to commit

        if os.path.exists(ConfigWrapper.RESOLV):
            with open(ConfigWrapper.RESOLV) as f:
                resolv_conf_text = f.read()
        else:
            resolv_conf_text = ''

        diff_lines = Differ().compare(resolv_conf_text.splitlines(True), scratchpad_text.splitlines(True))  # A list of strings with endlines

        # Get affected VRFs, if any.  This must be done before calling the parent class method, which deletes the scratchpad.
        vrfs_to_add = []
        vrfs_to_del = []
        stuff_to_add = False
        stuff_to_del = 0
        nameserver_diff_re = re.compile(r'''(?P<action>-|\+)\s*nameserver\s+(?P<ip>[.:\da-f]+)(?:\s+#\s+vrf\s+(?P<vrf>\w+))?''', re.IGNORECASE)

        for line in diff_lines:
            nameserver_diff = nameserver_diff_re.match(line)
            if nameserver_diff is not None:
                if nameserver_diff.group('action') == '+':
                    stuff_to_add = True
                    if nameserver_diff.group('vrf'):  # Neither None nor the empty string
                        vrfs_to_add.append((nameserver_diff.group('ip'), nameserver_diff.group('vrf')))
                else:  # '-'
                    stuff_to_del += 1
                    if nameserver_diff.group('vrf'):  # Neither None nor the empty string
                        vrfs_to_del.append((nameserver_diff.group('ip'), nameserver_diff.group('vrf')))

        if stuff_to_add:
            # The user added a nameserver.  Make sure this change isn't overwritten by DHCP upon rebooting.
            self.create_no_dns_update_file()  # Create the NO_DNS_UPDATE file, if it does not exist.
        elif stuff_to_del and (resolv_conf_text.count('nameserver') - stuff_to_del) == 0:
            assert not ((resolv_conf_text.count('nameserver') - stuff_to_del) < 0)
            # The user deleted the last nameserver entry.  Remove the NO_DNS_UPDATE file.
            if exists(ConfigWrapper.NO_DNS_UPDATE):
                os.remove(ConfigWrapper.NO_DNS_UPDATE)

        # Write the scratchpad to RESOLV_CONF, and delete the scratchpad, using the parent class method.
        success, modified_files = super(ConfigWrapper, self).commit_pending(verbose)

        if success:
            # Call executables to update VRF rules (FIBs).
            for ip, vrf in vrfs_to_add:
                check_output(['/usr/lib/vrf/vrf-dns-helper', 'dns_add', ip, vrf])
            for ip, vrf in vrfs_to_del:
                check_output(['/usr/lib/vrf/vrf-dns-helper', 'dns_del', ip, vrf])

        # Return the tuple returned by the parent class method.
        return (success, modified_files)
