#!/usr/bin/env python

# Copyright (C) 2016-2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    # dot1x RADIUS commands
    # FIXME add IPv6 later dot1x (add|del) dot1x radius server-ip <ipv4>|<ipv6>
    dot1x (add|del) dot1x radius server-ip <ipv4>
    dot1x add       dot1x radius authentication-port (default|<1-65535>)
    dot1x add       dot1x radius accounting-port (default|<1-65535>)
    dot1x add       dot1x radius accounting-port (default|<1-65535>)
    dot1x (add|del) dot1x radius shared-secret <text>
    dot1x add       dot1x mab-activation-delay (default|<5-30>)
    dot1x add       dot1x eap-reauth-period (default|<0-86400>)
    dot1x (add|del) dot1x parking-vlan-id <1-4094>

    # dot1x del all configs and stop hostapd
    dot1x del dot1x all

    # dot1x interface enabling or disabling commands
    dot1x (add|del) interface <interface> dot1x
    dot1x (add|del) interface <interface> dot1x  mab
    dot1x (add|del) interface <interface> dot1x  mab parking-vlan
    dot1x (add|del) interface <interface> dot1x  parking-vlan
    dot1x (add|del) interface <interface> dot1x  parking-vlan mab

    # dot1x show commands
    dot1x show dot1x [json]
    dot1x show dot1x status [json]
    dot1x show dot1x macs [json]
    dot1x show dot1x macs <interface> [json]
    dot1x show dot1x port-details [json]
    dot1x show dot1x port-details <interface> [json]
    dot1x show dot1x radius-details [json]
    dot1x show dot1x radius-details <interface> [json]
    dot1x example dot1x
    net show configuration dot1x

Options:
    dot1x : Configure, Enable, Delete or Show IEEE 802.1X EAPOL
    mab : Enable MAC Authentication Bypass with IEEE 802.1X EAPOL
    parking-vlan : Enable a IEEE 802.1X Parking VLAN for non-Authorized MAC Addresses.
    all : Everything
    stations : Show authenticated MAC addresses per interface
    macs : Media Access Control addresses
    port-details : Show counters from the IEEE8021-PAE-MIB for ports
    radius-details : Show counters from the RADIUS-CLIENT MIB (RFC 2618) for ports
    radius : IEEE 802.1X RADIUS Server Configuration
    authentication-port : RADIUS authentication port (default: 1812)
    mab-activation-delay : Seconds to wait before starting MAC Authentication Bypass (default: 30)
    eap-reauth-period : EAP Reauthentication period in seconds (0 to disable, default:0)
    parking-vlan-id : Configure a Isolated Parking VLAN ID for non-Authorized MAC Addresses.
    accounting-port : RADIUS accounting parameters (default: 1813)
    server-ip : Server IPv4 or IPv6 Address (no default, required)
    shared-secret : Configure RADIUS Shared Secret (no default, required)

"""

from nclu import (
    ifname_expand_glob,
    ifname_is_glob,
    ifname_is_globish,
    ConfigWrapperBase,
    PENDING_COMMIT_NOCHANGE,
    PENDING_COMMIT_CHANGE,
    PENDING_COMMIT_ERROR,
    ifname_is_valid,
    print_diff,
    tabulate_remove_unicode as tru
)

from collections import OrderedDict
from subprocess import check_output, CalledProcessError
from tabulate import tabulate
import json
import logging
import logging.handlers
import os
import re
import shutil
import glob

log = logging.getLogger('netd')

class ConfigWrapper(ConfigWrapperBase):
    def __init__(self, show_linux_command, color_diffs):
        self.color_diffs = color_diffs
        self.show_linux_command = show_linux_command
        self.commands = None
        # we can have only one value, e.g. ip addres, secret, port, interface
        self.values = None
        # interface may need to be expanded
        self.ifnames = None
        self.dot1x_scratchpad     = '/var/run/nclu/dot1x/'
        self.scratch_config_file  = '/var/run/nclu/dot1x/hostapd.conf'
        self.temp_config_file  = '/var/run/nclu/dot1x/temp_hostapd.conf'
        self.scratch_pending_delete_file  = '/var/run/nclu/dot1x/pending_delete'
        self.pending_config_file = '/var/run/nclu/dot1x/pending_config.json'
        self.config_file       = '/etc/hostapd.conf'
        self.config_attrs = ['eap_server',
                             'ieee8021x',
                             'driver',
                             'interfaces',
                             'mab_interfaces',
                             'parking_vlan_interfaces',
                             'parking_vlan_id',
                             'mab_activation_delay',
                             'eap_reauth_period',
                             'ctrl_interface',
                             'nas_identifier',
                             'auth_server_addr',
                             'auth_server_port',
                             'auth_server_shared_secret',
                             'acct_server_addr',
                             'acct_server_port',
                             'acct_server_shared_secret']
        self.config_defaults = OrderedDict([('ctrl_interface', '/var/run/hostapd'),
                                            ('eap_server', '0'),
                                            ('ieee8021x', '1'),
                                            ('driver', 'wired'),
                                            ('interfaces', None),
                                            ('mab_interfaces', None),
                                            ('parking_vlan_interfaces', None),
                                            ('parking_vlan_id', None),
                                            ('mab_activation_delay', '30'),
                                            ('eap_reauth_period', '0'),
                                            ('nas_identifier', 'localhost'),
                                            ('auth_server_addr', None),
                                            ('auth_server_port', '1812'),
                                            ('auth_server_shared_secret', None),
                                            ('acct_server_addr', None),
                                            ('acct_server_port', '1813'),
                                            ('acct_server_shared_secret', None)])
        # attributes the user can actually change
        self.iface_array = {'interfaces' : {'arg' : '<interface>',
                                         'config_list' : ['interfaces', 'dot1x'],
                                         'default' : None},
                            'mab_interfaces' : {'arg' : '<interface>',
                                                'config_list' : ['interfaces', 'mab'],
                                                'default' : None},
                            'parking_vlan_interfaces' : {'arg' : '<interface>',
                                                         'config_list' :
                                                         ['interfaces', 'parking-vlan'],
                                                         'default' : None}
                           }
        self.config_array =  {'mab_activation_delay' : {'arg' : '<5-30>',
                                                        'cli_name' : 'mab-activation-delay',
                                                        'default' : '30'},
                              'parking_vlan_id' : {'arg' : '<1-4094>',
                                                   'cli_name' : 'parking-vlan-id',
                                                   'default' : None},
                              'eap_reauth_period' : {'arg' : '<0-86400>',
                                                     'cli_name' : 'eap-reauth-period',
                                                     'default' : '0'},
                              'auth_server_addr' : {'arg' : '<ipv4>',
                                                    'cli_name' : 'server-ip',
                                                    'default' : None},
                              'auth_server_port' : {'arg' : '<1-65535>',
                                                    'cli_name' : 'authentication-port',
                                                    'default' : '1812'},
                              'auth_server_shared_secret' : {'arg' : '<text>',
                                                             'cli_name' : 'shared-secret',
                                                             'default' : None},
                              'acct_server_addr' : {'arg' : '<ipv4>',
                                                    'cli_name' : 'server-ip',
                                                    'default' : None},
                              'acct_server_port' : {'arg' : '<1-65535>',
                                                    'cli_name' : 'accounting-port',
                                                    'default' : '1813'},
                              'acct_server_shared_secret' : {'arg' : '<text>',
                                                             'cli_name' : 'shared-secret',
                                                             'default' : None}
                             }
        self.cli_array =  {'dot1x' : {'arg' : '<interface>',
                                      'config_list' : ['interfaces'],
                                      'default' : None},
                           'mab' : {'arg' : '<interface>',
                                          'config_list' : ['mab_interfaces'],
                                          'default' : None},
                           'server-ip' : {'arg' : '<ipv4>',
                                          'config_list' : ['auth_server_addr',
                                                           'acct_server_addr'],
                                          'default' : None},
                           'mab-activation-delay' : {'arg' : '<5-30>',
                                                     'config_list' :
                                                     ['mab_activation_delay'],
                                                     'default' : '30'},
                           'parking-vlan-id' : {'arg' : '<1-4094>',
                                                'config_list' :
                                                ['parking_vlan_id'],
                                                'default' : None},
                           'eap-reauth-period' : {'arg' : '<0-86400>',
                                                  'config_list' :
                                                  ['eap_reauth_period'],
                                                  'default' : '0'},
                           'authentication-port' : {'arg' : '<1-65535>',
                                                    'config_list' : ['auth_server_port'],
                                                    'default' : '1812'},
                           'accounting-port' : {'arg' : '<1-65535>',
                                                'config_list' : ['acct_server_port'],
                                                'default' : '1813'},
                           'shared-secret' : {'arg' : '<text>',
                                              'config_list' : ['auth_server_shared_secret',
                                                               'acct_server_shared_secret'],
                                              'default' : None}
                          }
        self.radius_detail_list = ['radiusAuthServerIndex',
                                   'radiusAuthServerAddress',
                                   'radiusAuthClientServerPortNumber',
                                   'radiusAuthClientRoundTripTime',
                                   'radiusAuthClientAccessRequests',
                                   'radiusAuthClientAccessRetransmissions',
                                   'radiusAuthClientAccessAccepts',
                                   'radiusAuthClientAccessRejects',
                                   'radiusAuthClientAccessChallenges',
                                   'radiusAuthClientMalformedAccessResponses',
                                   'radiusAuthClientBadAuthenticators',
                                   'radiusAuthClientPendingRequests',
                                   'radiusAuthClientTimeouts',
                                   'radiusAuthClientUnknownTypes',
                                   'radiusAuthClientPacketsDropped',
                                   'radiusAccServerIndex',
                                   'radiusAccServerAddress',
                                   'radiusAccClientServerPortNumber',
                                   'radiusAccClientRoundTripTime',
                                   'radiusAccClientRequests',
                                   'radiusAccClientRetransmissions',
                                   'radiusAccClientResponses',
                                   'radiusAccClientMalformedResponses',
                                   'radiusAccClientBadAuthenticators',
                                   'radiusAccClientPendingRequests',
                                   'radiusAccClientTimeouts',
                                   'radiusAccClientUnknownTypes',
                                   'radiusAccClientPacketsDropped']
        self.port_detail_list = ['dot1xPaePortProtocolVersion',
                                 'dot1xAuthPaeState',
                                 'dot1xAuthBackendAuthState',
                                 'dot1xAuthQuietPeriod',
                                 'dot1xAuthServerTimeout',
                                 'dot1xAuthReAuthPeriod',
                                 'dot1xAuthReAuthEnabled',
                                 'dot1xAuthEapolFramesRx',
                                 'dot1xAuthEapolFramesTx',
                                 'dot1xAuthEapolStartFramesRx',
                                 'dot1xAuthEapolLogoffFramesRx',
                                 'dot1xAuthEapolRespIdFramesRx',
                                 'dot1xAuthEapolRespFramesRx',
                                 'dot1xAuthEapolReqIdFramesTx',
                                 'dot1xAuthEapolReqFramesTx',
                                 'dot1xAuthInvalidEapolFramesRx',
                                 'dot1xAuthEapLengthErrorFramesRx',
                                 'dot1xAuthLastEapolFrameVersion',
                                 'dot1xAuthLastEapolFrameSource',
                                 'dot1xAuthAuthSuccessesWhileAuthenticating',
                                 'dot1xAuthAuthTimeoutsWhileAuthenticating',
                                 'dot1xAuthAuthFailWhileAuthenticating',
                                 'dot1xAuthSessionId',
                                 'dot1xAuthSessionAuthenticMethod',
                                 'dot1xAuthSessionTime',
                                 'dot1xAuthSessionTerminateCause',
                                 'dot1xAuthSessionUserName',
                                 'authMultiSessionId',
                                 'last_eap_type_as',
                                 'last_eap_type_sta',
                                 'connected_time']

    def convert(self, str):
        return int("".join(re.findall("\d*", str)))

    def do_show_config(self):
        '''
        This shows the config files in various forms.
        '''
        # if an interface is given, only show information for that interface
        running_config = self.get_config(self.config_file)
        if self.args.get('json'):
            print(json.dumps(running_config, sort_keys=True, indent=4))
        else:
            headers = ["Attribute", "Value"]
            table = []
            # need to show the existing running configs
            # all the radius configs should be the same
            # show running radius configs
            ifaces = running_config.get('interfaces', [])
            if '' in ifaces:
                ifaces.remove('')
            if len(ifaces) > 1:
                ifaces = sorted(ifaces, key=self.convert)
                ifaces = ','.join(ifaces)
            if not ifaces:
                ifaces = None
            table.append(['dot1x radius server-ip',
                          running_config.get('auth_server_addr')])
            table.append(['dot1x radius authentication-port',
                          running_config.get('auth_server_port', 1812)])
            table.append(['dot1x radius accounting-port',
                          running_config.get('acct_server_port', 1813)])
            table.append(['dot1x radius shared-secret',
                          running_config.get('auth_server_shared_secret', '')])
            table.append(['dot1x enabled interfaces', ifaces])
            print tabulate(tru(table), headers, tablefmt="simple")
        return True

    def three_column_results(self, command=None, attribute='Mac Addresses', mylist=[]):
        if self.args.get('<interface>'):
            running_ifname_set = self.get_ifnames(self.args.get('<interface>'))
        else:
            # we are only showing running interface info
            running_ifname_set = self.get_running_ifnames(config_file=self.config_file)
        details = {}
        for ifname in running_ifname_set:
            details[ifname] = {}
            try:
                result = check_output(['/usr/sbin/hostapd_cli', '-i', ifname,
                                       command])
                if not result:
                    details[ifname] = {'No Data' : ''}
                    result = ''
            except:
                details[ifname] = {'802.1x not enabled on interface' : ''}
                result = ''
            for line in result.split('\n'):
                if '=' in line:
                    a,b = line.split('=')
                    if a in mylist:
                        details[ifname][a] = b
                elif line and ':' in line:
                    # MACs are just a line with the mac
                    if details[ifname].get(attribute, None):
                        details[ifname][attribute].append(line)
                    else:
                        details[ifname][attribute] = [line]

        if self.args.get('json'):
            print(json.dumps(details, sort_keys=True, indent=4))
        else:
            headers = ["Interface", "Attribute", "Value"]
            table = []
            ifnames = sorted(running_ifname_set, key=self.convert)
            for ifname in ifnames:
                # first line should have interface name
                mykeys = sorted(details[ifname].keys())
                counter = 1
                for k in mykeys:
                    myvals = details[ifname][k]
                    if isinstance(myvals, list):
                        table.append([ifname, k, myvals[0]])
                        for mac in myvals[1:]:
                            table.append(['', '', mac])
                    else:
                        if counter == 1:
                            table.append([ifname, k, myvals])
                        else:
                            table.append(['', k, myvals])
                    counter += 1
                table.append(['','',''])
            print tabulate(tru(table), headers, tablefmt="simple")

    def do_show_macs(self):
        self.three_column_results(command='list_sta', attribute='MAC Addresses')
        return True

    def do_show_port_details(self):
        self.three_column_results(command='all_sta', mylist=self.port_detail_list)
        return True

    def do_show_radius_details(self):
        self.three_column_results(command='mib', mylist=self.radius_detail_list)
        return True

    def do_show_status(self):
        enabled = 'failed'
        active = 'failed'
        try:
            enabled = check_output(['/bin/systemctl', 'is-enabled', 'hostapd'])
        except:
            # when disabled, the command returns 1 so we have to catch it
            enabled = 'disabled'
        try:
            active = check_output(['/bin/systemctl', 'is-active', 'hostapd'])
        except:
            # in a failed state, systemctl returns non-zero, which is ok
            active = 'inactive'

        if self.args.get('json'):
            output = {'enabled_status': enabled.strip(),
                      'active_status': active.strip()}
            print(json.dumps(output, sort_keys=True, indent=4))
        else:
            print('IEEE802.1X Enabled Status: %s'%enabled.strip())
            print('IEEE802.1X Active Status: %s'%active.strip())
        return True

    def invalid_interfaces(self):
        for iface in self.get_ifnames(self.args.get('<interface>')):
            if re.match(r"""^swp\d+(?:s\d+)?$""", iface) is None:
                # has to be a physical port
                return "ERROR: interface %s is not a physical switchport" % iface
            if os.path.exists('/sys/class/net/%s/bonding_slave' % iface):
                # cannot be a bond slave
                return "ERROR: interface %s cannot be a bond member" % iface
            if (self.args.get('parking-vlan') or self.args.get('mab')) \
               and self.args.get("interface") and self.args.get('add') \
               and not os.path.exists('/sys/class/net/%s/brport/bridge' % iface):
                # Adding MAB and Parking VLAN require bridge ports
                return "ERROR: for MAB and Parking VLAN, %s must be a bridgeport" % iface
        return None

    def handle_interfaces(self, running_array = {}):
        # handle net add/del interface <foo> (dot1x|mab) separately
        dot1xattr = self.args.get('dot1x')
        mabattr = self.args.get('mab')
        pvlanattr = self.args.get('parking-vlan')
        add_del = self.args.get('add') or self.args.get('del')
        running_array.setdefault('interfaces', [])
        running_array.setdefault('mab_interfaces', [])
        running_array.setdefault('parking_vlan_interfaces', [])
        for iface in self.get_ifnames(self.args.get('<interface>')):
            if add_del == 'add':
                if dot1xattr:
                    # this is always there, otherwise we would not be here
                    if iface not in running_array['interfaces']:
                        running_array['interfaces'].append(iface)
                    else:
                        # already included, do nothing
                        pass
                if mabattr:
                    if iface not in running_array['mab_interfaces']:
                        running_array['mab_interfaces'].append(iface)
                if pvlanattr:
                    if iface not in running_array['parking_vlan_interfaces']:
                        running_array['parking_vlan_interfaces'].append(iface)
            if add_del == 'del':
                if (dot1xattr and not mabattr and not pvlanattr):
                    if iface in running_array['interfaces']:
                        running_array['interfaces'].remove(iface)
                    if iface in running_array['mab_interfaces']:
                        running_array['mab_interfaces'].remove(iface)
                    if iface in running_array['parking_vlan_interfaces']:
                        running_array['parking_vlan_interfaces'].remove(iface)
                if mabattr:
                    if iface in running_array['mab_interfaces']:
                        running_array['mab_interfaces'].remove(iface)
                if pvlanattr:
                    if iface in running_array['parking_vlan_interfaces']:
                        running_array['parking_vlan_interfaces'].remove(iface)
        running_array['interfaces'] = sorted(running_array['interfaces'],
                                             key=self.convert)
        running_array['mab_interfaces'] = sorted(running_array['mab_interfaces'],
                                                 key=self.convert)
        running_array['parking_vlan_interfaces'] = sorted(running_array['parking_vlan_interfaces'],
                                                          key=self.convert)
        return running_array

    def do_config(self):
        '''
        Save configs in a list to keep track of commands entered.
        Later on (show_pending or commit_pending) we will modify the
        current config by replaying this config file.
        '''
        if (os.path.exists(self.scratch_config_file)):
            running_array = self.get_config(self.scratch_config_file)
        else:
            # we always make a copy of the config if it does not exist
            shutil.copyfile(self.config_file, self.scratch_config_file)
            os.chmod(self.scratch_config_file, 0o600)
            running_array = self.get_config(self.scratch_config_file)

        add_del = self.args.get('add') or self.args.get('del')
        if self.args.get('interface'):
            rc = self.invalid_interfaces()
            if not rc:
                # handle the interface command
                running_array = self.handle_interfaces(running_array)
            else:
                print rc
                return False
        else:
            # handle all the other dot1x commands
            for k in self.config_array.keys():
                # we are looping over actual hostapd.conf config keywords
                # the actual arg value if they did not use default keyword
                argval = self.args.get(self.config_array[k]['arg'])
                # the word default was used in the net command
                defaultcli = self.args.get('default')
                # the keyword has a default
                defaultval = self.config_array[k]['default']
                # a particular nclu keyword was used like server-ip
                cliname = self.args.get(self.config_array[k]['cli_name'])
                if (defaultcli or str(argval)) and cliname:
                    # we have a value we need to deal with
                    if add_del == 'del':
                        if cliname == 'server-ip' and \
                           argval in running_array[k]:
                            # this is special in that both auth_server_addr and
                            # acct_server_addr should be added/deleted to the list
                            running_array[k].remove(argval)
                        else:
                            running_array[k] = None or defaultval
                    elif add_del == 'add':
                        if defaultcli:
                            running_array[k] = defaultval
                        elif cliname == 'server-ip':
                            # we can have up to 3 radius servers
                            # add to both auth and acct server_addr
                            if len(running_array[k]) < 3:
                                running_array[k].append(argval)
                            else:
                                print("ERROR: No more then three RADIUS Servers allowed.")
                                return False

                        else:
                            running_array[k] = argval
        oldscratch = self.get_config(self.scratch_config_file)
        changed = False
        for i in running_array.keys():
            if running_array.get(i, None) != oldscratch.get(i, None):
                changed = True
        if changed:
            self.save_scratch_config(running_array)
        else:
            print("\'%s\' is already configured or does not change the config" % \
                  ' '.join(self.argv_expanded))
            return False
        return True

    def get_ifnames(self, ifstr):
        # we are either adding or deleting an interface (or range)
        if not ifstr:
            return
        ifnames = []
        if ifname_is_glob(str(ifstr)):
            for x in ifname_expand_glob(str(ifstr)):
                ifnames.append(x)
        elif ifname_is_valid(str(ifstr)) and not ifname_is_globish(str(ifstr)):
            ifnames.append(ifstr)
        else:
            print("ERROR: a valid interface name or glob is needed")
            return False
        return ifnames

    def del_all(self, cli):
        # our job here is to wipe the config clean in the scratch area.
        # We are not stopping the hostapd daemon.  We need a better
        # NCLU mechanism to make the stopping of all daemons clear before
        # we start stopping daemons.
        # netd calls us directly and expects no return
        self.save_scratch_config({})
        return True

    def abort_pending(self):
        # need to delete the /var/run/nclu/dot1x configs
        for filename in glob.glob('%s*' % self.dot1x_scratchpad):
            os.remove(filename)
        return

    def commit_pending(self, verbose):
        '''
        This first checks for various missing or incorrect information:
        radius server-ip is required, auth server-port can default as can
        the acct server-port.   After verification, the config file is copied to
        /etc/hostapd.conf and the hostapd daemon is restarted.
        '''
        commit_ok = True
        modified = []

        # we are doing our own copy and systemctl calls
        if not os.path.exists(self.scratch_config_file):
            # there is no scratch config, bail
            for filename in glob.glob('%s*' % self.dot1x_scratchpad):
                os.remove(filename)
            return (commit_ok, modified)

        elif os.path.exists(self.scratch_config_file) and \
                os.path.exists(self.config_file) and \
                not print_diff(self.config_file,
                               self.scratch_config_file,
                               self.color_diffs):
            # Nothing has changed, but we could get
            # here if net del all is called and
            # creates an identical config in scratch.
            # Since every wrapper commit will get called, we have
            # to recheck for changed configs before proceeding.
            # Otherwise, we start or restart hostapd for no reason.
            for filename in glob.glob('%s*' % self.dot1x_scratchpad):
                os.remove(filename)
            return (commit_ok, modified)

        elif os.path.exists(self.scratch_config_file) and \
                os.path.exists(self.config_file) and \
                print_diff(self.config_file,
                           self.scratch_config_file,
                           self.color_diffs):
            # we have a config and it is different from the one running
            # so we must keep going
            pass
        else:
            # all other cases should bail out
            for filename in glob.glob('%s*' % self.dot1x_scratchpad):
                os.remove(filename)
            return (commit_ok, modified)

        # if any of the radius attributes changed, we currently need to restart
        # otherwise, a simple reload will handle interface adds and deletes.
        newconfig = self.get_config(self.scratch_config_file)
        oldconfig = self.get_config(self.config_file)
        # the only way we restart rather then reload is the following changed
        # attributes
        command = 'reload'
        for attr in ['auth_server_addr', 'acct_server_addr',
                     'auth_server_port', 'auth_server_shared_secret',
                     'acct_server_port', 'acct_server_shared_secret',
                     'eap_reauth_period', 'mab_activation_delay', 'parking_vlan_id']:
            if newconfig.get(attr, '') != oldconfig.get(attr, ''):
                command = 'restart'

        # there is a case where we could be running with no interfaces, but now
        # we have some, must do  restart.
        if len(oldconfig.get('interfaces', '')) == 0 and \
           len(newconfig.get('interfaces', '')) > 0:
            command = 'restart'
        # there is a case where we could be running with some interfaces but
        # now have none, just stop.
        if len(oldconfig.get('interfaces', '')) > 0 and \
           len(newconfig.get('interfaces', '')) == 0:
            command = 'stop'

        # copy the config to the right place.
        shutil.copyfile(self.config_file, self.temp_config_file)
        shutil.copyfile(self.scratch_config_file, self.config_file)
        os.chmod(self.config_file, 0o600)
        modified.append(self.config_file)

        try:
            try:
                enabled = check_output(['/bin/systemctl', 'is-enabled', 'hostapd'])
            except:
                enabled = check_output(['/bin/systemctl', 'enable', 'hostapd'])
            try:
                enabled = check_output(['/bin/systemctl', 'is-active', 'hostapd'])
                # if it is running, reload or restart or stop
                enabled = check_output(['/bin/systemctl', command, 'hostapd'])
            except:
                # it is not running, better just restart it
                if command != 'stop':
                    enabled = check_output(['/bin/systemctl', 'restart', 'hostapd'])
        except CalledProcessError as e:
            print('Error: attept to %s hostapd failed: %s'%(command, e))
            commit_ok = False

        # remove rest of files since we just finished the commit
        if commit_ok:
            for filename in glob.glob('%s*' % self.dot1x_scratchpad):
                os.remove(filename)
        elif os.path.exists(self.temp_config_file):
                shutil.copyfile(self.temp_config_file, self.config_file)
                modified = []

        #restart the hostapd daemon with the new interface configs
        return [commit_ok, modified]

    def get_running_ifnames(self, config_file=None, ifaces=set([])):
        # check for a running default file
        if os.path.exists(config_file):
            fd = open(config_file, 'r')
            default_lines = fd.read().strip().split('\n')
            default_lines = [l for l in default_lines
                             if l.startswith('interfaces=')]
            temp_str = default_lines[0].split('=')[1]
            if temp_str:
                temp_set = set(temp_str.strip().split(','))
                ifaces = ifaces | temp_set
            fd.close()
        return ifaces

    def get_config(self, config_file=None):
        # check for a running config in an interface config.
        config_array = {}
        config_array['auth_server_addr'] = []
        config_array['acct_server_addr'] = []
        if os.path.exists(config_file):
            config_fd = open(config_file, 'r')
            config_list = config_fd.read().strip().split('\n')
            config_list = [l for l in config_list \
                           if not l.startswith('#') and '=' in l]
            for l in config_list:
                if l.find('=') > -1:
                    i = l[:l.find('=')]
                    j = l[l.find('=')+1:]
                    if i == 'interfaces' or i == 'mab_interfaces' or \
                        i == 'parking_vlan_interfaces':
                        # interfaces are special, they are saved as a list
                        if j == '':
                            config_array[i.strip()]=[]
                        else:
                            config_array[i.strip()]=j.strip().split(',')
                    elif i == 'auth_server_addr' or i == 'acct_server_addr':
                        # we can have up to 3 radius servers
                        if j:
                            config_array[i.strip()].append(j.strip())
                    else:
                        config_array[i.strip()]=j.strip()
                else:
                    print("Error: invalid config file %s" % config_file)
            config_fd.close()
        return config_array

    def get_pending_config(self):
        pending_config = {}
        try:
            with open(self.pending_config_file) as infile:
                pending_config = json.load(infile)
        except:
            pass
        return pending_config

    def write_server_attrs(self, config_fd=None, server=None, config_array=None):
        if 'auth' in server:
            portname = 'auth_server_port'
            secretname = 'auth_server_shared_secret'
        elif 'acct' in server:
            portname = 'acct_server_port'
            secretname = 'acct_server_shared_secret'

        port = config_array.get(portname) or self.config_defaults[portname]
        secret = config_array.get(secretname) or self.config_defaults[secretname]
        if port:
            config_fd.write('%s=%s\n' % (portname, port))
        else:
            config_fd.write('%s=\n' % (portname))
        if secret:
            config_fd.write('%s=%s\n' % (secretname, secret))
        else:
            config_fd.write('%s=\n' % (secretname))
        return

    def save_scratch_config(self, config_array):
        # just overwrite it
        # enforce some parameters and reset them in the hostapd.conf file
        config_fd = open(self.scratch_config_file, 'w')
        os.chmod(self.scratch_config_file, 0o600)
        config_fd.write('#### Auto-generated config file: do not edit. ####\n')
        # when we save the config, we save the radius ports and secrets with the server.
        attrs = [l for l in self.config_attrs if l not in ['auth_server_port',
                                                           'auth_server_shared_secret',
                                                           'acct_server_port',
                                                           'acct_server_shared_secret']]
        for l in attrs:
            # if the config is specified, use it
            val = config_array.get(l) or self.config_defaults[l]
            if val and ('interfaces' in l):
                # interfaces are special in that a blank interfaces line
                # will generate a '' at the beginning when split.  Remove this
                if '' in val:
                    val.remove('')
                val = ','.join(val)
            # We can have up to 3 radius servers, handle them here
            if val and ('server_addr' in l):
                # the value should be a list of ip addresses
                # need to write the ports and shared secrets
                # the same time because order matters in hostapd.conf
                for ip in val:
                    # write the auth_server_addr or acct_server_addr
                    config_fd.write('%s=%s\n' % (l, ip))
                    self.write_server_attrs(config_fd=config_fd, server=l,
                                            config_array=config_array)
            elif 'server_addr' in l:
                # we have no servers configured, so we need to write out blanks
                config_fd.write('%s=\n' % (l))
                self.write_server_attrs(config_fd=config_fd, server=l,
                                        config_array=config_array)

            elif val:
                config_fd.write('%s=%s\n' % (l, val))
            else:
                config_fd.write('%s=\n' % (l))
        config_fd.close()
        return

    def show_pending(self, print_errors=True):
        # if any of the radius attributes changed, we currently need to restart
        # otherwise, a simple reload will handle interface adds and deletes.

        diff_pending = print_diff(self.config_file,
                                  self.scratch_config_file,
                                  self.color_diffs)
        if not diff_pending:
            return PENDING_COMMIT_NOCHANGE

        newconfig = self.get_config(self.scratch_config_file)
        oldconfig = self.get_config(self.config_file)
        server_addr = newconfig.get('auth_server_addr')
        secret = newconfig.get('auth_server_shared_secret')
        ifnames = newconfig.get('interfaces')
        pvlan_ifnames = newconfig.get('parking_vlan_interfaces')
        pvlan_id = newconfig.get('parking_vlan_id')
        if ifnames and not server_addr:
            print("\nERROR: Missing dot1x RADIUS server-ip address")
            return PENDING_COMMIT_ERROR
        if server_addr and not secret:
            print("\nERROR: Missing dot1x RADIUS shared-secret")
            return PENDING_COMMIT_ERROR
        if pvlan_ifnames and not pvlan_id:
            print("\nERROR: Parking VLAN Interfaces enabled without Parking VLAN ID")
            return PENDING_COMMIT_ERROR

        # the only way we restart rather then reload is the following changed
        # attributes
        command = 'reload'
        for attr in ['auth_server_addr', 'acct_server_addr',
                     'auth_server_port', 'auth_server_shared_secret',
                     'acct_server_port', 'acct_server_shared_secret',
                     'eap_reauth_period', 'mab_activation_delay', 'parking_vlan_id']:
            if newconfig.get(attr, '') != oldconfig.get(attr, ''):
                command = 'restart'

        # there is a case where we could be running with no interfaces, but now
        # we have some, must do  restart.
        if len(oldconfig.get('interfaces', '')) == 0 and \
           len(newconfig.get('interfaces', '')) > 0:
            command = 'restart'
        # there is a case where we could be running with some interfaces but
        # now have none, just stop.
        if len(oldconfig.get('interfaces', '')) > 0 and \
           len(newconfig.get('interfaces', '')) == 0:
            command = 'stop'
            print("\nWarning: with no dot1x interfaces, hostapd daemon will stop")

        if command == 'restart':
            print("\nWarning: committing changes will require hostapd daemon restart")
            print("         which will deauthorize all stations.")

        return PENDING_COMMIT_CHANGE

    def show_config_files(self, user_may_edit):
        if os.path.exists(self.config_file):
            print "%s" % self.config_file
            print "======================="

            # If the user has edit privs then do not bother hiding passwords
            if user_may_edit:
                with open(self.config_file, 'r') as fh:
                    print ''.join(fh.readlines())
            else:
                output = []
                with open(self.config_file, 'r') as fh:
                    for line in fh.readlines():
                        line = re.sub(r'''_shared_secret=(\S+)''',
                                      '_shared_secret=PASSWORD REMOVED', line)
                        output.append(line)
                print ''.join(output)
            print ''

        return

    def get_managed_files(self):
        files = []
        files.append(self.config_file)
        return files

    def show_config_commands(self, user_may_edit):
        supported = []
        unsupported = []
        running_array = self.get_config(self.config_file)
        # now show the supported commands
        for k in self.cli_array.keys():
            valkey = self.cli_array[k]['config_list'][0]
            if running_array.get(valkey, '') and \
               k != 'dot1x' and k != 'mab' and k != 'parking-vlan':
                val = running_array.get(valkey, '')
                if not user_may_edit and k == 'shared-secret':
                    supported.append('net add dot1x radius %s PASSWORD_REMOVED' % k)
                else:
                    if k == 'mab-activation-delay' or k == 'eap-reauth-period' or \
                        k == 'parking-vlan-id':
                        supported.append('net add dot1x %s %s' % (k, val))
                    elif k == 'server-ip':
                        # radius servers is a list
                        for ip in val:
                            supported.append('net add dot1x radius %s %s' % (k, ip))
                    else:
                        supported.append('net add dot1x radius %s %s' % (k, val))
        # interfaces is always a list
        for iface in running_array.get('interfaces', []):
            if iface in running_array.get('mab_interfaces', []) and \
               iface in running_array.get('parking_vlan_interfaces', []):
                supported.append('net add interface %s dot1x mab parking-vlan' % iface)
            elif iface in running_array.get('mab_interfaces', []):
                supported.append('net add interface %s dot1x mab' % iface)
            elif iface in running_array.get('parking_vlan_interfaces', []):
                supported.append('net add interface %s dot1x parking-vlan' % iface)
            elif iface in running_array.get('interfaces', []):
                supported.append('net add interface %s dot1x' % iface)
        return (supported, unsupported)

    def show_config_summary(self, user_may_edit, summary):
        running_array = self.get_config(self.config_file)
        for i in running_array.get('interfaces', []):
            k = 'INTERFACE:%s' % i
            summary.setdefault(k, OrderedDict())
            summary[k].setdefault('COMMAND_LIST', [])
            if i in running_array.get('mab_interfaces', []):
                summary[k]['COMMAND_LIST'].append('mab-enabled')
            if i in running_array.get('parking_vlan_interfaces', []):
                summary[k]['COMMAND_LIST'].append('parking-vlan-enabled')
            if i in running_array.get('interfaces', []):
                summary[k]['COMMAND_LIST'].append('dot1x-enabled')
        summary['dot1x'] = OrderedDict()
        if running_array.get('mab_activation_delay', ''):
            summary['dot1x']['mab-activation-delay %s' % \
                             running_array.get('mab_activation_delay', '')] = []
        if running_array.get('eap_reauth_period', ''):
            summary['dot1x']['eap-reauth-period %s' % \
                             running_array.get('eap_reauth_period', '')] = []
        if running_array.get('parking_vlan_id', ''):
            summary['dot1x']['parking-vlan-id %s' % \
                             running_array.get('parking_vlan_id', '')] = []
        summary['dot1x']['radius'] = []
        for k in self.cli_array.keys():
            valkey = self.cli_array[k]['config_list'][0]
            # dot1x interfaces will be added to each interface
            if running_array[valkey] and \
               k != 'dot1x' and k != 'mab' and k != 'mab-activation-delay' and \
               k != 'eap-reauth-period' and k != 'parking-vlan-id':
                val = running_array[valkey]
                if not user_may_edit and k == 'shared-secret':
                    summary['dot1x']['radius'].append('%s PASSWORD_REMOVED' % k)
                elif k == 'server-ip':
                    # radius server-ip is a list
                    for ip in val:
                        summary['dot1x']['radius'].append('%s %s' % (k, ip))
                else:
                    summary['dot1x']['radius'].append('%s %s' % (k, val))
        # interfaces key is special in that dot1x-enabled is shown with interfaces
        # even though they are not in the /etc/network/interfaces file
        return

    def get_managed_files(self):
        files = []
        return files

    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user.
        For an add, this method needs to create an interface config file hostapd and stage it
        in /var/run/nclu/dot1x.
        """
        #print('dot1x: eval_command_line_args')
        self.args = cli.args
        self.argv_expanded = cli.argv_expanded
        rc = True
        if self.args.get('show') and self.args.get('macs'):
            rc = self.do_show_macs()
        elif self.args.get('show') and self.args.get('port-details'):
            rc = self.do_show_port_details()
        elif self.args.get('show') and self.args.get('radius-details'):
            rc = self.do_show_radius_details()
        elif self.args.get('show'):
            rc = self.do_show_status()
        elif self.args.get('del') and self.args.get('all'):
            rc = self.del_all(cli)
        else:
            rc = self.do_config()
        return rc

# This is for testing only
if __name__ == '__main__':
    c = ConfigWrapper('foo', True)
    print c
