# Copyright (C) 2016-2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    # Interface configuration
    iface del bond <interface>
    iface (add|del) bridge
    iface add loopback lo
    iface (add|del) vrf <interface>
    iface (add|del) vrf mgmt

    iface (add|del) interface <interface>
    iface del vxlan <interface>
    iface (add|del) bond <interface> bridge trunk [vlans <number-range-list>]
    iface (add|del) interface <interface> bridge trunk [vlans <number-range-list>]
    iface (add|del) vlan <number-range-list>
    iface add clag peer sys-mac <mac-clag> interface <interface> (primary|secondary) [backup-ip <ipv4>]
    iface add clag peer sys-mac <mac-clag> interface <interface> (primary|secondary) [backup-ip <ipv4> vrf <text>]
    iface del clag peer
    iface add clag port bond <interface> interface <interface> clag-id <0-65535>
    iface del clag port bond <interface>

    iface show clag [our-macs|our-multicast-entries|our-multicast-router-ports|peer-macs|peer-multicast-entries|peer-multicast-router-ports|params|backup-ip|id] [verbose] [json]
    iface show clag macs [<mac>] [json]
    iface show clag peer-lacp-rate
    iface show clag verify-vlans [verbose]
    iface show clag status [verbose] [json]
    iface show bridge macs [<mac>|vlan <number>|<ip>|permanent|dynamic] [json]
    iface show bridge (link|mdb)
    iface show bridge spanning-tree [<interface>] [json]
    iface show bridge vlan [<number>] [json]

Options:

    0                            : 0, False, etc
    1                            : 1, True, etc
    4096                         : 4096
    8192                         : 8192
    12288                        : 12288
    16384                        : 16384
    20480                        : 20480
    24576                        : 24576
    28672                        : 28672
    36864                        : 36864
    40960                        : 40960
    45056                        : 45056
    49152                        : 49152
    53248                        : 53248
    57344                        : 57344
    61440                        : 61440
    100                          : 100M
    1000                         : 1G
    10000                        : 10G
    25000                        : 25G
    50000                        : 50G
    100000                       : 100G
    40000                        : 40G
    <interface-as-arg>           : An interface name "swp1" or glob "swp1-4 6 10-12"
    <interface-list>             : Example: "swp1 swp2 swp3"
    <interface-range-list>       : Example: "swp1=100 swp2=100"
    <interface-yes-no-0-1-list>  : Example: "swp1=yes swp2=no swp3=1,swp4=0"
    <interface-yes-no-auto-list> : Example: "swp1=yes swp2=no swp3=auto"
    <interface-yes-no-list>      : Example: "swp1=yes swp2=no"
    <interface-on-off-list>      : Example: "swp1=on swp2=off"
    <interface>                  : An interface name "swp1" or glob "swp1-4 6 10-12"
    <ip/prefixlen>               : An IPv4 or IPv6 address and prefix length
    <ip>                         : An IPv4 or IPv6 Address
    <ipv4-vrf-text>              : Example: "<ipv4> vrf <text>"
    <ipv4/prefixlen>             : An IPv4 address and prefix length
    <ipv4>                       : An IPv4 address
    <ipv6/prefixlen>             : An IPv6 address and prefix length
    <ipv6>                       : An IPv6 address
    <mac-ip/prefixlen-list>      : Example: "00:11:22:33:44:01 10.0.0.1/24 10.0.1.2/24"
    <mac>                        : A MAC address
    <mac-clag>                   : A MAC address from the 44:38:39:FF:00:00 to 44:38:39:FF:FF:FF range
    <number-interface-list>      : Example: "4 swp1 swp2"
    <number-ipv4-list>           : Example: "100=10.0.100.1 101=10.101.1"
    <number-range-list>          : Example: "1-5,8,10-12,15"
    <number-comma-range-list>    : Example: "1-5,8,10-12,15"
    <number>                     : Any integer
    <text>                       : Any text
    auto                         : Automatically
    autoneg                      : Auto negotiation
    backup-ip                    : Displays backup link info
    baser                        : Forward error correction mode
    basic-clag                   : Introduction to CLAG/MLAG configuration
    bond                         : Bond, port-channel, etc
    bond-slaves                  : Add physical ports to a bond
    bridge                       : A layer2 bridge
    trunk                        : A layer2 trunk, 'bridge trunk' is an alias for 'bridge vids'
    clag                         : Multi-Chassis Link Aggregation
    configuration                : Settings, configuration state, etc
    detail                       : Show detailed output
    details                      : Show detailed output
    dynamic                      : MACs learned dynamically
    duplex                       : Link duplex
    fec                          : Forward Error Correction
    full                         : Full duplex
    half                         : Half duplex
    hostname                     : System hostname
    id                           : Identifier
    ip                           : Internet Protocol version 4
    ipv6                         : Internet Protocol version 6
    l2-with-server-vlan-trunks   : Traditional high scale L2/virtualization deployment
    l3-uplinks-virtual-address   : Modern network virtualization/containter deployment
    link                         : An interface
    link-settings                : Physical link parameters
    lo                           : The loopback interface name
    loopback                     : A loopback interface
    macs                         : Media Access Control addresses
    mdb                          : Multicast group database entry
    mgmt                         : Management VRF
    multicast-querier            : Ipv4 multicast querier
    no                           : Negative, disable, etc
    off                          : Disable
    on                           : Enable
    our-macs                     : Displays the MACs learned on this switch
    our-multicast-entries        : Displays the multicast entries learned on this switch
    our-multicast-router-ports   : Displays the multicast router ports learned on this switch
    params                       : Display the parameters in use by clagd
    peer-lacp-rate               : Displays the peer's polling rate
    peer-macs                    : Displays the MACs learned on the peer switch
    peer-multicast-entries       : Displays the multicast entries learned on the peer switch
    peer-multicast-router-ports  : Displays the multicast router ports learned on the peer switch
    permanent                    : Intended to exist or function for a long, indefinite period
    rs                           : Reed-Solomon
    spanning-tree                : STP - Spanning Tree Protocol
    speed                        : Link speed
    status                       : Display the status of the clagd daemon
    verbose                      : Show detailed output
    verify-vlans                 : Verifies VLAN configuration with the peer
    version                      : Version number
    vlan                         : Virtual Local Area Network
    vlan-interfaces              : IP interfaces for VLANs
    vlans                        : Virtual Local Area Networks
    vni                          : VXLAN Network Identifier
    vrf <text>                   : Virtual Routing and Forwarding
    vxlan                        : Virtual Extensible LAN
    yes                          : Positive, enable, etc

    allow-untagged               : A bridge port interface may allow untagged packets
    args                         : Command line arguments
    bpduguard                    : Bridge Protocol Data Unit guard
    down                         : Disable, turn off, etc
    enable                       : To make able
    host                         : An end station
    igmp-querier-src             : Bridge igmp querier source
    lacp-bypass-allow            : Allow LACP bypass
    lacp-rate                    : Bond LACP rate
    mcquerier                    : Multicast querier
    mcsnoop                      : Multicast snooping
    miimon                       : Bond miimon
    min-links                    : Bond minimum number of links
    peer                         : A BGP, OSPF, PIM, etc neighbor
    peer-ip                      : The IP address of our peer
    portadminedge                : Edge state of the port
    portautoedge                 : Auto transition to/from edge state of the port
    portbpdufilter               : BPDU filter on a port
    portmcfl                     : Multicast fast leave
    portmcrouter                 : Multicast ports
    portnetwork                  : Bridge assurance capability for a port
    ports                        : Interfaces
    primary                      : The first or highest in rank or importance
    pvid                         : The native VLAN
    secondary                    : Next after the first in order, place, time, etc
    slaves                       : Bond slave interfaces
    stp                          : Spanning Tree Protocol
    sys-mac                      : The system ID of the CLAG pair
    treeprio                     : STP priority
    vids                         : VLAN IDs
    vlan-aware                   : A VLAN aware bridge
    vxlan-anycast-ip             : Anycast local IP address for dual connected VxLANs
    local-tunnelip               : vxlan local tunnel ip
    purge-remotes                : vxlan purge existing remote entries
    remoteip                     : remote IP address
    svcnodeip                    : vxlan service node IP address
    updelay                      : Bond updelay
    downdelay                    : Bond downdelay
    secondary                    : The second or next highest in rank or importance
    learning                     : The bridge port learning flag
    mld-version                  : Multicast mld version
    igmp-version                 : IGMP version
    mode                         : Mode
    802.3ad                      : LACP
    balance-xor                  : Static bond
    ad-actor-sys-prio            : 802.3ad system priority
    ad-actor-system              : 802.3ad system mac address
    num-grat-arp                 : Number of gratuitous ARP packets to send on failover
    num-unsol-na                 : Number of unsolicited IPv6 NA packets to send on failover
    use-carrier                  : Bond use carrier
    xmit-hash-policy             : Bond hash policy
    multicast-flood              : Bridge port multicast flood flag
    unicast-flood                : Bridge port unicast flood flag
    ageing                       : Ageing
    vlan-protocol                : Bridge VLAN protocol
    slow                         : Bond LACP rate slow
    vlan-stats                   : VLAN statistics
    arp-nd-suppress              : ARP ND suppression
    802.1q                       : 802.1q
    802.1ad                      : 802.1ad
    forward                      : Forwarding
    nd-suppress                  : Neighbor Discovery suppress
    igmp-stats                   : IGMP statistics
    mcstats                      : Multicast statistics
    mlag                         : Multi-Chassis Link Aggregation

"""

# At one point we experimented with a bond-master keyword but bond that
# bonds are pretty configurable already
#
#    iface (add|del) interface <interface> bond-master <interface-bond>
#    bond-master                  : Add a physical interface to a bond

from collections import OrderedDict
from copy import deepcopy
from ipaddr import IPv6Network, IPv4Network, AddressValueError, IPNetwork
from network_docopt import sort_ifaces, sort_for_humans, NetworkDocopt
from nclu import (print_diff,
    decode_dict,
    glob_msg,
    ifname_is_valid,
    ifname_is_globish,
    ifname_is_glob,
    ifname_expand_glob,
    is_mac_address,
    is_number_glob,
    PENDING_COMMIT_NOCHANGE,
    PENDING_COMMIT_CHANGE,
    PENDING_COMMIT_ERROR,
    get_interfaces,
    glob_to_numbers,
    globify_interface_commands,
    numbers_to_glob,
    Capturing,
    seconds_to_hhmmss,
    PENDING_COMMANDS_FILE,
    print_pending_add_del_commands,
    tabulate_remove_unicode as tru)
from pprint import pformat
from shutil import rmtree, copyfile
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate
import json
import logging
import logging.handlers
import os
import re
import shlex
import shutil
import subprocess
import sys
sys.path.append('/usr/share/ifupdown2/addons/')

log = logging.getLogger('netd')
addons_cmds = []
keyword_default = {}

all_interface_keywords = (
    'post-down',
    'post-up',
    'pre-down',
    'pre-up')

bond_interface_keywords = (
    'address',
    'address-virtual',
    'alias',
    'bond-ad-actor-sys-prio',
    'bond-ad-actor-system',
    'bond-downdelay',
    'bond-lacp-bypass-allow',
    'bond-lacp-rate',
    'bond-miimon',
    'bond-min-links',
    'bond-mode',
    'bond-num-grat-arp',
    'bond-num-unsol-na',
    'bond-slaves',
    'bond-updelay',
    'bond-use-carrier',
    'bond-xmit-hash-policy',
    'bridge-access',
    'bridge-allow-untagged',
    'bridge-multicast-flood',
    'bridge-nd-suppress',
    'bridge-learning',
    'bridge-pathcosts',
    'bridge-portmcfl',
    'bridge-portmcrouter',
    'bridge-pvid',
    'bridge-unicast-flood',
    'bridge-vids',
    'clag-id',
    'gateway',
    'ip-forward',
    'ip6-forward',
    'link-down',
    'mpls-enable',
    'mstpctl-bpduguard',
    'mstpctl-portadminedge',
    'mstpctl-portautoedge',
    'mstpctl-portbpdufilter',
    'mstpctl-portnetwork',
    'mstpctl-portp2p',
    'mstpctl-portpathcost',
    'mstpctl-portrestrrole',
    'mstpctl-portrestrtcn',
    'mstpctl-treeportcost',
    'mstpctl-treeportprio',
    'mtu',
    'netmask',
    'preferred-lifetime',
    'scope',
    'vlan-protocol',
    'vrf')

bridge_interface_keywords = (
    'alias',
    'bridge-ageing',
    'bridge-bridgeprio',
    'bridge-fd',
    'bridge-gcint',
    'bridge-hashel',
    'bridge-hashmax',
    'bridge-hello',
    'bridge-igmp-querier-src',
    'bridge-igmp-stats',
    'bridge-igmp-version',
    'bridge-maxage',
    'bridge-maxwait',
    'bridge-mclmc',
    'bridge-mclmi',
    'bridge-mcmi',
    'bridge-mcqi',
    'bridge-mcqifaddr',
    'bridge-mcqpi',
    'bridge-mcqri',
    'bridge-mcquerier',
    'bridge-mcqv4src',
    'bridge-mcrouter',
    'bridge-mcsnoop',
    'bridge-mcsqc',
    'bridge-mcsqi',
    'bridge-mcstats',
    'bridge-mld-version',
    'bridge-portprios',
    'bridge-ports',
    'bridge-pvid',
    'bridge-stp',
    'bridge-vids',
    'bridge-vlan-aware',
    'bridge-vlan-protocol',
    'bridge-vlan-stats',
    'bridge-waitport',
    'mstpctl-ageing',
    'mstpctl-fdelay',
    'mstpctl-forcevers',
    'mstpctl-hello',
    'mstpctl-maxage',
    'mstpctl-maxhops',
    'mstpctl-treeprio',
    'mstpctl-txholdcount')

loopback_interface_keywords = (
    'address',
    'alias',
    'clagd-vxlan-anycast-ip',
    'ip-forward',
    'ip6-forward',
    'netmask',
    'preferred-lifetime',
    'scope',
    'vxrd-src-ip',
    'vxrd-svcnode-ip',
    'vrf')

physical_interface_keywords = (
    'address',
    'address-virtual',
    'alias',
    'bridge-access',
    'bridge-allow-untagged',
    'bridge-learning',
    'bridge-multicast-flood',
    'bridge-nd-suppress',
    'bridge-pathcosts',
    'bridge-portmcfl',
    'bridge-portmcrouter',
    'bridge-pvid',
    'bridge-vids',
    'bridge-unicast-flood',
    'clagd-args',
    'clagd-backup-ip',
    'clagd-enable',
    'clagd-peer-ip',
    'clagd-priority',
    'clagd-sys-mac',
    'gateway',
    'ip-forward',
    'ip6-forward',
    'link-autoneg',
    'link-down',
    'link-duplex',
    'link-fec',
    'link-speed',
    'mpls-enable',
    'mstpctl-bpduguard',
    'mstpctl-portadminedge',
    'mstpctl-portautoedge',
    'mstpctl-portbpdufilter',
    'mstpctl-portnetwork',
    'mstpctl-portp2p',
    'mstpctl-portpathcost',
    'mstpctl-portrestrrole',
    'mstpctl-portrestrtcn',
    'mstpctl-treeportcost',
    'mstpctl-treeportprio',
    'mtu',
    'netmask',
    'preferred-lifetime',
    'scope',
    'vlan-protocol',
    'vrf')

vlan_interface_keywords = (
    'address',
    'address-virtual',
    'alias',
    'gateway',
    'ip-forward',
    'ip6-forward',
    'link-down',
    'mpls-enable',
    'mtu',
    'netmask',
    'preferred-lifetime',
    'scope',
    'vlan-id',
    'vlan-protocol',
    'vlan-raw-device',
    'vrf')

vxlan_interface_keywords = (
    'address',
    'alias',
    'bridge-access',
    'bridge-arp-nd-suppress',
    'bridge-learning',
    'bridge-multicast-flood',
    'bridge-nd-suppress',
    'bridge-unicast-flood',
    'gateway',
    'ip-forward',
    'ip6-forward',
    'link-down',
    'mstpctl-bpduguard',
    'mstpctl-portbpdufilter',
    'mtu',
    'netmask',
    'preferred-lifetime',
    'scope',
    'vxlan-ageing',
    'vxlan-id',
    'vxlan-learning',
    'vxlan-local-tunnelip',
    'vxlan-purge-remotes',
    'vxlan-remoteip',
    'vxlan-svcnodeip')

vrf_interface_keywords = (
    'address',
    'ip-forward',
    'ip6-forward',
    'preferred-lifetime',
    'scope',
    'vrf-table')

all_keywords = []

for keyword in list(set((
    ('bridge-trunk', ) +
    all_interface_keywords +
    bond_interface_keywords +
    bridge_interface_keywords +
    loopback_interface_keywords +
    physical_interface_keywords +
    vlan_interface_keywords +
    vxlan_interface_keywords +
    vrf_interface_keywords))):

    if keyword in ('post-down', 'post-up', 'pre-down', 'pre-up'):
        all_keywords.append(keyword)
    elif '-' in keyword:
        re_prefix = re.search('^(.*?)-(.*)$', keyword)

        if re_prefix:
            pre = re_prefix.group(1)
            post = re_prefix.group(2)

            # We cannot treat bond, clag, etc as keywords because they also
            # exist in "net add bond", "net add clag", etc
            if pre not in ('bond', 'clag', 'vlan', 'vxlan'):
                all_keywords.append(pre)
            all_keywords.append(post)
    else:
        all_keywords.append(keyword)

all_keywords = sorted(list(set(all_keywords)))

IFACE_DIR_BACKUP = '/var/run/nclu/iface-backup/'
ETC_NETWORK_INTERFACES_BACKUP = '/var/run/nclu/iface/etc_network_interfaces.backup'
ETC_NETWORK_INTERFACES = '/etc/network/interfaces'


def ifname_in_etc_network_interfaces(ifname):

    # If this file exists, the user is in the process of adding ifname.
    if os.path.exists('/var/run/nclu/iface/{0}'.format(ifname)):
        return True
    else:
        iface_line = 'iface {0}'.format(ifname)
        iface_inet_line = 'iface {0} inet'.format(ifname)

        with open(ETC_NETWORK_INTERFACES) as f:
            for line in f:
                if line.strip() == iface_line or line.startswith(iface_inet_line):
                    return True
    return False


def standardize_indent(line):
    """
    /etc/network/interfaces tends to be indented with four spaces but ifquery uses tabs.
    Note that this also strips newline characters.
    """

    indented = line.startswith(' ') or line.startswith('\t')
    line = line.strip()

    if line and indented:
        return '    ' + line
    else:
        return line


def write_new_interface(lines_to_write, prev_line, ifname, lines, persist):

    # Always have a blank line between interface stanzas
    if prev_line:
        lines_to_write.append("\n")

    for line in lines:
        lines_to_write.append(line)
    lines_to_write.append("\n")

    if persist:
        os.remove("/var/run/nclu/iface/%s" % ifname)


def get_ifnames(cli, is_delete=False):
    """
    Return interface names the user specified on the command line.  The behavior is different for
    "net add ..." than it is for "net del ...".
    """

    args = cli.args
    ifnames = []
    ifname = None
    ifnames_valid = True

    if len(cli.argv_expanded) >= 4 and cli.match[3].value and cli.match[3].value.text == '<interface>':
        ifname = args.get('<interface>')

        if isinstance(ifname, list):
            ifname = ifname[0]

    # Most (add|del) commands also have <interface> but not all of
    # them...'add vlans 10-20' for example
    if ifname:

        # If <interface> is a glob then populate ifnames with all of the
        # interface names produced by that glob else populate ifnames with <interface>
        if ifname_is_glob(str(ifname)):
            ifnames = ifname_expand_glob(str(ifname))

            # A glob such as "swp12-11" will pass ifupdown2's glob regexes (so
            # ifname_is_glob will return true) but will not expand becauase the
            # range is backwards.  Return false for this scenario.
            if not ifnames:
                print glob_msg.replace('INTERFACE', ifname)
                ifnames_valid = False
                return (ifnames_valid, ifnames)
        else:
            if not ifname_is_valid(ifname) or ifname_is_globish(ifname):
                ifnames_valid = False
                return (ifnames_valid, ifnames)
            ifnames = [ifname, ]

    else:
        # "net add bridge" is an exception.  We do not allow the user
        # to specify the bridge name so populate ifnames with 'bridge'.
        if args.get('bridge') and len(cli.argv_expanded) >= 3 and cli.argv_expanded[2] == 'bridge':
            ifnames.append('bridge')

        # Same for loopback...force the name to 'lo'.
        elif args.get('loopback') and len(cli.argv_expanded) >= 3 and cli.argv_expanded[2] == 'loopback':
            ifnames.append('lo')

        # If the user does "net add vlan 10-20" do NOT create SVIs for all of those vlans
        # If the user does "net add vlan 10-20 vrf BLUE" do create SVIs for all of those vlans
        elif args.get('vlan') and args.get('<number-range-list>'):
            if (not is_delete and len(cli.argv_expanded) > 4) or is_delete:
                vlan_numbers = glob_to_numbers(str(args.get('<number-range-list>')))

                for number in vlan_numbers:
                    ifnames.append("vlan%d" % number)

    return (ifnames_valid, ifnames)


def expand_bridge_ports_glob(cmd):
    """
    If 'cmd' contains 'bridge-ports glob swp1-4' expand that
    out to 'bridge-ports swp1 swp2 swp3 swp4'.
    """

    # compress duplicate whitespaces
    cmd = ' '.join(cmd.strip().split())

    # Make 'regex swp*' a single token so we can split by whitespace
    cmd = cmd.replace('regex ', 'regex-')

    # Make 'glob swp1-9' a single token so we can split by whitespace
    cmd = cmd.replace('glob ', 'glob-')

    final = []

    for token in cmd.split():
        if token.startswith('glob-'):
            glob_args_expanded = ifname_expand_glob(token.replace('glob-', 'glob ').split()[1])
            final.extend(glob_args_expanded)
        else:
            final.append(token)

    return ' '.join(final)


class Command(object):
    """
    Base class for the various CommandXYZ classes that are used to represent
    commands supported by ifupdown2 addon modules
    """

    def __init__(self, keyword, value):
        """
        value is a dictionary of attributes for the 'keyword' command.
        value is extracted from the _modinfo dictionary for the ifupdown2
        addon that owns this command.
        """
        self.keyword = keyword
        self.keyword_original = keyword
        self.default = None
        self.validvals = None
        self.rangemin = None
        self.rangemax = None
        self.deprecated = False
        self.help = value.get('help')
        self.usage_str_add = []
        self.usage_str_del = []
        self.validvals = []
        add_string = None
        del_string = None
        global keyword_default

        # There is some weird ifupdown1 historical reason that the address.py
        # addon cannot list <ipv4> as the validvals for netmask so we hard code
        # it here to work around that
        if self.keyword == 'netmask' and 'validvals' not in value:
            value['validvals'] = ['<ipv4>', ]

        # We only support vlan aware bridges, do not provide a means to disable it
        if self.keyword == 'bridge-vlan-aware':
            value['validvals'] = ['yes', ]

        self.default = value.get('default')

        if self.default is not None:
            keyword_default[self.keyword] = self.default

        if 'deprecated' in value:
            self.deprecated = value['deprecated']

        # bridge-pathcosts and mstpctl-portpathcost support configuring the keyword
        # under the physical interface or the bridge.  We only support configuring
        # these under the physical interface where validrange is used so delete the
        # validvals entry.
        if self.keyword in ('bridge-pathcosts', 'mstpctl-portpathcost') and 'validvals' in value and 'validrange' in value:
            del value['validvals']

        if 'validvals' in value:
            sorted_validvals = sorted(value['validvals'])

            # Compress the various ways of saying "yes or no" down to "yes or no"
            if (sorted_validvals == ['0', '1', 'no', 'yes'] or
                sorted_validvals == ['no', 'yes']):
                value['validvals'] = ['yes', 'no']

            # bond-mode takes a number of keyword, chop the numbers
            elif sorted_validvals == ['0', '1', '2', '3', '4', '5', '6', '802.3ad', 'active-backup', 'balance-alb', 'balance-rr', 'balance-tlb', 'balance-xor', 'broadcast']:
                # value['validvals'] = ['802.3ad', 'active-backup', 'balance-alb', 'balance-rr', 'balance-tlb', 'balance-xor', 'broadcast']
                # We only support two of the modes
                value['validvals'] = ['802.3ad', 'balance-xor']

            elif (sorted_validvals == ['0', '1', 'fast', 'slow']):
                value['validvals'] = ['fast', 'slow']

            # bridge-learning, bridge-multicast-flood and bridge-unicast-flood all
            # list <interface-on-off-list> to support the old bridge driver. We (nclu)
            # do not support the old bridge driver though so chop this option.
            elif (sorted_validvals == ['<interface-on-off-list>', 'off', 'on']):
                value['validvals'] = ['off', 'on']

            elif self.keyword == 'address-virtual4':
                value['validvals'] = ['<mac> <ipv4/prefixlen>', ]

            elif self.keyword == 'address-virtual6':
                value['validvals'] = ['<mac> <ipv6/prefixlen>', ]

            # These are keywords that can be configured under the bridge via <interface-yes-no-list>
            # or under the swp via 'yes/no'. We will restrict these keywords to just the swp so change
            # the validvals to 'yes/no'
            if self.keyword in ('bridge-portmcrouter',
                                'bridge-portmcfl',
                                'mstpctl-bpduguard',
                                'mstpctl-portadminedge',
                                'mstpctl-portautoedge',
                                'mstpctl-portbpdufilter',
                                'mstpctl-portnetwork',
                                'mstpctl-portp2p',
                                'mstpctl-portrestrrole',
                                'mstpctl-portrestrtcn',
                                'mstpctl-treeportprio'):
                value['validvals'] = ['yes', 'no']

            if self.default is not None:

                if self.default in ('0', 0, 'off', 'no'):
                    options_to_remove = ['0', 0, 'off', 'no']
                elif self.default in ('1', 1, 'on', 'yes'):
                    options_to_remove = ['1', 1, 'on', 'yes']
                else:
                    options_to_remove = [self.default, ]

                for tmp_default in options_to_remove:
                    if tmp_default in value['validvals']:
                        value['validvals'].remove(tmp_default)

            for val in value['validvals']:

                if isinstance(val, str):
                    if val == '<ipaddr>':
                        self.validvals.append('<ip>')

                    elif val == '<ipaddr/prefixlen>':
                        self.validvals.append('<ip/prefixlen>')

                    elif val == '<interface>':
                        self.validvals.append('<interface-as-arg>')

                    # 1.1.1.1 or "1.1.1.1 vrf BLUE"
                    elif val == '<ipv4-vrf-text>':
                        self.validvals.append('<ipv4> [vrf <text>]')

                    else:
                        self.validvals.append(val)

                elif isinstance(val, int):
                    self.validvals.append(str(val))

                else:
                    raise Exception("Unexpected validvals type %s for %s" % (type(val), val))

            add_string = "    iface add interface <interface> %s " % self.keyword
            del_string = "    iface del interface <interface> %s " % self.keyword

            if len(self.validvals) == 1:
                if not self.auto_add_yes():
                    add_string += self.validvals[0]
                    del_string += '[' + self.validvals[0] + ']'

            # The only time validvals will be [] is if the only option for the
            # command was the default value (which we removed).
            #
            # As of today, mstpctl-forcevers is the only command that falls
            # into this bucket.
            elif self.validvals:
                # (no|yes) just looks weird so do not sort this one
                if self.validvals == ['yes', 'no']:
                    add_string += "(yes|no)"
                    del_string += "[yes|no]"
                else:
                    add_string += "(%s)" % '|'.join(sort_for_humans(self.validvals))
                    del_string += "[%s]" % '|'.join(sort_for_humans(self.validvals))

            # We only support vlan aware bridges, do not provide a means to disable it
            if self.keyword == 'bridge-vlan-aware':
                del_string = ''

        elif 'validrange' in value:
            self.rangemin = int(value['validrange'][0])
            self.rangemax = int(value['validrange'][1])
            add_string = "    iface add interface <interface> %s <%d-%d>" % (self.keyword, self.rangemin, self.rangemax)
            del_string = "    iface del interface <interface> %s [<%d-%d>]" % (self.keyword, self.rangemin, self.rangemax)

        else:
            add_string = "    iface add interface <interface> %s <wildcard>" % self.keyword
            del_string = "    iface del interface <interface> %s [<wildcard>]" % self.keyword

        # babelfish scenarios
        # This is a special case...we override 'address (<ipv4/prefixlen>|<ipv6/prefixlen>)' with
        # "[ip] address (<ipv4/prefixlen>|<ipv6/prefixlen>|dhcp)" command
        if self.keyword == 'address4':
            add_string = "    iface add interface <interface> ip address (<ipv4/prefixlen>|dhcp)"
            del_string = "    iface del interface <interface> ip address [<ipv4/prefixlen>|dhcp]"
            self.keyword_original = 'address'
            self.keyword = ['ip', 'address']
            self.validvals.append('dhcp')

        elif self.keyword == 'address6':
            add_string = "    iface add interface <interface> ipv6 address (<ipv6/prefixlen>|dhcp)"
            del_string = "    iface del interface <interface> ipv6 address [<ipv6/prefixlen>|dhcp]"
            self.keyword_original = 'address'
            self.keyword = ['ipv6', 'address']
            self.validvals.append('dhcp')

        elif self.keyword == 'address-virtual4':
            add_string = "    iface add interface <interface> ip address-virtual <mac> <ipv4/prefixlen>"
            del_string = "    iface del interface <interface> ip address-virtual [<mac> <ipv4/prefixlen>]"
            self.keyword_original = 'address-virtual'
            self.keyword = ['ip', 'address-virtual']

        elif self.keyword == 'address-virtual6':
            add_string = "    iface add interface <interface> ipv6 address-virtual <mac> <ipv6/prefixlen>"
            del_string = "    iface del interface <interface> ipv6 address-virtual [<mac> <ipv6/prefixlen>]"
            self.keyword_original = 'address-virtual'
            self.keyword = ['ipv6', 'address-virtual']

        elif self.keyword == 'gateway4':
            add_string = "    iface add interface <interface> ip gateway <ipv4>"
            del_string = "    iface del interface <interface> ip gateway [<ipv4>]"
            self.keyword_original = 'gateway'
            self.keyword = ['ip', 'gateway']

        elif self.keyword == 'gateway6':
            add_string = "    iface add interface <interface> ipv6 gateway <ipv6>"
            del_string = "    iface del interface <interface> ipv6 gateway [<ipv6>]"
            self.keyword_original = 'gateway'
            self.keyword = ['ipv6', 'gateway']

        elif self.keyword == 'netmask4':
            add_string = "    iface add interface <interface> ip netmask <ipv4>"
            del_string = "    iface del interface <interface> ip netmask [<ipv4>]"
            self.keyword_original = 'netmask'
            self.keyword = ['ip', 'netmask']

        elif self.keyword == 'netmask6':
            add_string = "    iface add interface <interface> ipv6 netmask <ipv6>"
            del_string = "    iface del interface <interface> ipv6 netmask [<ipv6>]"
            self.keyword_original = 'netmask'
            self.keyword = ['ipv6', 'netmask']

        elif self.keyword == 'bridge-stp':
            add_string = "    iface add interface <interface> stp off"
            del_string = "    iface del interface <interface> stp off"
            self.keyword_original = 'bridge-stp'
            self.keyword = ['bridge', 'stp']

        # link-autoneg, link-duplex, link-fec and link-speed
        elif self.keyword.startswith('link-'):
            add_string = add_string.replace('link-', 'link ')
            del_string = del_string.replace('link-', 'link ')
            self.keyword = ['link', '-'.join(self.keyword.split('-')[1:])]

        elif self.keyword.startswith('bond-'):
            add_string = add_string.replace('bond-', 'bond ')
            del_string = del_string.replace('bond-', 'bond ')
            self.keyword = ['bond', '-'.join(self.keyword.split('-')[1:])]

        elif self.keyword.startswith('bridge-'):
            add_string = add_string.replace('bridge-', 'bridge ')
            del_string = del_string.replace('bridge-', 'bridge ')
            self.keyword = ['bridge', '-'.join(self.keyword.split('-')[1:])]

        elif self.keyword.startswith('clag-'):
            add_string = add_string.replace('clag-', 'clag ')
            del_string = del_string.replace('clag-', 'clag ')
            self.keyword = ['clag', '-'.join(self.keyword.split('-')[1:])]

        elif self.keyword.startswith('clagd-'):
            add_string = add_string.replace('clagd-', 'clag ')
            del_string = del_string.replace('clagd-', 'clag ')
            self.keyword = ['clag', '-'.join(self.keyword.split('-')[1:])]

        elif self.keyword.startswith('ip-'):
            add_string = add_string.replace('ip-', 'ip ')
            del_string = del_string.replace('ip-', 'ip ')
            self.keyword = ['ip', 'forward']

        elif self.keyword.startswith('ip6-'):
            add_string = add_string.replace('ip6-', 'ipv6 ')
            del_string = del_string.replace('ip6-', 'ipv6 ')
            self.keyword = ['ipv6', 'forward']

        # mstpctl-bpduguard
        elif self.keyword.startswith('mstpctl-'):
            add_string = add_string.replace('mstpctl-', 'stp ')
            del_string = del_string.replace('mstpctl-', 'stp ')
            self.keyword = ['stp', '-'.join(self.keyword.split('-')[1:])]

        elif self.keyword.startswith('vxlan-'):
            add_string = add_string.replace('vxlan-', 'vxlan ')
            del_string = del_string.replace('vxlan-', 'vxlan ')
            self.keyword = ['vxlan', '-'.join(self.keyword.split('-')[1:])]

        keyword_scope_defined = False

        int_types = ('bond', 'bridge', 'loopback', 'vlan', 'vxlan', 'vrf')
        if self.keyword_original in all_interface_keywords:
            keyword_scope_defined = True
            self.usage_str_add.append(add_string)
            self.usage_str_del.append(del_string)

            for int_type in int_types:

                # bridge type is special, we restrict its name to 'bridge'
                if int_type == 'bridge':
                    tmp_add_string = add_string.replace(' interface <interface> ', ' bridge ')
                    tmp_del_string = del_string.replace(' interface <interface> ', ' bridge ')

                elif int_type == 'loopback':
                    tmp_add_string = add_string.replace(' interface <interface> ', ' loopback lo ')
                    tmp_del_string = del_string.replace(' interface <interface> ', ' loopback lo ')

                elif int_type == 'vlan':
                    tmp_add_string = add_string.replace(' interface <interface> ', ' vlan <number-range-list> ')
                    tmp_del_string = del_string.replace(' interface <interface> ', ' vlan <number-range-list> ')

                else:
                    tmp_add_string = add_string.replace(' interface ', ' %s ' % int_type)
                    tmp_del_string = del_string.replace(' interface ', ' %s ' % int_type)

                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

        else:
            if self.keyword_original in bridge_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface <interface> ', ' bridge ')
                tmp_del_string = del_string.replace(' interface <interface> ', ' bridge ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword_original in bond_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' bond ')
                tmp_del_string = del_string.replace(' interface ', ' bond ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword_original in loopback_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface <interface> ', ' loopback lo ')
                tmp_del_string = del_string.replace(' interface <interface> ', ' loopback lo ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword_original in physical_interface_keywords:
                keyword_scope_defined = True
                self.usage_str_add.append(add_string)
                self.usage_str_del.append(del_string)

            if self.keyword_original in vlan_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface <interface> ', ' vlan <number-range-list> ')
                tmp_del_string = del_string.replace(' interface <interface> ', ' vlan <number-range-list> ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword_original in vxlan_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' vxlan ')
                tmp_del_string = del_string.replace(' interface ', ' vxlan ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

            if self.keyword_original in vrf_interface_keywords:
                keyword_scope_defined = True
                tmp_add_string = add_string.replace(' interface ', ' vrf ')
                tmp_del_string = del_string.replace(' interface ', ' vrf ')
                self.usage_str_add.append(tmp_add_string)
                self.usage_str_del.append(tmp_del_string)

        if not keyword_scope_defined:
            # We currently do not do anything with the deprecated commands
            # so don't bother spewing the log for these
            if not self.deprecated:
                log.info('Where does keyword %s belong?' % self.keyword)
            self.usage_str_add.append(add_string)
            self.usage_str_del.append(del_string)

    def __str__(self):
        return str(self.keyword)

    def auto_add_yes(self):
        """
        Return True if we should automagically add "yes" to a keyword when
        writing that keyword to an interface stanza.  Otherwise, return False.
        """
        return self.validvals == ['yes']

    def value_from_args(self, args):
        args_value_keyword = None

        if self.auto_add_yes():
            args_value_keyword = 'yes'

        elif self.validvals:
            for val in self.validvals:
                # Combos such as '<ipv4> [vrf <text>]' are a special case
                if ' ' in val:
                    values = []

                    for val_component in val.split():
                        if val_component.startswith('['):
                            val_component = val_component[1:]

                        if val_component.endswith(']'):
                            val_component = val_component[:-1]

                        tmp = args.get(val_component)
                        if tmp:
                            values.append(tmp)

                    if values:
                        return ' '.join(values)
                else:
                    if args.get(val) is not None:
                        args_value_keyword = val
                        break

        elif self.rangemax:
            args_value_keyword = "<{0}-{1}>".format(self.rangemin, self.rangemax)

        elif self.default:
            args_value_keyword = self.default

        else:
            if args.get('<wildcard>'):
                args_value_keyword = '<wildcard>'

        if args_value_keyword is None:
            return None

        return args.get(args_value_keyword) if args_value_keyword.startswith('<') else args_value_keyword

    def keywords_in_args(self, args):

        # 99% of commands have only one keyword, the exceptions are the ones
        # like "ipv6 address-virtual" where we created two Command objects
        # from a single keyword. In that scenario self.keyword is a list
        # instead of a string.
        if isinstance(self.keyword, list):
            keywords_to_check = self.keyword
        else:
            keywords_to_check = [self.keyword, ]

        args_has_all_keywords = True
        for keyword in keywords_to_check:
            if args.get(keyword) is None:
                args_has_all_keywords = False
                break
        return args_has_all_keywords


class CommandAddDelMultiLine(Command):
    """
    Commands that are allowed to appear multiple times in and interface stanza...example:

    auto swp1
    iface swp1
        address 1.1.1.1/24
        address 2.2.2.1/24
    """
    pass


class CommandAddDelMultiValue(Command):
    """
    Commands that can only appear once in an interface stanza but can have a list of values...example:

    auto br0
    iface br0
        bridge-ports swp1 swp2 swp7
    """
    pass


class CommandSetClear(Command):
    """
    Commands that can only appear one time in an interface stanza...example:

    auto vni100
    iface vni100
        vxlan-id 1.1.1.1
    """
    pass


def build_doc_string(blacklist):
    """
    Parse the _modinfo dictionary for the various ifupdownaddons modules.
     _modinfo contains information about the commands supported by the addon.

    We use this to:
    - modify __doc__ to include the CLI to support each command
    - create a Command object for each command
    """
    global __doc__
    global addons_cmds

    state = 'LIMBO'
    usage = []
    options = []

    # "/sbin/ifquery --syntax-help -o json" dumps a dictionary of
    # the addons and all of the keywords they support.
    cmd = ['/sbin/ifquery', '--syntax-help', '-o', 'json']
    try:
        output = subprocess.check_output(cmd)
    except subprocess.CalledProcessError as e:
        output = '{}'
        log.warning("'%s' failed due to %s" % (' '.join(cmd), str(e)))

    try:
        ifupdown_syntax = json.loads(output, object_hook=decode_dict)
    except ValueError:
        ifupdown_syntax = {}
        log.warning("'%s' produced invalid json output\n%s\n" % (' '.join(cmd), pformat(output)))

    for addon_name in ifupdown_syntax.iterkeys():
        addon_attrs = ifupdown_syntax[addon_name].get('attrs', {})

        for (keyword, value) in addon_attrs.iteritems():

            # Some keywords are corner cases so we do not list them in order to
            # avoid clutter when the user does TAB complete
            if keyword in blacklist:
                continue

            # babelfish
            # This would look weird (keyword 'ip' followed by an ipv6 address)
            #   net add interface <interface> ip gateway (<ipv4>|<ipv6>)
            #
            # So we create two branches for the 'gateway' keyword
            #   net add interface <interface> ip gateway <ipv4>
            #   net add interface <interface> ipv6 gateway <ipv6>
            if keyword == 'address':
                keywords = ('address4', 'address6')
            elif keyword == 'address-virtual':
                keywords = ('address-virtual4', 'address-virtual6')
            elif keyword == 'gateway':
                keywords = ('gateway4', 'gateway6')
            elif keyword == 'netmask':
                keywords = ('netmask4', 'netmask6')
            else:
                keywords = (keyword, )

            for keyword in keywords:
                # Commands like 'address' where you are allowed to have more than one of
                # them per interface
                if 'multiline' in value:
                    cmd = CommandAddDelMultiLine(keyword, value)

                # Commands like "bridge-ports" where you can have
                # "bridge-ports swp2 swp3 swp4"
                elif 'multivalue' in value:
                    cmd = CommandAddDelMultiValue(keyword, value)

                # Commands where you can only have one (vxlan-id)
                else:
                    cmd = CommandSetClear(keyword, value)
                addons_cmds.append(cmd)

    # ====================
    # __doc__ construction
    # ====================

    # Parse the initial docstring, extract the Usage and Options lines
    for line in __doc__.splitlines():
        if line == 'Usage:':
            state = 'USAGE'
        elif line == 'Options:':
            state = 'OPTIONS'
        else:
            if state == 'LIMBO':
                pass
            elif state == 'USAGE':
                usage.append(line)
            elif state == 'OPTIONS':
                options.append(line)
            else:
                raise Exception("Invalid state %s, line is %s" % (state, line))

    addon_usage = []
    # Build all of the Usage and Options lines
    for cmd in addons_cmds:
        if not cmd.deprecated:
            for line in cmd.usage_str_add:
                addon_usage.append(line)

            for line in cmd.usage_str_del:
                addon_usage.append(line)

            if cmd.help:
                options.append("%s : %s" % (cmd.keyword_original, cmd.help))

    addon_usage.sort()
    usage.extend(addon_usage)

    __doc__ = '\nUsage:\n' + '\n'.join(usage) + '\n\nOptions:\n' + '\n'.join(options)

    return __doc__
    # End of __doc__ construction


class EtcNetworkInterface(object):
    """
    Used to represent an interface configuration stanza in the format expected in
    /etc/network/interfaces.
    """

    def __init__(self, name, ifquery_all):
        """
        From the output of "ifquery -a", extract the stanza for the given interface
        (name), and store it, as a list of lines, in self.ifquery.
        """

        self.content = []
        self.name = str(name)
        self.filename = "/var/run/nclu/iface/{0}".format(self.name)  # Pending changes
        self.ifquery = []

        # Extract the configuration for this interface from the "ifquery -a" output.
        found_auto = False
        found_iface = False
        found_name = False

        for line in ifquery_all.splitlines():
            line = standardize_indent(line)
            re_auto = re.search("^auto\s+{0}$".format(self.name), line)
            re_iface = re.search("^iface\s+{0}$".format(self.name), line)
            re_iface_inet = re.search("^iface\s+{0}\s+inet".format(self.name), line)

            if re_auto:
                found_name = True

                if not found_auto:
                    self.ifquery.append(line)
                    found_auto = True

            elif re_iface or re_iface_inet:
                found_name = True

                if not found_iface:
                    self.ifquery.append(line)
                    found_iface = True

            elif found_name:
                # We've started parsing the next interface.
                if line.startswith('auto ') or line.startswith('source '):
                    found_name = False
                elif line and line not in self.ifquery:
                    self.ifquery.append(line)

        if os.path.isfile(self.filename):
            # Populate self.content with the pending changes file.
            self.load_iface_state()
        else:
            # Populate self.content with self.ifquery, etc.
            self.init_iface_state()

    def __str__(self):
        return self.name

    def add_cmd(self, cmd):
        self.content.append(cmd)

    def del_cmd(self, cmd, match_exact, give_feedback):

        # If an interface is configured with "vxlan-id 500" and the user
        # does "net del int swp1 vxlan-id 500" then look for this exact line
        if match_exact:
            if cmd not in self.content:
                if give_feedback:
                    print "%s's configuration does not have '%s'" % (self, cmd.strip())
            else:
                self.content.remove(cmd)

        # But if the user did "net del int swp1 vxlan-id" then we just need to
        # look for the lines that start with "    vxlan-id "
        else:
            lines_to_remove = []

            # Do not change the size of the list as we are looping over it, store
            # the ones to delete and then come back and delete them
            for tmp_cmd in self.content:
                if tmp_cmd.startswith('%s ' % cmd):
                    lines_to_remove.append(tmp_cmd)

            if lines_to_remove:
                for tmp_cmd in lines_to_remove:
                    self.content.remove(tmp_cmd)
            else:
                if give_feedback:
                    print "%s's configuration does not have '%s'" % (self, cmd.strip())

    def mod_cmd(self, old_cmd, new_cmd, give_feedback):

        if old_cmd == new_cmd:
            if give_feedback:
                print "%s's configuration already has '%s'" % (self, new_cmd.strip())
            return

        for (i, line) in enumerate(self.content):
            if line == old_cmd:
                self.content[i] = new_cmd
                return

    def get_cmd(self, keyword):
        keyword = keyword.strip() + ' '

        for line in self.content:
            if line.lstrip().startswith(keyword):
                return line
        return None

    def has_cmd(self, keyword):
        return self.get_cmd(keyword) is not None

    def get_keyword_value(self, keyword):
        # line will be something like "mtu 1500" or "bridge-ports swp1 swp2" so
        # split it and return all values after the keyword
        line = self.get_cmd(keyword)

        if line:
            line = line.strip()

            if ' ' in line:
                return ' '.join(line.split()[1:])
        return None

    def get_keyword_value_list(self, keyword):
        """
        For commands like "address" where there can be multiple lines for a keyword
        """
        keyword = keyword.strip() + ' '
        results = []

        for line in self.content:
            line = line.strip()
            if line.startswith(keyword):
                results.append(line.replace(keyword, ''))
        return results

    def is_bond(self):
        return self.has_cmd('bond-slaves')

    def will_become_bond(self, cli):
        return cli.args.get('bond') is not None and cli.args.get('slaves') is not None

    def is_bridge(self):
        return self.has_cmd('bridge-ports') or self.has_cmd('bridge-vlan-aware')

    def will_become_bridge(self, cli):
        return cli.args.get('bridge') is not None and (cli.args.get('ports') is not None or cli.args.get('vlan-aware') is not None)

    def is_layer2(self):
        # You also need to look at the bridge bridge-ports to see if this iface is listed
        return self.has_cmd('bridge-access') or self.has_cmd('bridge-vids')

    def is_loopback(self):
        for line in self.content:
            if line.rstrip().endswith('inet loopback'):
                return True
        return False

    def is_vxlan(self):
        return self.has_cmd('vxlan-id') or self.has_cmd('vxlan-local-tunnelip')

    def will_become_vxlan(self, cli):
        return cli.args.get('vxlan') is not None and (cli.args.get('id') is not None or cli.args.get('local-tunnelip') is not None)

    def is_vlan_interface(self):
        return self.has_cmd('vlan-raw-device') or re.search('^bridge\.\d+$', self.name)

    def will_become_vlan_interface(self, cli):
        return cli.args.get('vlan') is not None and cli.args.get('raw-device') is not None

    def is_vrf(self):
        return self.has_cmd('vrf-table')

    def exists(self):
        """
        If we are editing this interface via iface, then the interface exists.
        self.content will contain the output of ifquery.  If the interface
        is not defined in /etc/network/interfaces, then ifquery will return
        an error, and self.content will be empty.
        """
        return os.path.isfile(self.filename) or self.content

    def validate_net_cmd_vs_ifname(self, parent_wrapper, cli):
        """
        The user must use "net add bond" if editing a bond interface, "net add bridge"
        for the bridge interface, etc. Return True if they used the correct syntax, else
        return False.
        """
        argv_expanded_str = ' '.join(map(str, cli.argv_expanded))

        net_add_del_bond = 'net add bond ' in argv_expanded_str or 'net del bond ' in argv_expanded_str
        net_add_del_vxlan = 'net add vxlan ' in argv_expanded_str or 'net del vxlan ' in argv_expanded_str

        # Sanity check that the correct "net (add|del) ..." syntax was used to
        # edit this type of interface.
        if self.is_bridge() or self.will_become_bridge(cli) or self.name == 'bridge':
            net_add_del_bridge = 'net add bridge' in argv_expanded_str or 'net del bridge' in argv_expanded_str

            if not net_add_del_bridge:
                print "ERROR: 'net (add|del) bridge ...' must be used to edit the bridge interface"
                return False

        elif self.is_bond() or self.will_become_bond(cli):
            net_del_clag_port = 'net del clag port' in argv_expanded_str

            if not net_add_del_bond and not net_del_clag_port:
                print "ERROR: 'net (add|del) bond ...' must be used to edit a bond interface"
                return False

        elif net_add_del_bond:
            print "ERROR: %s is not a bond interface. Use 'net add bond %s bond slaves ...' to make %s a bond interface" % (self.name, self.name, self.name)
            return False

        elif self.is_loopback() or self.name == 'lo':
            net_add_del_loopback = 'net add loopback ' in argv_expanded_str or 'net del loopback ' in argv_expanded_str

            if not net_add_del_loopback:
                print "ERROR: 'net (add|del) loopback ...' must be used to edit a loopback interface"
                return False

        elif self.is_vlan_interface() or self.will_become_vlan_interface(cli):
            net_add_del_vlan = 'net add vlan ' in argv_expanded_str or 'net del vlan ' in argv_expanded_str

            if not net_add_del_vlan:
                print "ERROR: 'net (add|del) vlan ...' must be used to edit a vlan interface"
                return False

        elif self.is_vxlan() or self.will_become_vxlan(cli):

            if not net_add_del_vxlan:
                print "ERROR: 'net (add|del) vxlan ...' must be used to edit a vxlan interface"
                return False

        elif net_add_del_vxlan:
            print "ERROR: 'net (add|del) interface ...' must be used, %s is not a vxlan interface" % self.name
            return False

        elif self.is_vrf():
            net_add_del_vrf = 'net add vrf ' in argv_expanded_str or 'net del vrf ' in argv_expanded_str

            if not net_add_del_vrf:
                print "ERROR: 'net (add|del) vrf ...' must be used to edit a vrf interface"
                return False

        # If the user enters "<name>.100", then check that the name is either a physical interface or bond
        # and not "bridge" (which indicates a SVI).  If so, return True.  Otherwise, return False.
        # TODO - Remove the bond slaves parameter.  It is a work-around for bond-slaves not being implemented, which makes tests fails.

        base_name = re.match(r"""^(?P<base>\S+?)(\.\d+)+$""", self.name)

        if base_name is not None:
            base = base_name.group('base')
            base_iface = EtcNetworkInterface(base, parent_wrapper.ifquery_all)
            bond_slaves = cli.args.get('bond') is not None and cli.args.get('slaves') is not None

            if base == 'bridge':
                print "ERROR: 'net (add|del) vlan ...' must be used to edit a vlan interface."
                return False
            elif not (self.is_physical(base, True) or bond_slaves or self.is_bond() or base_iface.is_bond()):
                print "ERROR: For interface names of the form <name>.<number>, the interface must be physical or a bond."
                return False

        elif argv_expanded_str.startswith('net add interface'):
            # If the user is doing 'net add interface' then the interface must be
            # physical. The exception to this is if they are configuring the
            # breakout keyword, for that scenario there may be a physical interface
            # that is not listed in /sys/class/net so only check that the interface
            # name begins with "swpX"
            if cli.args.get('breakout'):
                if not re.match(r"""^swp\d+$""", self.name):
                    print "ERROR: {0} is not a physical interface on this switch.".format(self.name)
                    return False
            else:
                if not self.is_physical():
                    return False

        return True

    def write_file(self):
        """
        Write self.content to a file in /var/run/nclu/iface/
        """

        with open(self.filename, 'w') as fh:
            commands_to_sort = []
            found_iface = False

            for line in self.content:

                if found_iface:
                    commands_to_sort.append(line)

                # 99% of the time, this will only be the 'auto' and 'iface' lines.
                else:
                    fh.write(line + '\n')

                if line.startswith('iface '):
                    found_iface = True

            # Sort all of the commands in this stanza.  This is
            # to make the stanza easier for the user to read.
            for line in sort_for_humans(commands_to_sort):
                fh.write(line + '\n')

    def is_swpx(self):
        return re.match(r"""^swp\d+(?:s\d+)?$""", self.name) is not None

    def is_physical(self, base_name=None, suppress_error=False):
        """
        Verify the interface is named like swp4, swp4s1 (breakout), or "eth0" and that it has an
        associated file in the directory /sys/class/net/.
        """
        if base_name is None:
            base_name = self.name
        if base_name == 'eth0' or (re.match(r"""^swp\d+(?:s\d+)?$""", base_name) is not None and base_name in os.listdir('/sys/class/net/')):
            return True
        if not suppress_error:
            print "ERROR: {0} is not a physical interface on this switch.".format(base_name)
        return False

    def add_interface(self):
        if self.content:
            print "{0} already exists in /etc/network/interfaces.  No need to add it.".format(self.name)
            return False

        # Add the basic commands needed (auto and iface) for this interface.
        self.add_cmd("auto {0}".format(self.name))

        if self.name == 'lo':
            self.add_cmd("iface {0} inet loopback".format(self.name))
        else:
            self.add_cmd("iface {0}".format(self.name))

        self.write_file()
        return True

    def del_interface(self):
        """
        The name not withstanding, this method marks an interface for deletion by creating
        a file with the extension ".del".  No changes are made to ETC_NETWORK_INTERFACES.
        """

        self.content = []
        filename_del = self.filename + '.del'
        filename_original = self.filename + '.original'
        valid_interface = False

        if os.path.isfile(self.filename):
            if os.path.getsize(self.filename):
                valid_interface = True

            # Remove the pending changes file.
            os.remove(self.filename)

        if os.path.isfile(filename_original) and not os.path.getsize(filename_original):
            # Remove the empty .original file.
            os.remove(filename_original)

            if not valid_interface:
                print "{0} does not exist to delete.".format(self.name)

            return

        if not os.path.isfile(filename_del):
            # Write an empty .del file so we know this is one that we are trying to delete.
            # The .del file will be deleted by commit_pending().
            open(filename_del, 'w').close()

    def del_interface_content(self):
        """
        Strip all uncommented content from the interface stanzas.
        """

        self.content = [line for line in self.content if line.startswith('auto ') or line.startswith('iface ') or line.startswith('#')]
        self.write_file()

    def conf_key_value_multiple_lines(self, add, key, value, args):
        """
        iface config lines where there can be more than one.  Example:
            address 1.1.1.1/32
            address 1.1.1.2/32
        """
        iface_dhcp = "iface {0} inet dhcp".format(self.name)
        iface_dhcp6 = "iface {0} inet6 dhcp".format(self.name)
        iface_static = "iface {0}".format(self.name)
        give_feedback = True

        if key == 'address' and value == 'dhcp':
            if add:
                if iface_static in self.content:

                    if args.get('ipv6'):
                        self.mod_cmd(iface_static, iface_dhcp6, False)
                    else:
                        self.mod_cmd(iface_static, iface_dhcp, False)

                elif iface_dhcp in self.content:
                    if args.get('ipv6'):
                        self.mod_cmd(iface_dhcp, iface_dhcp6, False)

                elif iface_dhcp6 in self.content:
                    if not args.get('ipv6'):
                        self.mod_cmd(iface_dhcp6, iface_dhcp, False)

                # Delete any static IPs assigned
                if self.get_cmd('address'):
                    self.del_cmd('    address', False, give_feedback)
            else:
                if iface_dhcp in self.content:
                    self.mod_cmd(iface_dhcp, iface_static, False)

                elif iface_dhcp6 in self.content:
                    self.mod_cmd(iface_dhcp6, iface_static, False)

        else:
            # make sure iface line does not contain 'dhcp'
            if key == 'address':
                if iface_dhcp in self.content:
                    self.mod_cmd(iface_dhcp, iface_static, False)

            if add:
                cmd = "    %s %s" % (key, value)

                if cmd in self.content:
                    if give_feedback:
                        print "%s's configuration already has '%s'" % (self, cmd.strip())
                else:
                    self.add_cmd(cmd)
            else:
                if value:
                    cmd = "    %s %s" % (key, value)
                    self.del_cmd(cmd, True, give_feedback)
                else:
                    cmd = "    %s" % key
                    self.del_cmd(cmd, False, give_feedback)

        self.write_file()

    def conf_key_value_multiple_values(self, add, key, values, give_feedback=True):
        """
        Some commands (bridge-ports for example) take a list of values from which
        we need to be able to add/del.

        values - A string of comma-separated values, or the empty string; possibly a glob
        """

        # This prevents undesirable downstream assertion failures.
        if not isinstance(values, str):
            values = str(values)

        assert isinstance(values, str), "values={0}; type(values)={1}".format(values, type(values))
        ifaces_to_add = []

        # Return early here if we know that all we need to do is nuke all
        # lines for this keyword.  This is the case when add=False and values=''.
        if not add and not values:
            cmd = "    {0}".format(key)
            self.del_cmd(cmd, False, give_feedback)
            self.write_file()
            return ifaces_to_add

        # numeric globs...used by bridge-vids
        if is_number_glob(values):
            # CM-16899: If the user issues a command like "net add bridge bridge port 12",
            # error out because the user forgot the rest of the switchport name.  This
            # should not actually be a number glob.
            if key.lower() in ('bridge-ports', 'bond-slaves'):
                print "ERROR: '{0}' is not a valid list of interfaces.".format(values)
                return []

            old_cmd = self.get_cmd("    {0} ".format(key))

            if old_cmd:
                values = glob_to_numbers(values)  # A sorted, deduplicated list
                old_values = ','.join(old_cmd.split()[1:])
                old_values = glob_to_numbers(old_values)

                if add:
                    if old_values:
                        final_values = old_values + values
                    else:
                        final_values = values
                else:
                    for x in values:
                        if x in old_values:
                            old_values.remove(x)
                    final_values = old_values

                if final_values:
                    glob = numbers_to_glob(final_values)
                    new_cmd = "    %s %s" % (key, glob.replace(',', ' '))

                    if new_cmd != old_cmd:
                        self.mod_cmd(old_cmd, new_cmd, give_feedback)
                else:
                    self.del_cmd(old_cmd, True, give_feedback)

            else:
                if add:
                    cmd = "    {0} {1}".format(key, values.replace(',', ' '))
                    self.add_cmd(cmd)

        # interfaces
        else:
            if key.startswith('address-virtual'):
                values = values.split()
            else:
                values_globs_expanded = ifname_expand_glob(values)  # A list of strings
                values = sort_for_humans(list(set(values_globs_expanded)))

            assert isinstance(values, list)

            # These two keywords accept interfaces as their argument.  Validate that
            # names beginning with "swp*" are valid physical interfaces.
            if add and key in ('bridge-ports', 'bond-slaves'):
                all_interfaces = os.listdir('/sys/class/net/')
                to_remove = []

                for ifname in values:
                    if re.match(r"""^swp\d+$""", ifname) and ifname not in all_interfaces:
                        to_remove.append(ifname)

                for ifname in to_remove:
                    if give_feedback:
                        print "{0} is not a physical interface on this switch".format(ifname)
                    values.remove(ifname)

                if not values:
                    add = False

            old_cmd = self.get_cmd("    %s " % key)

            if old_cmd:
                old_cmd_expanded = expand_bridge_ports_glob(old_cmd)
                old_values = old_cmd_expanded.split()[1:]

                if add:

                    # The order of the values matters for address-virtual so we
                    # can't just create a set and convert back to a list
                    if key.startswith('address-virtual'):
                        final_values = []
                        for value in (old_values + values):
                            if is_mac_address(value):
                                continue

                            if value not in final_values:
                                final_values.append(value)
                        final_values = sorted(final_values)

                    else:
                        final_values = sort_for_humans(list(set(old_values + values)))
                else:
                    for x in values:
                        if x in old_values:
                            old_values.remove(x)
                        else:
                            if give_feedback:
                                print '"iface %s" %s does not contain %s' % (self, key, x)

                    if key.startswith('address-virtual'):
                        final_values = old_values
                    else:
                        final_values = sort_for_humans(old_values)

                if final_values:
                    new_cmd = "    %s %s" % (key, ' '.join(final_values))
                    new_cmd = new_cmd.replace('regex-', 'regex ')

                    if new_cmd != old_cmd:
                        self.mod_cmd(old_cmd, new_cmd, give_feedback)

                        if key in ('bridge-ports', 'bond-slaves'):
                            # If there is a "regex FOO" that we converted to "regex-FOO"
                            # for sorting purposes, do not try to create an interface
                            # called "regex-FOO"
                            final_values = [x for x in final_values if not x.startswith('regex')]
                            ifaces_to_add = final_values
                else:
                    self.del_cmd(old_cmd, True, give_feedback)
            else:
                if add:
                    cmd = "    %s %s" % (key, ' '.join(values))
                    self.add_cmd(cmd)

                    # Automatically add the interfaces specifed in bridge-ports or bond-slaves
                    if key in ('bridge-ports', 'bond-slaves'):
                        ifaces_to_add = values

        self.write_file()

        return ifaces_to_add

    def conf_key_value_solo(self, add, key, value, give_feedback=True):
        """
        iface configs lines where there can be only one. Example:
        bridge_fd 1
        """
        key = key.rstrip()

        if add:
            new_cmd = "    %s %s" % (key, value)
            old_cmd = self.get_cmd('    %s ' % key)

            if old_cmd:
                self.mod_cmd(old_cmd, new_cmd, give_feedback)
            else:
                self.add_cmd(new_cmd)
        else:
            if value:
                cmd = "    %s %s" % (key, value)
                self.del_cmd(cmd, True, give_feedback)
            else:
                cmd = "    %s" % key
                self.del_cmd(cmd, False, give_feedback)
        self.write_file()

    def load_iface_state(self):
        """
        Populate self.content with lines from the pending changes file.
        """

        self.content = []
        filename_original = self.filename + '.original'

        if os.path.exists(self.filename):
            filename = self.filename
        elif os.path.exists(filename_original):
            filename = filename_original
        else:
            filename = None
            log.debug("{0}: {1} and {2} do not exist".format(self, self.filename, filename_original))

        if filename is not None:
            with open(filename) as f:
                for line in f:
                    line = standardize_indent(line)

                    # Ignore blank lines.
                    if line:
                        self.content.append(line)

    def init_iface_state(self):
        """
        Populate self.content with the current /etc/network/interfaces config,
        as output by "ifquery -a", for the given interface, unless a .del
        file exists.
        """

        if not os.path.isfile(self.filename + '.del'):
            self.content = self.ifquery

        # Remember the /etc/network/interfaces state of this interface. We
        # use this later to determine if someone is modifying the file and
        # modifying via iface (not supported).
        #
        # This file will be removed by commit_pending().
        with open(self.filename + '.original', 'w') as f:
            for line in self.content:
                f.write("{0}\n".format(line))


def show_bridge_vlan(args):
    # Build the vlans dict from the bridge json output
    try:
        vlan_show_output = check_output(['/sbin/bridge', '-c', '-j', 'vlan', 'show'])
    except Exception:
        vlan_show_output = '{}'

    try:
        vlans = json.loads(vlan_show_output)
    except Exception:
        vlans = {}

    # iproute2 does not yet support json so we must screen scrape :(
    try:
        ip_link_show_output = check_output(['/sbin/ip', '-d', 'link', 'show'])
    except Exception:
        ip_link_show_output = ''

    # Now combine the output from the two...basically just add
    # the VNI for each interface to the vlans dict
    ifname = None
    found_vni = False
    target_vlan = args.get('<number>')

    if target_vlan:
        ifnames_to_delete = []
        for ifname in vlans.keys():
            has_target_vlan = False
            vlans_to_delete = []

            for vlan_dict in vlans[ifname]:
                vlan_start = vlan_dict.get('vlan')
                vlan_end = vlan_dict.get('vlanEnd')

                if not vlan_end:
                    vlan_end = vlan_start

                if target_vlan >= vlan_start and target_vlan <= vlan_end:
                    has_target_vlan = True
                    vlan_dict['vlan'] = target_vlan
                    if 'vlanEnd' in vlan_dict:
                        del vlan_dict['vlanEnd']
                else:
                    vlans_to_delete.append(vlan_dict)

            for vlan_dict in vlans_to_delete:
                vlans[ifname].remove(vlan_dict)

            if not has_target_vlan:
                ifnames_to_delete.append(ifname)

        for ifname in ifnames_to_delete:
            del vlans[ifname]

    for line in ip_link_show_output.splitlines():
        '''
        We are interested in the 1st and 3rd lines:

        93: vxlan10101: <BROADCAST,MULTICAST,UP,LOWER_UP> ....
            link/ether 06:36:34:68:a1:1b brd ff:ff:ff:ff:ff:ff promiscuity 1
            vxlan id 10101 local 6.0.0.17 srcport 0 0 dstport 4789 nolearning ageing 300
        '''
        re_ifname = re.match('\d+: (.*?):', line)
        re_vni = re.match('\s+vxlan id (\d+)', line)

        if re_ifname:
            ifname = re_ifname.group(1)
        elif re_vni and ifname in vlans:
            vni = int(re_vni.group(1))
            found_vni = True

            for vlan_dict in vlans[ifname]:
                vlan_dict['vni'] = vni

    if args.get('json'):
        sys.stdout.write(json.dumps(vlans, indent=4) + '\n')
    else:
        header = ['Interface',
                  'VLAN',
                  'Flags']

        if found_vni:
            header.append('VNI')

        data = []
        sorted_interfaces = sort_ifaces(vlans.keys())
        for ifname in sorted_interfaces:

            for vlan_dict in vlans[ifname]:
                vlan_start = vlan_dict.get('vlan')
                vlan_end = vlan_dict.get('vlanEnd')
                flags = vlan_dict.get('flags', '')

                if flags:
                    flags = ', '.join(flags)

                if vlan_end:
                    vlan_range = "%d-%d" % (vlan_start, vlan_end)
                else:
                    vlan_range = vlan_start

                if found_vni:
                    vni = vlan_dict.get('vni', '')
                    data.append((ifname, vlan_range, flags, vni))
                else:
                    data.append((ifname, vlan_range, flags))

                # No need to repeat the ifname on every row
                ifname = ''

        sys.stdout.write('\n' + tabulate(tru(data), header) + '\n\n\n')


class ConfigWrapper(object):

    def __init__(self, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs
        self.ifquery_all = None
        self.etc_network_interfaces_init_mod_time = None
        self.ifaces = {}
        self.working_dir = '/var/run/nclu/iface/'
        self.init_working_dir()

        '''
        # This was causing issues where lo:1 would be deleted but the config
        # lines were not being moved under lo...commenting this out for now
        # to reduce risk for GA

        # When netd first starts we need to "add" the lo interface so that we
        # can migrate the configs that are under lo:1, lo:2, etc to be under lo
        self.get_ifquery_all()
        lo = EtcNetworkInterface('lo', self.ifquery_all)
        lo.write_file()
        '''

    def init_working_dir(self):

        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir, mode=0600)

    def get_managed_files(self):
        files = []
        files.append(ETC_NETWORK_INTERFACES)
        return files

    def get_service_cmd_for_file(self, filename):
        cmds = []

        if filename == ETC_NETWORK_INTERFACES:
            cmds.append(('ifreload', '-a'))

        return cmds

    def get_ifquery_all(self):
        if self.ifquery_all:
            if self.etc_network_interfaces_changed():
                return
        else:
            MAX_ATTEMPTS = 6
            SLEEP_AMOUNT = 10

            for x in range(MAX_ATTEMPTS):
                try:
                    self.ifquery_all = check_output(['/sbin/ifquery', '-a'], stderr=STDOUT)
                    self.etc_network_interfaces_init_mod_time = os.path.getmtime(ETC_NETWORK_INTERFACES)
                    break
                except Exception as e:
                    log.warning("Attempt %d/%d: 'ifquery -a' failed, will sleep %ds and try again" % (x+1, MAX_ATTEMPTS, SLEEP_AMOUNT))

                    # This is ugly but we are doing this because the nclu/time.py module means
                    # we cannot import the normal "time" module which contains the sleep function.
                    subprocess.call(['/bin/sleep', str(SLEEP_AMOUNT)])
            else:
                log.warning("'ifquery -a' failed due to '%s'" % e)
                sys.exit(1)

    def expand_ifname(self, ifnames):
        ifnames_expanded = []

        if isinstance(ifnames, str) or isinstance(ifnames, unicode):
            ifnames = [str(ifnames), ]

        for ifname in ifnames:
            if ifname_is_glob(str(ifname)):
                for x in ifname_expand_glob(str(ifname)):
                    ifnames_expanded.append(x)

            elif ifname_is_valid(str(ifname)) and not ifname_is_globish(str(ifname)):
                ifnames_expanded.append(ifname)

        return ifnames_expanded

    def add_interfaces(self, ifname):
        ifnames = self.expand_ifname(ifname)
        self.get_ifquery_all()

        for ifname in ifnames:

            # Quagga can create a static route with these interface names as
            # the nexthop.  Do not create these interfaces.
            if ifname in ('null0', 'Null0', 'blackhole', 'reject'):
                continue

            if not ifname_in_etc_network_interfaces(ifname):
                iface = EtcNetworkInterface(ifname, self.ifquery_all)
                iface.add_interface()

            # If the user is adding swp10.50 add swp10 (if needed).
            if '.' in ifname:
                base = ifname.split('.')[0]
                if re.match(r"""^(swp\d+)(?:s\d+)?$""", base) and not ifname_in_etc_network_interfaces(base):
                    iface = EtcNetworkInterface(base, self.ifquery_all)
                    iface.add_interface()

    def breakout_interface(self, args, ifname, first):
        breakout_ifname_regex = ifname + 's\d+'
        iface = EtcNetworkInterface(ifname, self.ifquery_all)
        ifaces= {}

        for line in self.ifquery_all.splitlines():
            line = standardize_indent(line)
            re_name = re.search("^auto\s+(\S+)\s*$", line)

            if re_name:
                tmp_ifname = re_name.group(1)
                ifaces[tmp_ifname] = EtcNetworkInterface(tmp_ifname, self.ifquery_all)

        if os.path.exists('/var/run/nclu/iface/'):
            for filename in os.listdir('/var/run/nclu/iface/'):
                if not filename.endswith('.original') and not filename.endswith('.del'):
                    if filename not in ifaces:
                        ifaces[filename] = EtcNetworkInterface(filename, self.ifquery_all)

        # delete breakout ports
        if args.get('disabled') or args.get('1x') or args.get('/4'):

            for (tmp_ifname, tmp_iface) in ifaces.iteritems():
                if re.match(breakout_ifname_regex, tmp_ifname):
                    if tmp_iface.exists():
                        tmp_iface.del_interface()

        if args.get('disabled'):
            if iface.exists():
                iface.del_interface()

        elif args.get('1x'):
            if not iface.exists():
                iface.add_interface()

        elif args.get('2x'):
            if iface.exists():
                iface.del_interface()

            # Create the two breakout ports
            for x in xrange(2):
                tmp_ifname = "%ss%d" % (ifname, x)
                tmp_iface = EtcNetworkInterface(tmp_ifname, self.ifquery_all)

                if not tmp_iface.exists():
                    tmp_iface.add_interface()

            # If we moved for four breakout ports to two, delete #s 3 and 4
            for x in xrange(2,4):
                tmp_ifname = "%ss%d" % (ifname, x)
                tmp_iface = EtcNetworkInterface(tmp_ifname, self.ifquery_all)

                if tmp_iface.exists():
                    tmp_iface.del_interface()

        elif args.get('4x'):
            if iface.exists():
                iface.del_interface()

            # Create the four breakout ports
            for x in xrange(4):
                tmp_ifname = "%ss%d" % (ifname, x)
                tmp_iface = EtcNetworkInterface(tmp_ifname, self.ifquery_all)

                if not tmp_iface.exists():
                    tmp_iface.add_interface()

        elif args.get('/4'):
            if not first:
                iface.del_interface()

    def breakout_interfaces(self, args):
        ifnames = self.expand_ifname(args.get('<interface>'))
        self.get_ifquery_all()
        first = True

        for ifname in ifnames:
            self.breakout_interface(args, ifname, first)
            first = False

    def abort_pending(self):
        if os.path.exists('/var/run/nclu/iface/'):
            for filename in os.listdir('/var/run/nclu/iface/'):
                full_filename = os.path.join('/var/run/nclu/iface/', filename)

                if os.path.isdir(full_filename):
                    rmtree(full_filename)
                else:
                    os.remove(full_filename)

        self.ifquery_all = None
        self.ifaces = {}
        self.etc_network_interfaces_init_mod_time = None

    def del_all(self, cli):
        self.get_ifquery_all()

        for ifname in get_interfaces():
            if ifname not in ('eth0', 'eth1', 'mgmt'):
                iface = EtcNetworkInterface(ifname, self.ifquery_all)

                if iface.exists():

                    # If this is a loopback or a physical interface, just delete the content
                    # and leave the auto/iface lines alone.  There are lots of services in
                    # linux that need a lo interface so we cannot delete it.  For the swp
                    # interfaces, we want to leave them in so the interface comes up after the
                    # net commit. This makes configuration easier because you can see who your
                    # LLDP peers are.
                    if iface.is_swpx() or iface.is_loopback():
                        iface.del_interface_content()
                    else:
                        iface.del_interface()

    def commit_pending(self, _verbose):
        """
        Make iface changes persistent by updating /etc/network/interfaces.
        Return a tuple like (bool, [str]).  The boolean indicates if the
        commit succeeded.  The list of strings indicates the files
        modified, if any.
        """

        commit_ok = True
        modified = []

        # Sanity check the configuration via "ifup --syntax-check".
        self._save_or_show_pending(persist=False, verbose=False)

        if os.path.exists('/var/run/nclu/iface/interfaces.tmp'):
            try:
                check_output(['/sbin/ifup',
                              '--syntax-check',
                              '--verbose',
                              '--all',
                              '--interfaces',
                              '/var/run/nclu/iface/interfaces.tmp'],
                              stderr=STDOUT)
                os.remove('/var/run/nclu/iface/interfaces.tmp')
            except CalledProcessError as e:
                print "\nERROR: We cannot commit the current pending changes due to the following errors\n" +\
                      "they would introduce in /etc/network/interfaces:\n\n%s\n\n" %\
                    '\n'.join(line for line in e.output.splitlines() if not line.startswith('info'))

                os.remove('/var/run/nclu/iface/interfaces.tmp')
                commit_ok = False
                return (commit_ok, modified)

        if os.path.exists(IFACE_DIR_BACKUP):
            shutil.rmtree(IFACE_DIR_BACKUP)

        if os.path.exists(ETC_NETWORK_INTERFACES_BACKUP):
            os.remove(ETC_NETWORK_INTERFACES_BACKUP)

        shutil.copyfile(ETC_NETWORK_INTERFACES, ETC_NETWORK_INTERFACES_BACKUP)
        shutil.copytree('/var/run/nclu/iface/', IFACE_DIR_BACKUP)

        commit_ok = self._save_or_show_pending(persist=True, verbose=_verbose)
        self.ifquery_all = None
        self.ifaces = {}
        self.etc_network_interfaces_init_mod_time = None

        # Return a list of the files modified.  iface can only modify /etc/network/interfaces.
        if commit_ok:
            if os.path.exists(IFACE_DIR_BACKUP):
                shutil.rmtree(IFACE_DIR_BACKUP)

            if os.path.exists(ETC_NETWORK_INTERFACES_BACKUP):
                os.remove(ETC_NETWORK_INTERFACES_BACKUP)

            return (commit_ok, [ETC_NETWORK_INTERFACES])
        else:
            # These files should always exist, but it can't hurt to check.

            if os.path.exists('/var/run/nclu/iface/'):
                shutil.rmtree('/var/run/nclu/iface/')

            if os.path.exists(IFACE_DIR_BACKUP):
                shutil.move(IFACE_DIR_BACKUP, '/var/run/nclu/iface/')

            if os.path.exists(ETC_NETWORK_INTERFACES_BACKUP):
                shutil.move(ETC_NETWORK_INTERFACES_BACKUP, ETC_NETWORK_INTERFACES)

            return (commit_ok, [])

    def show_pending(self, print_errors=True):
        """
        print_errors will be False when the user does a 'net commit' because
        the commit will run --syntax-check.  We do not want to double print
        any errors.
        """
        diff_pending = PENDING_COMMIT_NOCHANGE
        self._save_or_show_pending(persist=False, verbose=False)
        if print_diff(ETC_NETWORK_INTERFACES, '/var/run/nclu/iface/interfaces.tmp', self.color_diffs):
            diff_pending = PENDING_COMMIT_CHANGE

        if os.path.exists('/var/run/nclu/iface/interfaces.tmp'):

            if print_errors:
                try:
                    check_output(['/sbin/ifup',
                                  '--syntax-check',
                                  '--verbose',
                                  '--all',
                                  '--interfaces',
                                  '/var/run/nclu/iface/interfaces.tmp'],
                                  stderr=STDOUT)
                except CalledProcessError as e:
                    print "\nERROR: pending changes will introduce the following errors in /etc/network/interfaces\n\n%s\n\n" % \
                        '\n'.join(line for line in e.output.splitlines() if not line.startswith('info'))
                    diff_pending = PENDING_COMMIT_ERROR

            os.remove('/var/run/nclu/iface/interfaces.tmp')

        return diff_pending

    def show_config_files(self, user_may_edit):
        print "/etc/network/interfaces"
        print "======================="

        if os.path.exists(ETC_NETWORK_INTERFACES):
            with open(ETC_NETWORK_INTERFACES) as f:
                print f.read() + '\n'

    def show_config_commands(self, user_may_edit):
        commands = []
        ifaces = OrderedDict()

        # Set self.ifquery_all to the output of "ifquery -a".
        self.get_ifquery_all()

        name_re = re.compile(r"""^auto\s+(\S+)""")
        vlan_re = re.compile(r"""vlan-?(\d+)""")

        for line in self.ifquery_all.splitlines():
            line = standardize_indent(line)
            name_match = name_re.search(line)

            # Only consider the first line of an interface stanza.
            if name_match is None:
                continue

            ifname = name_match.group(1)
            assert not ifname.isdigit(), "Digits are not valid interface names (ifname={0}).".format(ifname)
            iface = EtcNetworkInterface(ifname, self.ifquery_all)

            if ifname not in ifaces:
                ifaces[ifname] = []

                if (not iface.content or
                    (len(iface.content) == 2 and
                     iface.content[0].startswith('auto ') and
                     'inet dhcp' not in iface.content[1] and
                     'inet6 dhcp' not in iface.content[1])):

                    # The interface does not have any content whose 'net add' will
                    # auto create the interface.  Add a 'net add'.

                    if ifname == 'bridge':
                        commands.append('net add bridge')

                    elif ifname == 'lo' or iface.is_loopback():
                        commands.append('net add loopback {0}'.format(ifname))

                    elif iface.is_bond():
                        commands.append('net add bond {0}'.format(ifname))

                    elif iface.is_vxlan():
                        commands.append('net add vxlan {0}'.format(ifname))

                    elif iface.is_vlan_interface():
                        vlan_match = vlan_re.search(ifname)

                        if vlan_match:
                            commands.append('net add vlan {0}'.format(vlan_match.group(1)))
                        else:
                            commands.append('net add interface {0}'.format(ifname))

                    elif iface.is_vrf():
                        commands.append('net add vrf {0}'.format(ifname))

                    else:
                        commands.append('net add interface {0}'.format(ifname))

            # Parse the lines of the interface's stanza.
            for command in iface.content:
                foo = command.strip().split()
                key = foo[0]
                value = foo[1:]

                # Special case:  We translate "iface eth0 inet dhcp" to "net add interface eth0 address dhcp".
                if ' '.join(value).endswith('inet dhcp'):
                    key = 'ip address'
                    value = ['dhcp', ]
                elif ' '.join(value).endswith('inet6 dhcp'):
                    key = 'ipv6 address'
                    value = ['dhcp', ]

                # Only consider stanza commands.
                if key in ('auto', 'iface'):
                    continue

                value = ' '.join(value)

                if is_number_glob(value):
                    value = numbers_to_glob(value.split())

                if key in ('bridge-ports', 'bond-slaves', 'bridge-vids'):
                    value = ','.join(value.split())

                elif key in ('post-up', 'post-down', 'up', 'down'):
                    value = "'" + value + "'"

                # If the user wrote the default for a keyword in /etc/network/interfaces,
                # skip it. We do this because we remove the default values from the parse
                # tree so the 'net add' for a keyword with its default value will fail.
                if key in keyword_default and keyword_default[key] == value:  # keyword_default is a global dictionary
                    continue

                # babelfish for 'net show config commands'
                if key == 'address':
                    if value != 'dhcp':
                        try:
                            IPv4Network(value)
                            key = 'ip address'
                        except AddressValueError:
                            key = 'ipv6 address'

                elif key == 'address-virtual':
                    key = 'ip address-virtual'

                elif key == 'gateway':
                    try:
                        IPv4Network(value)
                        key = 'ip gateway'
                    except AddressValueError:
                        key = 'ipv6 gateway'

                elif key == 'netmask':
                    try:
                        IPv4Network(value)
                        key = 'ip netmask'
                    except AddressValueError:
                        key = 'ipv6 netmask'

                # link-autoneg, link-duplex, link-fec and link-speed
                elif key.startswith('link-'):
                    key = key.replace('link-', 'link ')

                elif key.startswith('bond-'):
                    key = key.replace('bond-', 'bond ')

                elif key == 'bridge-stp':
                    key = 'stp'

                elif key.startswith('bridge-'):
                    key = key.replace('bridge-', 'bridge ')

                elif key.startswith('mstpctl-'):
                    key = key.replace('mstpctl-', 'stp ')

                elif key.startswith('clagd-'):
                    key = key.replace('clagd-', 'clag ')

                elif key.startswith('clag-'):
                    key = key.replace('clag-', 'clag ')

                elif key.startswith('vxlan-'):
                    key = key.replace('vxlan-', 'vxlan ')

                if ifname == 'bridge':
                    ifaces[ifname].append('net add bridge {0} {1}'.format(key, value))

                elif ifname == 'lo' or iface.is_loopback():
                    ifaces[ifname].append('net add loopback {0} {1} {2}'.format(ifname, key, value))

                elif iface.is_bond():
                    ifaces[ifname].append('net add bond {0} {1} {2}'.format(ifname, key, value))

                elif iface.is_vxlan():
                    ifaces[ifname].append('net add vxlan {0} {1} {2}'.format(ifname, key, value))

                elif iface.is_vlan_interface():
                    vlan_match = vlan_re.search(ifname)

                    if vlan_match:
                        ifaces[ifname].append('net add vlan {0} {1} {2}'.format(vlan_match.group(1), key, value))
                    else:
                        ifaces[ifname].append('net add interface {0} {1} {2}'.format(ifname, key, value))

                elif iface.is_vrf():
                    ifaces[ifname].append('net add vrf {0} {1} {2}'.format(ifname, key, value))

                else:
                    ifaces[ifname].append('net add interface {0} {1} {2}'.format(ifname, key, value))

                cmd_added = ifaces[ifname][-1]

                # There can be multiple IPs listed for address-virtual, but
                # these were added via multiple 'net add' commands.
                if 'address-virtual' in cmd_added:
                    up_to_mac_v4 = ' '.join(cmd_added.split()[0:7]).replace('ipv6 address-virtual', 'ip address-virtual')
                    up_to_mac_v6 = up_to_mac_v4.replace('ip address-virtual', 'ipv6 address-virtual')
                    ips = cmd_added.split()[7:]
                    ifaces[ifname].remove(cmd_added)

                    for ip in ips:
                        try:
                            IPv4Network(ip)
                            ifaces[ifname].append("{0} {1}".format(up_to_mac_v4, ip))
                        except AddressValueError:
                            ifaces[ifname].append("{0} {1}".format(up_to_mac_v6, ip))

        for ifname in sort_ifaces(ifaces.keys()):
            for command in ifaces[ifname]:
                commands.append(command)

        commands = globify_interface_commands(commands)

        # Create a temporary NetworkDocopt object so we can see
        # which 'commands' are supported. We need to replace the
        # 'iface' in our __doc__ with 'net'
        cli = NetworkDocopt(__doc__.replace('    iface ', '    net '))
        supported = []
        not_supported = []

        for cmd in commands:
            # We do not want the user to see the "Command not found"
            # error if this is an unsupported command so use Capturing
            # but do nothing with the output.
            with Capturing():
                cli.run(shlex.split(cmd))

            if cli.match:
                supported.append(cmd)
            else:
                # For keywords like "link-down" where we automagically add the "yes"
                # we do not include the obvious "yes" in the parse tree so attempt
                # the command without that yes and see if it works
                if cmd.endswith(' yes'):
                    cmd = cmd[0:-4]

                    with Capturing():
                        cli.run(shlex.split(cmd))

                    if cli.match:
                        supported.append(cmd)
                    else:
                        not_supported.append(cmd)
                else:
                    not_supported.append(cmd)
            cli.re_init()

        return (supported, not_supported)

    def show_config_summary(self, user_may_edit, summary):
        self.get_ifquery_all()

        name_re = re.compile("^auto\s+(\S+)\s*$")

        for line in self.ifquery_all.splitlines():
            line = standardize_indent(line)
            name_match = name_re.search(line)

            if name_match:
                ifname = name_match.group(1)
                iface = EtcNetworkInterface(ifname, self.ifquery_all)
                ifname = "INTERFACE:%s" % ifname

                if ifname not in summary:
                    summary[ifname] = OrderedDict()

                if 'COMMAND_LIST' not in summary[ifname]:
                    summary[ifname]['COMMAND_LIST'] = []

                for command in sorted(iface.content):
                    foo = command.strip().split()
                    key = foo[0]
                    value = foo[1:]

                    if ' '.join(value).endswith('inet dhcp'):
                        key = 'address'
                        value = ['dhcp']
                    elif ' '.join(value).endswith('inet6 dhcp'):
                        key = 'ipv6 address'
                        value = ['dhcp']

                    if key in ('auto', 'iface'):
                        continue

                    summary[ifname]['COMMAND_LIST'].append('{0} {1}'.format(key, ' '.join(value)))

    def bridge_del(self, bridge):

        if not bridge:
            return

        # Call vlan_del() for the vlan glob in the bridge's bridge-vids
        # This will remove all references to those VLANs
        for line in bridge.content:
            if line.strip().startswith('bridge-vids'):
                vlan_glob = line.strip().replace('bridge-vids ', '')
                vlan_glob = ','.join(vlan_glob.strip().split())
                self.vlan_del(bridge, vlan_glob, True)
                break

        bridge.del_interface()

    def vlan_add(self, bridge, vlan_glob):
        bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan_glob)

    def vlan_del(self, bridge, vlan_glob, nuke_l2=False):
        if not bridge:
            return

        bridge.conf_key_value_multiple_values(False, 'bridge-vids', vlan_glob, give_feedback=False)
        vlans_to_del = glob_to_numbers(str(vlan_glob))

        for vlan in vlans_to_del:
            ifname = "vlan%d" % vlan
            iface = EtcNetworkInterface(ifname, self.ifquery_all)
            self.svi_del(iface)

        # remove any access interface in these vlans
        # remove these vlans from the bridge-vids of all trunks
        for ifquery_line in self.ifquery_all.splitlines():
            ifquery_line = standardize_indent(ifquery_line)
            re_name = re.search("^auto\s+(\S+)\s*$", ifquery_line)

            if re_name and re_name.group(1) != 'bridge':
                ifname = re_name.group(1)
                iface = EtcNetworkInterface(ifname, self.ifquery_all)

                # See if this is an SVI named "bridge.10"...if so and if
                # vlan 10 is one of the vlans we are deleting then delete
                # this interface and 'continue'
                re_strict_svi_name = re.search('^bridge\.(\d+)$', ifname)
                if re_strict_svi_name:
                    vlan_id = int(re_strict_svi_name.group(1))

                    if vlan_id in vlans_to_del:
                        iface.del_interface()
                        continue

                # We deepcopy here because we could end up deleting an
                # entry from iface.content
                for line in deepcopy(iface.content):

                    # If this interface is in a vlan that is being deleted then:
                    # - remove the bridge-access line
                    # - remove this interface from the bridge's bridge-ports
                    if 'bridge-access' in line:
                        bridge_access_vlan = int(line.strip().split()[1])

                        if bridge_access_vlan in vlans_to_del:
                            iface.conf_key_value_solo(False, 'bridge-access', bridge_access_vlan)
                            bridge.conf_key_value_multiple_values(False, 'bridge-ports', ifname, give_feedback=False)

                    # If this is a trunk remove the vlans from the trunk's bridge-vids
                    elif 'bridge-vids' in line:
                        self.trunk_del(bridge, iface, vlan_glob)

                    # If this is a SVI for a vlan that is being deleted then delete the SVI
                    # Look for both vlan-id and vlan-raw-device to be 100% sure this is a SVI
                    elif 'vlan-id' in line:
                        vlan_id = int(line.strip().split()[1])

                        if vlan_id in vlans_to_del:
                            found_vlan_raw_device = False

                            for line2 in iface.content:
                                if 'vlan-raw-device' in line2:
                                    found_vlan_raw_device = True
                                    break

                            if found_vlan_raw_device:
                                iface.del_interface()
                                break

                    # nuke_l2 will be true if the user did 'net del bridge' and
                    # this vlan was listed under the bridge's bridge-vids
                    #
                    # Delete all of the bridge-* and mstpctl-* commands
                    elif nuke_l2 and (line.lstrip().startswith('bridge-') or line.lstrip().startswith('mstpctl-')):
                        foo = line.strip().split()
                        key = foo[0]
                        value = ' '.join(foo[1:])
                        iface.conf_key_value_solo(False, key, value)

    def access_add(self, bridge, iface, vlan):

        # If this interface is currently configured as a trunk, remove that configuration
        if iface.has_cmd('bridge-vids') or iface.has_cmd('bridge-pvid'):
            self.trunk_del(bridge, iface, None)

        # add this interface to bridge bridge-ports
        bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)

        # add this vlan to bridge bridge-vids
        bridge.conf_key_value_multiple_values(True, 'bridge-vids', str(vlan), give_feedback=False)

        # add bridge-access for this interface
        iface.conf_key_value_solo(True, 'bridge-access', vlan)

    def access_del(self, bridge, iface, vlan):

        # Remove this interface from bridge bridge-ports
        # We must do this else the interface would instantly become a trunk
        if ifname_in_etc_network_interfaces('bridge'):
            bridge = EtcNetworkInterface('bridge', self.ifquery_all)
            bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name)

        # remove the bridge-access line for this interface
        iface.conf_key_value_solo(False, 'bridge-access', vlan)

    def vrf_add(self, iface):
        iface.conf_key_value_solo(True, 'vrf-table', 'auto', give_feedback=False)

    def vrf_del(self, iface):
        iface.del_interface()

        # If any interfaces are configured to be in this vrf we must move
        # them out of this vrf and back to the default table
        for ifquery_line in self.ifquery_all.splitlines():
            ifquery_line = standardize_indent(ifquery_line)
            re_name = re.search("^auto\s+(\S+)\s*$", ifquery_line)

            if re_name:
                ifname = re_name.group(1)
                iface_tmp = EtcNetworkInterface(ifname, self.ifquery_all)
                iface_tmp_vrf = iface_tmp.get_keyword_value('vrf')

                if iface_tmp_vrf and iface_tmp_vrf == iface.name:
                    iface_tmp.del_cmd('    vrf %s' % iface.name, True, True)
                    iface_tmp.write_file()

    def trunk_add(self, bridge, iface, vlan_glob):

        # If this interface is currently configured as 'access', remove that configuration
        if iface.has_cmd('bridge-access'):
            self.access_del(bridge, iface, None)

        # add this interface to bridge bridge-ports
        bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)

        if vlan_glob:
            # add these vlans to the trunk bridge-vids
            iface.conf_key_value_multiple_values(True, 'bridge-vids', vlan_glob)

            # add these vlans to bridge bridge-vids
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', vlan_glob, give_feedback=False)

    def trunk_del(self, bridge, iface, vlan_glob):
        deconfigure_trunk = False

        # We are removing specific vlans from this trunk
        if vlan_glob:
            initial_bridge_vids = iface.get_keyword_value('bridge-vids')
            iface.conf_key_value_multiple_values(False, 'bridge-vids', vlan_glob)
            final_bridge_vids = iface.get_keyword_value('bridge-vids')

            # If we have removed the last bridge-vid then this
            # interface should no longer be a trunk
            if initial_bridge_vids and not final_bridge_vids:
                deconfigure_trunk = True

        # If the user did a del without specifying any vlans
        # then this interface should no longer be a trunk
        else:
            deconfigure_trunk = True

        if deconfigure_trunk:
            # Remove this interface from bridge bridge-ports
            if ifname_in_etc_network_interfaces('bridge'):
                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name)

            if iface.has_cmd('bridge-vids'):
                bridge_vids = iface.get_keyword_value('bridge-vids')
                iface.conf_key_value_multiple_values(False, 'bridge-vids', bridge_vids)

            if iface.has_cmd('bridge-pvid'):
                iface.conf_key_value_solo(False, 'bridge-pvid', None)

    def svi_add(self, bridge, iface):
        # Note that these regexes cannot both match.
        re_bridge_sub_int = re.match(r"""^bridge\.(?P<sub_int>\d+)$""", iface.name, re.I)
        re_vlan_number = re.match(r"""^vlan-?(?P<vlan_int>\d+)$""", iface.name, re.I)

        if (re_bridge_sub_int is not None or re_vlan_number is not None) and not iface.exists():
            iface.add_interface()

        if re_bridge_sub_int is not None:
            sub_int = int(re_bridge_sub_int.group('sub_int'))
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', sub_int, give_feedback=True)
            return True

        elif re_vlan_number is not None:
            vlan_int = int(re_vlan_number.group('vlan_int'))
            iface.conf_key_value_solo(True, 'vlan-raw-device', 'bridge', give_feedback=False)
            iface.conf_key_value_solo(True, 'vlan-id', vlan_int, give_feedback=False)
            bridge.conf_key_value_multiple_values(True, 'bridge-vids', str(vlan_int), give_feedback=False)
            return True

        else:
            # This is the case when neither regular expression matches.  It indicates a programming error.
            log.warning("svi_add() should not be here.")
            return False

    def svi_del(self, iface):
        if iface.exists():
            iface.del_interface()

    def add_svi_interfaces(self, svi_ifnames):
        self.get_ifquery_all()
        bridge = EtcNetworkInterface('bridge', self.ifquery_all)

        if not bridge.exists():
            bridge.add_interface()
            bridge.conf_key_value_solo(True, 'bridge-vlan-aware', 'yes', give_feedback=False)

        for ifname in svi_ifnames:
            iface = EtcNetworkInterface(ifname, self.ifquery_all)

            if not iface.exists():
                self.svi_add(bridge, iface)

    def vxlan_add(self, bridge, iface, vxlan_id):
        re_vxlan_number = re.match(r"""^vxlan-?(?P<vxlan_id>\d+)$""", iface.name, re.I)
        re_vni_number = re.match(r"""^vni-?(?P<vxlan_id>\d+)$""", iface.name, re.I)

        if vxlan_id:

            if iface.exists():
                if not iface.is_vxlan():
                    print "ERROR: {0} is not a vxlan.".format(iface.name)
                    return False
            else:
                iface.add_interface()

            iface.conf_key_value_solo(True, 'vxlan-id', vxlan_id, give_feedback=False)
            iface.conf_key_value_solo(True, 'mstpctl-bpduguard', 'yes', give_feedback=False)
            iface.conf_key_value_solo(True, 'mstpctl-portbpdufilter', 'yes', give_feedback=False)
            bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)
            return True

        elif re_vxlan_number is not None or re_vni_number is not None:

            if re_vxlan_number is not None:
                vxlan_id = int(re_vxlan_number.group('vxlan_id'))
            elif re_vni_number is not None:
                vxlan_id = int(re_vni_number.group('vxlan_id'))

            if not iface.exists():
                iface.add_interface()

            iface.conf_key_value_solo(True, 'vxlan-id', vxlan_id, give_feedback=False)
            iface.conf_key_value_solo(True, 'mstpctl-bpduguard', 'yes', give_feedback=False)
            iface.conf_key_value_solo(True, 'mstpctl-portbpdufilter', 'yes', give_feedback=False)
            bridge.conf_key_value_multiple_values(True, 'bridge-ports', iface.name, give_feedback=False)
            return True

        else:
            print "ERROR: A vxlan interface may be named 'vxlan10, vxlan-20, etc', or you can specify\n"\
                  "the vxlan-id via 'net add vxlan FOO vxlan id 10'.\n\n"
            return False

    def vxlan_del(self, bridge, iface, vlan):
        iface.del_interface()

        if ifname_in_etc_network_interfaces('bridge'):
            bridge = EtcNetworkInterface('bridge', self.ifquery_all)
            bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name, give_feedback=False)

    def keywords_configured(self, args):
        """
        Return True if a keyword under an interface is configured

        We use this to tell if the user did:
            net add int swp1
            net del int swp2
        """
        for keyword in all_keywords:
            if args.get(keyword):
                # log.info("keywords_configured %s is configured" % keyword)
                return True
        return False

    def get_iface(self, ifname):
        self.get_ifquery_all()

        if ifname in self.ifaces:
            iface = self.ifaces[ifname]
            iface.load_iface_state()
        else:
            self.ifaces[ifname] = EtcNetworkInterface(ifname, self.ifquery_all)
            iface = self.ifaces[ifname]

        return iface

    def eval_command_line_args(self, cli):
        """
        - args is a dictionary produced by NetworkDocopt based on command line input from the user
        - Create an EtcNetworkInterface object for the interface we wish to modify
        - Find the Command object for the command keyword specified by the user (address, bridge-ports, etc)
        - Call the appropriate conf_XXX() function for this Command
        """
        args = cli.args
        ifnames = []
        command_ok = True

        if args.get('add') or args.get('del'):
            self.get_ifquery_all()
            bridge = None
            vlan_glob = args.get('<number-range-list>') if args.get('<number-range-list>') is not None else args.get('<number-comma-range-list>')
            (ifnames_valid, ifnames) = get_ifnames(cli)

            if not ifnames_valid:
                return False

            # 'add', for all of these, requires us to create a vlan-aware bridge.
            # 'del' requires an EtcNetworkInterface object for the bridge (if it exists).
            if ((args.get('clag') and (args.get('peer') or args.get('host'))) or
                args.get('bridge') or
                args.get('vlan') or
                args.get('vxlan')):

                if args.get('add'):

                    if ifname_in_etc_network_interfaces('bridge'):
                        bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                    else:
                        # Create an interface called "bridge".
                        bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                        bridge.add_interface()

                    # Make the bridge interface vlan aware.
                    bridge.conf_key_value_solo(True, 'bridge-vlan-aware', 'yes', give_feedback=False)

                elif args.get('del'):
                    if ifname_in_etc_network_interfaces('bridge'):
                        bridge = EtcNetworkInterface('bridge', self.ifquery_all)

            # iface (add|del) vlan <number-range-list>
            if args.get('vlan') and vlan_glob:
                if args.get('add'):
                    self.vlan_add(bridge, vlan_glob)
                elif args.get('del'):
                    if len(cli.argv_expanded) == 4:
                        self.vlan_del(bridge, vlan_glob)

            argv_expanded_string = ' '.join(map(str, cli.argv_expanded))

            if (argv_expanded_string.startswith('net add interface eth0 vrf') and
                argv_expanded_string != 'net add interface eth0 vrf mgmt'):
                print "ERROR: only 'vrf mgmt' is allowed for eth0"
                return False


            mgmt_vrf_msg = """
************************************************************************************
NOTE: Enabling or disabling Management VRF will cause all SSH sessions to disconnect
on the next 'net commit'. This only happens the first time you do a 'net commit'
after enabling or disabling Management VRF.
***********************************************************************************
"""
            # Macros for Management VRF
            if argv_expanded_string == 'net add vrf mgmt' or argv_expanded_string == 'net add interface eth0 vrf mgmt':

                # 'net add vrf mgmt' is a macro that does
                # - net add vrf mgmt vrf-table auto
                # - net add vrf mgmt address 127.0.0.1/8
                # - net add int eth0 vrf mgmt
                iface = EtcNetworkInterface('mgmt', self.ifquery_all)

                if not iface.exists():
                    iface.add_interface()

                iface.conf_key_value_solo(args.get('add'), 'vrf-table', 'auto', args)
                iface.conf_key_value_multiple_lines(args.get('add'), 'address', '127.0.0.1/8', args)

                # Only auto add the eth0 'vrf mgmt' line if the user did 'net add vrf mgmt'.
                # If we were to add it here then later (when the normal code path for a 'net add'
                # runs) we would print a "eth0's configuration already has 'vrf mgmt'" message
                # which would be confusing for the user.
                if argv_expanded_string == 'net add vrf mgmt':
                    eth0 = EtcNetworkInterface('eth0', self.ifquery_all)

                    if not eth0.exists():
                        eth0.add_interface()
                        eth0.conf_key_value_multiple_lines(args.get('add'), 'address', 'dhcp', args)
                    eth0.conf_key_value_solo(args.get('add'), 'vrf', 'mgmt', args)

                print mgmt_vrf_msg

            elif argv_expanded_string == 'net del vrf mgmt' or argv_expanded_string == 'net del interface eth0 vrf mgmt':
                iface = EtcNetworkInterface('mgmt', self.ifquery_all)
                if iface.exists():
                    iface.del_interface()

                if argv_expanded_string == 'net del vrf mgmt':
                    eth0 = EtcNetworkInterface('eth0', self.ifquery_all)

                    if eth0.exists():
                        eth0.conf_key_value_solo(False, 'vrf', 'mgmt', args)

                print mgmt_vrf_msg

            # Macros for clag configuration
            if argv_expanded_string.startswith('net add clag peer'):
                bondname = 'peerlink'
                interfaces = args.get('<interface>')
                bondname_sub = bondname + '.4094'

                if 'peerlink' in interfaces:
                    print "ERROR: You must specify the interfaces that will go in the 'peerlink' bond to our fellow clag switch"
                    return False

                bond = EtcNetworkInterface(bondname, self.ifquery_all)
                if not bond.exists():
                    bond.add_interface()
                ifaces_to_add = bond.conf_key_value_multiple_values(True, 'bond-slaves', interfaces)

                if ifaces_to_add:
                    self.add_interfaces(ifaces_to_add)

                # Add the peerlink bond to the bridge.
                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                bridge.conf_key_value_multiple_values(True, 'bridge-ports', bondname)

                bondsub = EtcNetworkInterface(bondname_sub, self.ifquery_all)
                if not bondsub.exists():
                    bondsub.add_interface()

                if args.get('primary'):
                    bondsub.conf_key_value_multiple_lines(True, 'address', '169.254.1.1/30', args)
                    bondsub.conf_key_value_solo(True, 'clagd-peer-ip', '169.254.1.2')
                    bondsub.conf_key_value_solo(True, 'clagd-priority', '1000')
                elif args.get('secondary'):
                    bondsub.conf_key_value_multiple_lines(True, 'address', '169.254.1.2/30', args)
                    bondsub.conf_key_value_solo(True, 'clagd-peer-ip', '169.254.1.1')
                    bondsub.conf_key_value_solo(True, 'clagd-priority', '2000')

                bondsub.conf_key_value_solo(True, 'clagd-sys-mac', args.get('<mac-clag>'))

                if args.get('backup-ip'):
                    if args.get('vrf'):
                        backup_ip = "%s vrf %s" % (args.get('<ipv4>'), args.get('<text>'))
                    else:
                        backup_ip = args.get('<ipv4>')

                    bondsub.conf_key_value_solo(True, 'clagd-backup-ip', backup_ip)

                return True

            # iface del clag peer [bond <interface>]
            elif argv_expanded_string.startswith('net del clag peer'):
                bondname = 'peerlink'
                bondname_sub = bondname + '.4094'

                bond = EtcNetworkInterface(bondname, self.ifquery_all)
                bondsub = EtcNetworkInterface(bondname_sub, self.ifquery_all)

                if bond.exists():
                   bond.del_interface()

                if bondsub.exists():
                   bondsub.del_interface()

                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                bridge.conf_key_value_multiple_values(False, 'bridge-ports', bondname)
                return True

            # iface add clag port bond <interface> interface <interface> clag-id <0-65535>
            elif argv_expanded_string.startswith('net add clag port'):
                (bondname, interfaces) = args.get('<interface>')
                bond = EtcNetworkInterface(bondname, self.ifquery_all)
                if not bond.exists():
                    if not ifname_is_valid(bond.name):
                        return False
                    bond.add_interface()
                ifaces_to_add = bond.conf_key_value_multiple_values(True, 'bond-slaves', interfaces)
                bond.conf_key_value_solo(True, 'clag-id', args.get('<0-65535>'))

                if ifaces_to_add:
                    self.add_interfaces(ifaces_to_add)

                # add the host facing bond to the bridge
                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                if not bridge.exists():
                    bridge.add_interface()
                    bridge.conf_key_value_solo(True, 'bridge-vlan-aware', 'yes', give_feedback=False)
                bridge.conf_key_value_multiple_values(True, 'bridge-ports', bondname)

                return True

            # iface del clag port bond <interface>
            elif argv_expanded_string.startswith('net del clag port'):
                bondname = args.get('<interface>')
                bond = EtcNetworkInterface(bondname, self.ifquery_all)

                if bond.exists():
                    bond.del_interface()

                # del the host facing bond from the bridge
                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                bridge.conf_key_value_multiple_values(False, 'bridge-ports', bondname)
                return True

            for ifname in ifnames:
                iface = EtcNetworkInterface(ifname, self.ifquery_all)

                if args.get('add'):
                    if ((ifname.startswith('vlan') or ifname.startswith('bridge.')) and not iface.exists()):
                        if not self.svi_add(bridge, iface):
                            continue

                    if args.get('vxlan') and not iface.exists():
                        vxlan_id = args.get('<1-16777214>')

                        if not self.vxlan_add(bridge, iface, vxlan_id) or vxlan_id:
                            continue

                    if args.get('bond'):
                        if args.get('slaves') or iface.is_bond():
                            pass
                        else:
                            print "ERROR: %s is not a bond" % ifname
                            command_ok = False
                            continue

                    if ifname.startswith('lo:'):
                        print "ERROR: The loopback interface must be named 'lo', please use 'net add loopback lo'"
                        command_ok = False
                        continue

                    if args.get('clag'):
                        if args.get('sys-mac'):
                            if not re.search('\.\d+$', ifname):
                                print "ERROR: 'clag sys-mac' can only be configured on a sub-interface such as swp1.4094"
                                command_ok = False
                                continue
                        elif args.get('peer-ip'):
                            if not re.search('\.\d+$', ifname):
                                print "ERROR: 'clag peer-ip' can only be configured on a sub-interface such as swp1.4094"
                                command_ok = False
                                continue

                # If we are doing an "add" to an interface that does not exist, add it automatically.
                if args.get('add') and not iface.exists():

                    if not iface.add_interface():
                        continue

                # The bridge-trunk, vlan, vxlan, etc keywords are special, these
                # are not from an ifupdown2 addon.  They are keywords that we
                # explicitly added in the __doc__ so we must handle these here and not
                # in the for loop for the addons.
                #
                # bridge-access is from an ifupdown2 addon but we want to overwrite it and
                # treat it as a macro so we handle it here instead of via the addon for loop.
                #
                # iface add interface <interface> bridge access <1-4094>
                if args.get('bridge') and args.get('access'):
                    vlan = args.get('<1-4094>')

                    if args.get('add'):
                        self.access_add(bridge, iface, vlan)
                    elif args.get('del'):
                        self.access_del(bridge, iface, vlan)

                #
                # iface (add|del) interface <interface> bond-master <interface-bond>
                #
                # elif args.get('bond-master'):
                #    bond = EtcNetworkInterface(args.get('<interface-bond>'), self.ifquery_all)
                #
                #    if args.get('add'):
                #        if not bond.exists():
                #            bond.add_interface()
                #        bond.conf_key_value_multiple_values(True, 'bond-slaves', ifname)
                #
                #    elif args.get('del'):
                #        if bond.exists():
                #            bond.conf_key_value_multiple_values(False, 'bond-slaves', ifname)
                #        else:
                #            print "ERROR: bond %s does not exist" % bond

                # The user can configure a trunk via bridge-trunk or bridge-vids.
                # bridge-vids can be configured on the bridge interface itself though
                # which is used to add support for a vlan, not configure a trunk.  Check
                # to see if bridge-vids is being configured but not on the 'bridge' interface
                #
                # iface (add|del) interface <interface> bridge-trunk [vlans <number-range-list>]
                elif (args.get('interface') or args.get('bond')) and args.get('bridge') and (args.get('trunk') or args.get('vids')):
                    if args.get('add'):
                        self.trunk_add(bridge, iface, vlan_glob)
                    elif args.get('del'):
                        self.trunk_del(bridge, iface, vlan_glob)

                # iface (add|del) vxlan <interface> vlan <1-4094>
                elif args.get('vxlan') and not self.keywords_configured(args):
                    vlan = args.get('<1-4094>')

                    if args.get('add'):
                        pass
                    elif args.get('del'):
                        self.vxlan_del(bridge, iface, vlan)

                # iface del bridge
                elif args.get('del') and args.get('bridge') and len(cli.argv_expanded) == 3:
                    self.bridge_del(bridge)

                # net add vlan mtu
                elif args.get('add') and args.get('vlan') and args.get('mtu'):
                    vlan_mtu = args.get('<552-9216>')
                    iface.conf_key_value_solo(True, 'mtu', vlan_mtu)

                    # get a list of bridge-ports
                    bridge_ports = bridge.get_keyword_value('bridge-ports')

                    if bridge_ports is not None:
                        for bridge_port in bridge_ports.split():
                            bridge_port_iface = EtcNetworkInterface(bridge_port, self.ifquery_all)
                            bridge_port_iface_mtu = bridge_port_iface.get_keyword_value('mtu')

                            if not bridge_port_iface.is_vxlan() and (bridge_port_iface_mtu is None or int(bridge_port_iface_mtu) < int(vlan_mtu)):
                                bridge_port_iface.conf_key_value_solo(True, 'mtu', vlan_mtu)


                # word-around for the word 'vrf'.  vrf is an interface type
                #   iface (add|del) vrf BLUE
                #
                # And it can be used to put an interface in a vrf
                #   iface (add)del int swp1 vrf BLUE
                #
                # If the length is 4 we know it is 'iface (add|del) vrf BLUE' so handle
                # that here as a one of
                elif len(cli.argv_expanded) == 4 and cli.argv_expanded[2] == 'vrf':
                    if cli.argv_expanded[1] == 'add':
                        self.vrf_add(iface)
                    elif cli.argv_expanded[1] == 'del':
                        self.vrf_del(iface)

                elif args.get('del') and args.get('address'):

                    if args.get('<ipv4/prefixlen>'):
                        iface.conf_key_value_multiple_lines(False, 'address', args.get('<ipv4/prefixlen>'), args)
                    elif args.get('<ipv6/prefixlen>'):
                        iface.conf_key_value_multiple_lines(False, 'address', args.get('<ipv6/prefixlen>'), args)
                    elif (args.get('ip') or args.get('ipv6')) and args.get('dhcp') and args.get('interface'):
                        # net del interface <interface> (ip|ipv6) address dhcp
                        assert re.match(r"""net del interface \S+ ip(v6)? address dhcp""", argv_expanded_string) is not None, \
                            "Unexpected command for this block of code."

                        # The line to modify /etc/network/interfaces looks like "iface <interface> inet(6) dhcp".
                        # The change is to remove the "inet(6) dhcp" ending.
                        old_cmd = 'iface {0} inet{1} dhcp'.format(iface.name, '6' if (' ipv6 ' in argv_expanded_string) else '')
                        new_cmd = 'iface {0}'.format(iface.name)

                        if not iface.mod_cmd(old_cmd, new_cmd, True):
                            # Couldn't modify the appropriate line in /etc/network/interfaces.
                            return False

                        # Successfully deleted the configuration command.  Persist the change.
                        iface.write_file()
                    else:
                        for ip_address in iface.get_keyword_value_list('address'):
                            ip_address = IPNetwork(ip_address)

                            if args.get('ip'):
                                if isinstance(ip_address, IPv4Network):
                                    iface.conf_key_value_multiple_lines(False, 'address', ip_address, args)
                            elif args.get('ipv6'):
                                if isinstance(ip_address, IPv6Network):
                                    iface.conf_key_value_multiple_lines(False, 'address', ip_address, args)

                # A "normal" (non macro) command that comes from an ifupdown2 addon
                else:

                    # Find the Command object that matches
                    for cmd in addons_cmds:
                        if not cmd.deprecated and cmd.keywords_in_args(args):
                            ifaces_to_add = []

                            # Another work-around for the 'vrf' keyword collision
                            if cmd.keyword == 'vrf' and len(cli.argv_expanded) >= 3 and cli.argv_expanded[2] == 'vrf':
                                continue

                            # bridge-stp is split into 'bridge' 'stp' while mstpctl-bpduguard
                            # is split into 'stp' 'bpduguard' so if we only look for 'stp' we
                            # will think the user configured 'bridge stp' when they may have
                            # configured 'stp bpduguard'
                            if 'bridge' in cmd.keyword and 'stp' in cmd.keyword and 'bridge stp off' not in argv_expanded_string:
                                continue

                            # clag-id and vxlan-id will both have a keyord of "id" (due to
                            # splitting on the hyphen) so we must do a special check.
                            if cmd.keyword == 'id':
                                if cmd.keyword_original == 'clag-id':
                                    if not args.get('clag'):
                                        continue

                                elif cmd.keyword_original == 'vxlan-id':
                                    if not args.get('vxlan'):
                                        continue

                            # If this is the first time we've configured a particular
                            # vrf, we must add the 'vrf-table auto' line.
                            if args.get('add') and not args.get('vrf-table'):

                                # net add vrf BLUE
                                if cli.argv_expanded[2] == 'vrf':
                                    self.vrf_add(iface)

                                # net add int swp1 vrf BLUE
                                elif len(cli.argv_expanded) == 6 and cli.argv_expanded[4] == 'vrf':
                                    vrf_name = args.get('<text>')
                                    vrf_iface = EtcNetworkInterface(vrf_name, self.ifquery_all)

                                    if not vrf_iface.exists():
                                        vrf_iface.add_interface()
                                        vrf_iface.conf_key_value_solo(True, 'vrf-table', 'auto', give_feedback=False)

                            if args.get('add') and not iface.exists():
                                # We are adding or modifying an interface that does not exist, so add it.
                                iface.add_interface()

                            if isinstance(cmd, CommandAddDelMultiLine):
                                iface.conf_key_value_multiple_lines(args.get('add'),
                                                                    cmd.keyword_original,
                                                                    cmd.value_from_args(args),
                                                                    args)

                            elif isinstance(cmd, CommandAddDelMultiValue):

                                # address-virtual is special, there can be multiple lines of address-virtual
                                # and each of those lines is multivalue (there can be multiple IPs for the
                                # virtual MAC)
                                if cmd.keyword_original == 'address-virtual':
                                    custom_keyword = "address-virtual %s" % args.get('<mac>')

                                    if iface.get_cmd(custom_keyword):
                                        use_multi_value = True
                                    else:
                                        use_multi_value = False

                                    if use_multi_value:
                                        ifaces_to_add = iface.conf_key_value_multiple_values(args.get('add'),
                                                                                             custom_keyword,
                                                                                             cmd.value_from_args(args))
                                    else:
                                        iface.conf_key_value_multiple_lines(args.get('add'),
                                                                            cmd.keyword_original,
                                                                            cmd.value_from_args(args),
                                                                            args)
                                else:
                                    ifaces_to_add = iface.conf_key_value_multiple_values(args.get('add'),
                                                                                         cmd.keyword_original,
                                                                                         cmd.value_from_args(args))

                            elif isinstance(cmd, CommandSetClear):
                                iface.conf_key_value_solo(args.get('add'),
                                                          cmd.keyword_original,
                                                          cmd.value_from_args(args))

                            else:
                                raise Exception("command %s is an invalid type %s" % (cmd, type(cmd)))

                            # Automagically add the interfaces listed under bridge-ports, bond-slaves, etc
                            self.add_interfaces(ifaces_to_add)
                            break
                    else:
                        # If we are here the user did one of the following
                        #  net add int swp20
                        #  net del int swp20
                        if args.get('add'):
                            if not iface.exists():
                                iface.add_interface()
                        elif args.get('del'):
                            if ifname == 'lo':
                                print "Deleting the loopback interface is not supported"
                                command_ok = False
                            elif ifname == 'eth0' or ifname == 'eth1':
                                print "Deleting the %s interface is not supported" % ifname
                                command_ok = False
                            # This was handled earlier via vlan_del()
                            elif args.get('vlan'):
                                pass
                            else:
                                iface.del_interface()

                                # Remove this interface from the bridge's bridge-ports
                                bridge = EtcNetworkInterface('bridge', self.ifquery_all)
                                if bridge.exists():
                                    bridge.conf_key_value_multiple_values(False, 'bridge-ports', iface.name, give_feedback=False)

                                # If there are any sub-interfaces for the interface we are deleting
                                # we must delete them too
                                for tmp_ifname in get_interfaces():
                                    if '.' in tmp_ifname:
                                        base  = '.'.join(tmp_ifname.split('.')[0:-1])
                                        extension = tmp_ifname.split('.')[-1]

                                        if base == iface.name:
                                            tmp_iface = EtcNetworkInterface(tmp_ifname, self.ifquery_all)

                                            if tmp_iface.exists():
                                                if bridge.exists():
                                                    bridge.conf_key_value_multiple_values(False, 'bridge-ports', tmp_ifname, give_feedback=False)
                                                tmp_iface.del_interface()

        elif args.get('show'):
            if args.get('clag'):

                if args.get('macs') or args.get('our-macs') or args.get('peer-macs'):
                    # special case where we run two commands and produce our own output

                    data = []
                    ourmacs = []
                    peermacs = []
                    use_json = args.get('json')

                    if args.get('macs') or args.get('our-macs'):
                        try:
                            ourmacs = json.loads(check_output(['/usr/bin/clagctl', '-j', 'dumpourmacs']))
                        except Exception as e:
                            ourmacs = []

                    if args.get('macs') or args.get('peer-macs'):
                        try:
                            peermacs = json.loads(check_output(['/usr/bin/clagctl', '-j', 'dumppeermacs']))
                        except Exception as e:
                            peermacs = []

                    for value in ourmacs:
                        if use_json:
                            data.append({
                                'vlanId' : int(value['vlanId']),
                                'interface' : value['outIf'],
                                'mac' : value['mac'],
                                'owner' : 'local'
                            })
                        else:
                            data.append((int(value['vlanId']), value['outIf'], value['mac'], 'local'))

                    for value in peermacs:
                        if use_json:
                            data.append({
                                'vlanId' : int(value['vlanId']),
                                'interface' : value['outIf'],
                                'mac' : value['mac'],
                                'owner' : 'peer'
                            })
                        else:
                            data.append((int(value['vlanId']), value['outIf'], value['mac'], 'peer'))

                    data = sorted(data)

                    if args.get('json'):
                        sys.stdout.write(json.dumps(data, indent=4) + '\n')
                    else:
                        header = ('VLAN', 'Interface', 'Dynamic MAC', 'Owner')
                        sys.stdout.write('\n' + tabulate(tru(data), header) + '\n\n\n')

                else:
                    cmd = []
                    cmd.append('/usr/bin/clagctl')

                    if args.get('verbose'):
                        cmd.append('-v')

                    if args.get('json'):
                        cmd.append('--json')

                    if len(cli.argv_expanded) > 3:
                        for keyword in cli.argv_expanded[3:]:

                            if keyword == 'json':
                                pass

                            elif keyword == 'verbose':
                                pass

                            elif keyword == 'our-multicast-entries':
                                cmd.append('dumpourmcast')

                            elif keyword == 'our-multicast-router-ports':
                                cmd.append('dumpourrport')

                            elif keyword == 'peer-multicast-entries':
                                cmd.append('dumppeermcast')

                            elif keyword == 'peer-multicast-router-ports':
                                cmd.append('dumppeerrport')

                            elif keyword == 'peer-lacp-rate':
                                cmd.append('peerlacprate')

                            elif keyword == 'backup-ip':
                                cmd.append('showbackupip')

                            elif keyword == 'id':
                                cmd.append('showclagid')

                            elif keyword == 'verify-vlans':
                                cmd.append('verifyvlans')

                            else:
                                cmd.append(keyword)

                    try:
                        output = check_output(cmd)
                    except CalledProcessError as e:
                        if 'Unable to communicate with clagd.' not in e.output.strip():
                            output = "\nWARNING: '%s' failed due to:\n%s\n" % (' '.join(cmd), e)
                        else:
                            output = ''
                    sys.stdout.write(output)

            elif args.get('bridge'):
                cmds = []
                cmd = []

                if args.get('macs'):
                    cmd.append('/sbin/bridge')
                    cmd.append('-s')
                    cmd.append('-j')
                    cmd.append('fdb')
                    cmd.append('show')
                    cmds.append(cmd)

                    target_mac = args.get('<mac>')
                    target_ip = args.get('<ip>')
                    target_permanent = args.get('permanent')
                    target_dynamic = args.get('dynamic')

                    if args.get('vlan'):
                        target_vlan = args.get('<number>')
                    else:
                        target_vlan = None

                    try:
                        bridge = json.loads(check_output(cmd))
                    except ValueError as e:
                        bridge = {}
                        output = "\nWARNING: '{0}' failed due to:\n{1}\n".format(' '.join(cmd), e)
                        print output

                    # To test output from someone's switch cut-n-paste their 'bridge -j -s fdb show' here
                    # bridge = json.loads(""" """)

                    if args.get('json'):
                        final = []

                        for row in bridge:

                            if target_mac and row.get('mac') != target_mac:
                                continue

                            if target_vlan and row.get('vlan') != target_vlan:
                                continue

                            if target_ip and row.get('dst') != target_ip:
                                continue

                            if target_permanent and row.get('state') != 'permanent':
                                continue

                            if target_dynamic and row.get('state') == 'permanent':
                                continue

                            final.append(row)
                        sys.stdout.write(json.dumps(final))
                    else:
                        header = ['VLAN',
                                  'Master',
                                  'Interface',
                                  'MAC',
                                  'TunnelDest',
                                  'State',
                                  'Flags',
                                  'LastSeen']
                        data_unsorted = []

                        for row in bridge:

                            if target_mac and row.get('mac') != target_mac:
                                continue

                            if target_vlan and row.get('vlan') != target_vlan:
                                continue

                            if target_ip and row.get('dst') != target_ip:
                                continue

                            if target_permanent and row.get('state') != 'permanent':
                                continue

                            if target_dynamic and row.get('state') == 'permanent':
                                continue

                            # sort_for_humans needs a string so
                            # - build a tuple the represents this row
                            # - join the tuple with a ::: seperator
                            # - add that joined string to data_unsorted
                            # - sort data_unsorted via sort_for_humans
                            # - split each row on ::: and make it a tuple again
                            row_tuple = (row.get('vlan') if row.get('vlan') else 'untagged',
                                         row.get('master'),
                                         row.get('dev'),
                                         row.get('mac'),
                                         row.get('dst'),
                                         row.get('state'),
                                         ', '.join(row.get('flags')) if row.get('flags') else '',
                                         seconds_to_hhmmss(row.get('updated')))
                            data_unsorted.append(':::'.join(map(str, row_tuple)))

                        data = []
                        for row in sort_for_humans(data_unsorted):
                            row_list = []
                            for x in row.split(':::'):
                                if x == 'None':
                                    row_list.append(None)
                                else:
                                    row_list.append(x)
                            data.append(tuple(row_list))

                        sys.stdout.write('\n' + tabulate(tru(data), header) + '\n\n')
                    cmds = []

                elif args.get('link'):
                    cmd.append('/sbin/bridge')
                    cmd.append('link')
                    cmd.append('show')
                    cmds.append(cmd)

                elif args.get('mdb'):
                    cmd.append('/sbin/bridge')
                    cmd.append('mdb')
                    cmd.append('show')
                    cmds.append(cmd)

                elif args.get('vlan'):
                    show_bridge_vlan(args)

                elif args.get('spanning-tree'):
                    ifname = args.get('<interface>')

                    if ifname:
                        if ifname == 'bridge':
                            cmd.append('/sbin/mstpctl')
                            cmd.append('showbridge')

                            if args.get('json'):
                                cmd.append('json')

                            cmd.append('bridge')
                            cmds.append(cmd)
                        else:
                            cmd.append('/sbin/mstpctl')
                            cmd.append('showportdetail')
                            cmd.append('bridge')

                            if args.get('json'):
                                cmd.append('json')
                            cmd.append(ifname)
                            cmds.append(cmd)

                    else:
                        cmd.append('/sbin/mstpctl')
                        cmd.append('showbridge')

                        if args.get('json'):
                            cmd.append('json')

                        cmd.append('bridge')
                        cmds.append(cmd)

                        if args.get('json'):
                            try:
                                showbridge = json.loads(check_output(cmd))
                            except Exception as e:
                                showbridge = {}
                                output = "\nWARNING: '%s' failed due to:\n%s\n" % (' '.join(cmd), e)
                                sys.stdout.write(output)

                        cmd = []
                        cmd.append('/sbin/mstpctl')
                        cmd.append('showportdetail')
                        cmd.append('bridge')

                        if args.get('json'):
                            cmd.append('json')

                        cmds.append(cmd)

                        # This is a special case because we have to combined the output from two json commands
                        if args.get('json'):
                            try:
                                showportdetail = json.loads(check_output(cmd))
                            except Exception as e:
                                showportdetail = {}
                                output = "\nWARNING: '%s' failed due to:\n%s\n" % (' '.join(cmd), e)
                                sys.stdout.write(output + '\n')

                            combined = {}

                            for (key, value) in showbridge.iteritems():
                                combined[key] = value

                            for (key, value) in showportdetail.iteritems():
                                combined[key] = value

                            sys.stdout.write(json.dumps(combined) + '\n')
                            cmds = []

                for cmd in cmds:
                    try:
                        output = check_output(cmd)
                    except Exception as e:
                        output = "\nWARNING: '%s' failed due to:\n%s\n" % (' '.join(cmd), e)
                    sys.stdout.write(output + '\n')

        return command_ok

    def etc_network_interfaces_changed(self):
        """
        Return True if some other user modified /etc/network/interfaces.
        Otherwise, return False.
        """

        if self.etc_network_interfaces_init_mod_time is None:
            # We haven't started making changes to /etc/network/interfaces yet.
            return False

        etc_network_interfaces_mod_time = os.path.getmtime(ETC_NETWORK_INTERFACES)

        if self.etc_network_interfaces_init_mod_time != etc_network_interfaces_mod_time:

            if os.path.exists(PENDING_COMMANDS_FILE):
                print "\nERROR: /etc/network/interfaces was modified by another user"
                print_pending_add_del_commands()
                return True

            # If the user hasn't staged any changes, adopt the new copy of the file.
            else:
                self.etc_network_interfaces_init_mod_time = etc_network_interfaces_mod_time

        return False

    def _save_or_show_pending(self, persist, verbose):
        """
        If persist is True:
        - write changes to /etc/network/interfaces
        - delete various tmp files, such as .del and .original
        - 'ifreload -a' to apply changes

        Otherwise:
        - write changes to /var/run/nclu/iface/interfaces.tmp
        """

        if self.etc_network_interfaces_changed():
            return False

        if persist:
            target_filename = ETC_NETWORK_INTERFACES
        else:
            target_filename = '/var/run/nclu/iface/interfaces.tmp'
            self.init_working_dir()
            copyfile(ETC_NETWORK_INTERFACES, target_filename)

        commit_ok = True
        iface_files = {}      # key:value pairs look like <filename>:[lines of file]
        iface_delete = set()  # Interfaces to delete
        iface_written = []

        # Make a first pass to:
        # - find all interfaces with a .del file
        # - remove the .original files
        for ifname in os.listdir('/var/run/nclu/iface/'):
            filename = "/var/run/nclu/iface/{0}".format(ifname)

            if filename.endswith('.del'):
                ifname = ifname[:-4]
                iface_delete.add(ifname)

                if persist:
                    os.remove(filename)

                    if os.path.isfile("/var/run/nclu/iface/{0}".format(ifname)):
                        os.remove("/var/run/nclu/iface/{0}".format(ifname))

            elif filename.endswith('.original') and persist:
                os.remove(filename)

        # Make a second pass to load the content for interfaces that have been modified
        # but are not being deleted.
        for ifname in os.listdir('/var/run/nclu/iface/'):

            # Ignore these.
            if (ifname in ('interfaces_for_diff', 'interfaces.tmp', 'etc_network_interfaces.backup') or
                ifname.endswith('.del') or ifname.endswith('.original')):
                continue

            filename = "/var/run/nclu/iface/{0}".format(ifname)

            # Ignore directories.  This should only happen if a user creates one manually.
            if not os.path.isfile(filename):
                print "ERROR: commit_pending() found {0} which is not a file (ignoring)".formt(filename)
                continue

            with open(filename) as f:
                iface_files[ifname] = list(f)

        # Sanity check that mako templates are not in use.
        with open(ETC_NETWORK_INTERFACES) as f:
            for line in f:
                line = line.strip()
                if line.startswith('<%') or line.startswith('% '):
                    print "/etc/network/interfaces contains mako templates, net does not support mako templates"
                    return False

        # r+ opens the file for reading and writing.
        with open(target_filename, 'r+') as fh:

            # Read in all lines, and reset the file handle back to the start of the file.
            lines = list(fh)
            fh.seek(0)

            state = 'WRITE'
            prev_line = None
            lines_to_write = []

            # Write all of the lines back to the file, but update the
            # stanzas for interfaces with /var/run/nclu/iface/ files.
            # This will maintain the original order of the stanzas.
            for line in lines:
                line = standardize_indent(line).rstrip()
                re_auto = re.search(r"""^\s*auto\s+(\S+)\s*$""", line)

                if re_auto:
                    ifname = re_auto.group(1)
                    # re_old_loopback = re.search("^\s*auto\s+lo:\d+\s*$", line)

                    # Look to see if any of the interfaces in iface_files need to be
                    # written before this interface in order to keep them sorted
                    # in /etc/network/interfaces.
                    ifaces_to_add_here = []
                    sorted_interfaces = sort_ifaces(iface_files.keys() + [ifname, ])

                    for x in sorted_interfaces:
                        if x == ifname:
                            break
                        ifaces_to_add_here.append(x)

                    # log.info("%s: line '%s'" % (ifname, line))
                    # log.info("%s: sorted_interface %s" % (ifname, sorted_interfaces))
                    # log.info("%s: interfaces to add here %s" % (ifname, ifaces_to_add_here))

                    if ifaces_to_add_here:
                        # We are at an 'auto' line and we need to insert interfaces
                        # prior to that auto line.  That part is easy but what if
                        # there are comments before this auto line?  We need to
                        # insert our interfaces before those comments.
                        comments = []

                        for x in reversed(lines_to_write):
                            if x.startswith('#'):
                                comments.append(x)
                            else:
                                break

                        # Remove the comments for this 'auto'.
                        if comments:
                            comments_to_chop = -1 * len(comments)
                            lines_to_write = lines_to_write[:comments_to_chop]

                        # Add the interfaces that need to go before this auto.
                        for x in ifaces_to_add_here:
                            write_new_interface(lines_to_write, prev_line, x, iface_files[x], persist)
                            del iface_files[x]
                            iface_written.append(x)

                        # Restore the comments for this 'auto'.
                        lines_to_write.extend(reversed(comments))

                    # Do not write any of the configuration lines for an interface we are trying to delete.
                    if ifname in iface_delete:
                        log.debug("{0}: remove from {1}".format(ifname, target_filename))
                        iface_delete.remove(ifname)

                        if state == 'WRITE':
                            # log.debug("WRITE->SKIP: %s" % line)
                            state = 'SKIP'

                    # This was causing issues where lo:1 would be deleted but the config
                    # lines were not being moved under lo...commenting this out for now
                    # to reduce risk for GA

                    # Do not write the old style lo:1 interface, their content was merged into
                    # the 'auto lo' interface
#                    elif re_old_loopback:
#                        if state == 'WRITE':
                            # log.debug("WRITE->SKIP: %s" % line)
#                            state = 'SKIP'

                    # This is an interface that has been modified via iface. Write the config lines
                    # from the iface file to target_filename
                    elif ifname in iface_files:
                        log.debug("{0}: update in {1}".format(ifname, target_filename))

                        for iface_line in iface_files[ifname]:
                            iface_line = iface_line.rstrip()
                            log.debug("{0}: write '{1}'".format(ifname, iface_line))
                            lines_to_write.append(iface_line + '\n')
                        lines_to_write.append('\n')

                        if persist:
                            os.remove("/var/run/nclu/iface/{0}".format(ifname))
                        del iface_files[ifname]
                        iface_written.append(ifname)

                        if state == 'WRITE':
                            # log.debug("WRITE->SKIP: %s" % line)
                            state = 'SKIP'

                    # We write interfaces in a very specific order so if this
                    # is true it means an interface was in /etc/network/interfaces
                    # but not in the order used by nclu.  So the first time the
                    # user makes a change we'll move the interface section to be
                    # in the correct order...avoid writing the same interface twice.
                    elif ifname in iface_written:
                        if state == 'WRITE':
                            state = 'SKIP'

                    else:
                        # The interface that has not been modified via iface.
                        iface_written.append(ifname)

                        if state == 'SKIP':
                            # log.debug("SKIP->WRITE: %s" % line)
                            state = 'WRITE'

                elif line.startswith('source '):
                    if state == 'SKIP':
                        # log.debug("SKIP->WRITE: %s" % line)
                        state = 'WRITE'

                elif state == 'SKIP' and line.startswith('#'):
                    # We are parsing comments.  Set state back to WRITE.
                    state = 'WRITE'
                    # log.debug("SKIP->WRITE: %s" % line)

                if state == 'WRITE':
                    lines_to_write.append(line + '\n')
                elif state == 'SKIP':
                    pass
                else:
                    raise ValueError("state = {0}".format(state))

                prev_line = line

            # Write any remaining iface files to target_filename.
            # These will be sorted in human readable order
            # log.info("%s: ifaces left to add %s\n" % (ifname, iface_files.keys()))
            for ifname in sort_ifaces(iface_files.keys()):
                write_new_interface(lines_to_write, prev_line, ifname, iface_files[ifname], persist)

            # Now that we have constructed what we want to write, write it.

            prev_line = None
            number_lines_to_write = len(lines_to_write)

            for (i, line) in enumerate(lines_to_write):
                if line.strip():
                    fh.write(line)
                else:
                    # Do not print multiple blank lines in a row
                    if prev_line is None or prev_line.strip():

                        # Do not print a blank line at the end
                        if i < number_lines_to_write - 1:
                            fh.write(line)
                prev_line = line

            # If we deleted lines we need to truncate the rest of the fh so that we
            # do not end up with the lines at the end printed twice
            fh.truncate()

        if persist:
            if verbose:
                print "Updating /etc/network/interfaces and applying changes via 'ifreload -va'"
                try:
                    output = check_output(['/sbin/ifreload', '-va'], close_fds=True, stderr=STDOUT)
                except CalledProcessError as e:
                    output = "\nERROR: 'ifreload -va' failed due to:\n%s\n%s\n" % (e.output.strip(), e)
                    commit_ok = False

            else:
                if self.show_linux_command:
                    print "Updating /etc/network/interfaces, and applying changes via 'ifreload -a'"

                try:
                    output = check_output(['/sbin/ifreload', '-a'], close_fds=True, stderr=STDOUT)
                except CalledProcessError as e:
                    output = "\nERROR: 'ifreload -a' failed due to:\n%s\n%s\n" % (e.output.strip(), e)
                    commit_ok = False

            if output:
                print output

        return commit_ok

    def validate_net_cmd_vs_ifnames(self, cli):
        args = cli.args

        if args.get('add') or args.get('del'):

            if args.get('vlan-id') and args.get('<number-range-list>'):
                vlan_numbers = glob_to_numbers(str(args.get('<number-range-list>')))

                if len(vlan_numbers) > 1:
                    print "ERROR: cannot assign vlan-id to a range of vlans"
                    return False

            self.get_ifquery_all()
            (ifnames_valid, ifnames) = get_ifnames(cli)

            if not ifnames_valid:
                return False

            if ifnames:
                for ifname in ifnames:
                    iface = EtcNetworkInterface(ifname, self.ifquery_all)

                    # If this returned False an ERROR message was already printed
                    if not iface.validate_net_cmd_vs_ifname(self, cli):
                        return False

        return True
