# Copyright (C) 2016-2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:

    # Linux Basics
    netmisc add hostname <text>
    netmisc show hostname
    netmisc show version
    netmisc show vrf
    net show configuration dhcp [json]

Options:
    help  : Show this screen and exit
    dhcp  : Dynamic Host Configuration Protocol
    relay : IPv4 relay agent
    relay6 : IPv6 relay agent
    version : A software version number
    server : IPv4 server
    server6 : IPv6 server
    unnumbered : An interface without an IP address
    downstream : A downstream interface
    upstream : An upstream interface
"""

from collections import OrderedDict
from nclu import (
    decode_dict,
    file_set_key_value,
    files_match,
    get_interfaces,
    ifname_expand_glob,
    ifname_is_glob,
    PENDING_COMMIT_NOCHANGE,
    PENDING_COMMIT_CHANGE,
    ifnames_to_glob,
    numbers_to_glob,
    print_diff,
    NCLU_VERSION,
    Capturing,
    tabulate_remove_unicode as tru
)

from network_docopt import sort_for_humans
from os.path import exists
from shutil import copyfile
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate
import json
import logging
import logging.handlers
import os
import re
import shlex
import shutil
import sys

try:
    from netshowlib import (
        interface_related,
        check as vendor_check,
        ShowInterfaces,
        CumulusShowInterfaces,
        ShowNeighbors,
        CumulusShowNeighbors,
        ShowSystem,
        CumulusShowSystem,
        exec_command,
        lldp_cacheinfo,
        read_symlink)
    netshow_present = True
except ImportError:
    netshow_present = False

log = logging.getLogger('netd')

# File name constants

# DHCP IPv4 relay
ISC_DHCP_RELAY4 = '/etc/default/isc-dhcp-relay'
ISC_DHCP_RELAY_SCRATCHPAD4 = '/var/run/nclu/netmisc/isc_dhcp_relay'
ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP = '/var/run/nclu/netmisc/isc_dhcp_relay.backup'

# DHCP IPv6 relay
ISC_DHCP_RELAY6 = '/etc/default/isc-dhcp-relay6'
ISC_DHCP_RELAY_SCRATCHPAD6 = '/var/run/nclu/netmisc/isc_dhcp_relay6'
ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP = '/var/run/nclu/netmisc/isc_dhcp_relay6.backup'

# DHCP IPv4 server
DHCPD_CONFIG4 = '/etc/dhcp/dhcpd.conf'
DHCPD_PID4 = '/var/run/dhcpd.pid'
ISC_DHCP_SERVER4 = '/etc/default/isc-dhcp-server'
ISC_DHCP_SERVER_SCRATCHPAD4 = '/var/run/nclu/netmisc/isc_dhcp_server'
ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP = '/var/run/nclu/netmisc/isc_dhcp_server.backup'

# DHCP IPv6 server
DHCPD_CONFIG6 = '/etc/dhcp/dhcpd6.conf'
DHCPD_PID6 = '/var/run/dhcpd6.pid'
ISC_DHCP_SERVER6 = '/etc/default/isc-dhcp-server6'
ISC_DHCP_SERVER_SCRATCHPAD6 = '/var/run/nclu/netmisc/isc_dhcp_server6'
ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP = '/var/run/nclu/netmisc/isc_dhcp_server6.backup'

# ports.conf
PORTS_CONF = '/etc/cumulus/ports.conf'
PORTS_CONF_SCRATCHPAD = '/var/run/nclu/netmisc/ports_conf'
PORTS_CONF_SCRATCHPAD_BACKUP = '/var/run/nclu/netmisc/ports_conf.backup'

# hostname
ETC_HOSTS = '/etc/hosts'
ETC_HOSTNAME = '/etc/hostname'
ETC_HOSTS_SCRATCHPAD = '/var/run/nclu/netmisc/etc_hosts'
ETC_HOSTS_SCRATCHPAD_BACKUP = '/var/run/nclu/netmisc/etc_hosts.backup'
ETC_HOSTNAME_SCRATCHPAD = '/var/run/nclu/netmisc/etc_hostname'
ETC_HOSTNAME_SCRATCHPAD_BACKUP = '/var/run/nclu/netmisc/etc_hostname.backup'
EXECLIST = '/var/run/nclu/netmisc/execlist'


def ifname_to_ports_conf_index(ifname):
    re_swp = re.match(r"""^swp(\d+)$""", ifname)
    if re_swp is not None:
        return int(re_swp.group(1))
    return None


def ports_conf_edit(args):
    command_ok = True

    if not exists(PORTS_CONF_SCRATCHPAD):
        if not exists(PORTS_CONF):
            print 'ERROR: %s does not exist' % PORTS_CONF
            return
        copyfile(PORTS_CONF, PORTS_CONF_SCRATCHPAD)

    with open(PORTS_CONF_SCRATCHPAD) as f:
        lines = [line.strip() for line in f]

    for new_speed in ('1x', '2x', '4x', '/4', 'disabled'):
        if args.get(new_speed):
            break
    else:
        # This should never happen.
        print "ERROR: speed for port was not specified"
        return False

    ports_to_edit = []
    ifname = args.get('<interface>')

    if ifname_is_glob(ifname):
        for x in ifname_expand_glob(ifname):
            ports_conf_index = ifname_to_ports_conf_index(x)

            if ports_conf_index:
                ports_to_edit.append(ports_conf_index)
            else:
                print "ERROR: The 'breakout' command can only be used on physical interfaces such as swp1, swp2, etc."
                command_ok = False
    else:
        ports_conf_index = ifname_to_ports_conf_index(ifname)

        if ports_conf_index:
            ports_to_edit.append(ports_conf_index)
        else:
            print "ERROR: The 'breakout' command can only be used on physical interfaces such as swp1, swp2, etc."
            command_ok = False

    if command_ok:
        port_number_setting_re = re.compile(r"""^(?P<port_number>\d+)\s*=\s*(?P<port_setting>\S+)\s*$""")

        with open(PORTS_CONF_SCRATCHPAD, 'w') as f:
            for line in lines:
                port_number_setting_match = port_number_setting_re.match(line)

                if port_number_setting_match:
                    port_number = int(port_number_setting_match.group('port_number'))
                    port_setting = port_number_setting_match.group('port_setting')

                    if port_number in ports_to_edit:
                        if port_setting == new_speed:
                            print "'{0}' configuration already has '{1}'.".format(port_number, new_speed)
                        f.write("{0}={1}\n".format(port_number, new_speed))
                    else:
                        f.write("{0}={1}\n".format(port_number, port_setting))
                else:
                    f.write(line + '\n')

    return command_ok


def dhcp_relay_edit(args):
    """
    The guts for the following:

    netmisc add dhcp relay server <ipv4>
    netmisc del dhcp relay server [<ipv4>]

    netmisc add dhcp relay interface <interface>
    netmisc del dhcp relay interface [<interface>]

    # Upstream, for IPv6, specifies the interface to which queries from clients, and other relay
    # agents, should be forwarded.  The user has the option to specify a unicast/multicast address.

    netmisc add dhcp relay6 upstream <interface> [<ipv6>]
    netmisc del dhcp relay6 upstream [<interface>]

    # Downstream, for IPv6, specifies the interface on which queries will be received from clients
    # and other relay agents.  The user has the option to specify a link address.

    netmisc add dhcp relay6 downstream <interface> [<ipv6>]
    netmisc del dhcp relay6 downstream [<interface>]

    IPv4 and IPv6 configurations live in different files.
    https://docs.cumulusnetworks.com/display/DOCS/Configuring+DHCP+Relays+and+Servers
    """
    # Does the user want to modify settings for IPv4 (relay) or IPv6 (relay6)?
    RELAY_FILE = ISC_DHCP_RELAY4 if args.get('relay') else ISC_DHCP_RELAY6
    SCRATCHPAD = ISC_DHCP_RELAY_SCRATCHPAD4 if args.get('relay') else ISC_DHCP_RELAY_SCRATCHPAD6

    # Create a default configuration file if a configuration file does not already exist.
    if not exists(SCRATCHPAD):
        if not exists(RELAY_FILE):
            with open(RELAY_FILE, 'w') as f:
                f.write('SERVERS=""\n')
                f.write('INTF_CMD=""\n')
                f.write('OPTIONS=""\n')
        copyfile(RELAY_FILE, SCRATCHPAD)

    # Read the configuration file.
    with open(SCRATCHPAD) as f:
        dhcp_config = f.read()

    # Using the appropriate helper function, update the text of dhcp_config.
    # If something goes wrong, the helper functions return None.

    # IPv4 commands
    if args.get('add') and args.get('relay'):
        if args.get('server'):
            # net add dhcp relay server <ipv4>
            dhcp_config = add_dhcp_relay_server(args.get('<ipv4>'), dhcp_config)
        elif args.get('interface'):
            # net add dhcp relay interface <interface>
            dhcp_config = add_dhcp_relay_interface(args.get('<interface>'), dhcp_config)

    elif args.get('del') and args.get('relay'):
        if args.get('server'):
            # net del dhcp relay server [<ipv4>]
            dhcp_config = del_dhcp_relay_server(args.get('<ipv4>'), dhcp_config)
        elif args.get('interface'):
            # net del dhcp relay interface [<interface>]
            dhcp_config = del_dhcp_relay_interface(args.get('<interface>'), dhcp_config)

    # IPv6 commands
    elif args.get('add') and args.get('relay6'):
        if args.get('upstream'):      # server
            # net add dhcp relay6 upstream <interface> [<ipv6>]
            dhcp_config = add_dhcp_relay6_server(args.get('<interface>'), args.get('<ipv6>'), dhcp_config)
        elif args.get('downstream'):  # interface
            # net add dhcp relay6 downstream <interface> [<ipv6>]
            dhcp_config = add_dhcp_relay6_interface(args.get('<interface>'), args.get('<ipv6>'), dhcp_config)

    elif args.get('del') and args.get('relay6'):
        if args.get('upstream'):      # server
            # net del dhcp relay6 upstream [<interface>]
            dhcp_config = del_dhcp_relay6_server(args.get('<interface>'), dhcp_config)
        elif args.get('downstream'):  # interface
            # net del dhcp relay6 downstream [<interface>]
            dhcp_config = del_dhcp_relay6_interface(args.get('<interface>'), dhcp_config)

    else:
        # This shouldn't happen.  It indicates a programming error.
        raise RuntimeError("Cannot determine which DHCP relay command to use.")

    command_ok = dhcp_config is not None

    if command_ok:
        # Write the modified configuration file to disk.
        with open(SCRATCHPAD, 'w') as f:
            f.write(dhcp_config)

    return command_ok


def add_dhcp_relay_server(ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return
    the modified string that results from adding a new server IPv4 address.

    If the given IP is already in the SERVERS string, return None to indicate
    an error.

    net add dhcp relay server <ipv4>
    """
    # Create a string containing the existing server IPs in the configuration file.
    server_line_re = re.compile(r'''^\s*#?\s*SERVERS\s*=\s*"(?P<ips>[.\d ]*)"''', re.IGNORECASE | re.MULTILINE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    server_ips = server_line.group('ips')  # Might be the empty string

    if ip in server_ips:
        print 'The DHCP relay SERVERS configuration already has the IP address {0}.'.format(ip)
        return None  # Error

    # Add the new IP and/or interface to the configuration file text.
    # Don't include a lead blank if the SERVER string does not already have values.
    if server_ips:
        return server_line_re.sub('SERVERS="{0} {1}"'.format(server_ips, ip), dhcp_config)
    return server_line_re.sub('SERVERS="{0}"'.format(ip), dhcp_config)


def add_dhcp_relay_interface(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from adding one or more new IPv4 relay interfaces.

    If any of the given interfaces are already in the INTF_CMD string, return None to
    indicate an error.

    net add dhcp relay interface <interface>

    interface - A non-empty string; an interface name or glob
    """

    assert isinstance(interface, unicode) and interface

    # Create a list of strings from the interface(s) the user specified.
    ifnames = ifname_expand_glob(interface)

    # Create a string containing the existing interface commands in the configuration file.
    intf_cmd_line_re = re.compile(r'''^\s*#?\s*INTF_CMD\s*=\s*"(?P<interfaces>[ \S]*?)"''', re.IGNORECASE | re.MULTILINE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMD="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-i swp1 -i swp2"; might be the empty string.

    for ifname in ifnames:
        if re.search(r'''\b{0}\b'''.format(ifname), intf_cmd) is not None:
            print 'The DHCP relay INTF_CMD configuration already has the interface {0}.'.format(ifname)
            return None  # Error

    # Add the new interface(s) to the configuration file text.
    # Don't include a lead blank if the INTF_CMD string does not already have values.
    if intf_cmd:
        return intf_cmd_line_re.sub('INTF_CMD="{0} -i {1}"'.format(intf_cmd, ' -i '.join(ifnames)), dhcp_config)
    return intf_cmd_line_re.sub('INTF_CMD="-i {0}"'.format(' -i '.join(ifnames)), dhcp_config)


def del_dhcp_relay_server(ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return
    the modified string that results from removing a given server IPv4 address.
    If no IPv4 address is given, remove all server IPv4 addresses.

    If the given IP is not in the SERVERS string, return None to indicate
    an error.

    net del dhcp relay server [<ipv4>]
    """
    server_line_re = re.compile(r'''SERVERS\s*=\s*"(?P<ips>[.\d ]*)"''', re.IGNORECASE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    if not ip:
        # The user did not specifiy an IP to remove.  Remove all IPs.
        return server_line_re.sub('SERVERS=""', dhcp_config)

    # Create a string containing the existing server IPs in the configuration file.
    server_ips = server_line.group('ips')  # Might be the empty string

    if ip not in server_ips:
        print 'The DHCP relay SERVERS configuration does not include the IP address {0}.'.format(ip)
        return None  # Error

    # Remove the given IPv4 address from the configuration.
    server_ips = re.sub(ip, '', server_ips)

    # Replace runs of spaces with a single space.
    server_ips = re.sub(r"""\s{2,}""", ' ', server_ips)

    # Replace the contents of SERVERS in the configuration file.
    return server_line_re.sub('SERVERS="{0}"'.format(server_ips.strip()), dhcp_config)


def del_dhcp_relay_interface(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from removing one or more IPv4 relay interfaces.

    If any of the given interfaces are already in the INTF_CMD string, return None to
    indicate an error.

    net del dhcp relay interface [<interface>]
    """
    intf_cmd_line_re = re.compile(r'''INTF_CMD\s*=\s*"(?P<interfaces>[ \S]*?)"''', re.IGNORECASE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    if not interface:
        # The user did not specifiy an interface to remove.  Remove all interfaces.
        return intf_cmd_line_re.sub('INTF_CMD=""', dhcp_config)

    # Create a list of the interface(s) the user specified.  The input might be a glob.
    ifnames = ifname_expand_glob(interface)  # A list of strings

    # Create a string containing the existing interface commands in the configuration file.
    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-i swp1 -i swp2"; might be the empty string.

    for ifname in ifnames:
        if re.search(r'''\b{0}\b'''.format(ifname), intf_cmd) is None:
            print 'The DHCP relay INTF_CMD configuration does not include the interface {0}.'.format(ifname)
            return None  # Error

        # Remove the interface from the configuration.
        intf_cmd = re.sub(r'''-i \b{0}\b'''.format(ifname), '', intf_cmd, re.IGNORECASE)

    # Replace runs of spaces with a single space.
    intf_cmd = re.sub(r"""\s{2,}""", ' ', intf_cmd)

    # Replace the contents of INTF_CMD in the configuration file.
    return intf_cmd_line_re.sub(r'''INTF_CMD="{0}"'''.format(intf_cmd.strip()), dhcp_config)


def add_dhcp_relay6_server(interface, ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from adding a new server (upstream) interface name.
    The user optionally can specify an IPv6 address, but this function will accept
    ip=None.

    If the given interface or IPv6 is already in the SERVERS string, return None to
    indicate an error.

    net add dhcp relay6 upstream <interface> [<ipv6>]
    """
    ifnames = ifname_expand_glob(interface) if ifname_is_glob(interface) else [interface]  # A list of strings

    # Create a string containing the existing server interfaces and IPs in the configuration file.
    server_line_re = re.compile(r'''^\s*#?\s*SERVERS\s*=\s*"(?P<interfaces>[-u :\da-z%]*)"''', re.IGNORECASE | re.MULTILINE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    server_interfaces = server_line.group('interfaces')  # Might be the empty string
    server_interfaces_list = server_interfaces.replace('-u', '').strip().split()

    # Test for error conditions.
    for ifname in ifnames:
        if ip:
            new_server_interface = ip + '%' + ifname
        else:
            new_server_interface = ifname

        if new_server_interface in server_interfaces_list:
            print 'The DHCP relay SERVERS configuration already has {0}.'.format(new_server_interface)
            return None  # Error

    # Add the new interface(s), and mabye IPv6, to the configuration file text.
    # Don't include a lead blank if the SERVER string does not already have values.
    if server_interfaces:
        if ip:
            return server_line_re.sub('SERVERS="{0} -u {1}%{2}"'.format(server_interfaces, ip, ' -u {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return server_line_re.sub('SERVERS="{0} -u {1}"'.format(server_interfaces, ' -u '.join(ifnames)), dhcp_config)
    else:
        if ip:
            return server_line_re.sub('SERVERS="-u {0}%{1}"'.format(ip, ' -u {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return server_line_re.sub('SERVERS="-u {0}"'.format(' -u '.join(ifnames)), dhcp_config)


def add_dhcp_relay6_interface(interface, ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from adding a new relay interface (downstream).
    The user optionally can specify an IPv6 address, but this function will accept
    ip=None.

    If the given interface or IPv6 is already in the INTF_CMD string, return None to
    indicate an error.

    net add dhcp relay6 downstream <interface> [<ipv6>]
    """

    ifnames = ifname_expand_glob(interface)  # A list of strings

    # Create a string containing the existing server interfaces and IPs in the configuration file.
    intf_cmd_line_re = re.compile(r'''^\s*#?\s*INTF_CMD\s*=\s*"(?P<interfaces>[-l :\da-z%]*)"''', re.IGNORECASE | re.MULTILINE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-l 1::%vlan100"; might be the empty string.

    # Test for error conditions.
    for ifname in ifnames:
        if re.search(r'''\b{0}\b'''.format(ifname), intf_cmd) is not None:
            print 'The DHCP relay INTF_CMD configuration already has the interface {0}.'.format(ifname)
            return None  # Error
    if ip and ip in intf_cmd:
            print 'The DHCP relay INTF_CMD configuration already has the IPv6 address {0}.'.format(ip)
            return None  # Error

    # Add the new interface(s), and mabye IPv6, to the configuration file text.
    # Don't include a lead blank if the INTF_CMD string does not already have values.
    if intf_cmd:
        if ip:
            return intf_cmd_line_re.sub('INTF_CMD="{0} -l {1}%{2}"'.format(intf_cmd, ip, ' -l {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return intf_cmd_line_re.sub('INTF_CMD="{0} -l {1}"'.format(intf_cmd, ' -l '.join(ifnames)), dhcp_config)
    else:
        if ip:
            return intf_cmd_line_re.sub('INTF_CMD="-l {0}%{1}"'.format(ip, ' -l {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return intf_cmd_line_re.sub('INTF_CMD="-l {0}"'.format(' -l '.join(ifnames)), dhcp_config)


def del_dhcp_relay6_server(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return
    the modified string that results from removing a given server IPv6 interface.
    If no IPv6 interface is given, remove all server IPv6 interfaces.

    If the given IPv6 interface is not in the SERVERS string, return None to indicate
    an error.

    net del dhcp relay6 upstream [<interface>]
    """
    server_line_re = re.compile(r'''SERVERS\s*=\s*"(?P<interfaces>[-u :\da-z%]*)"''', re.IGNORECASE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    if not interface:
        # The user did not specifiy an IP to remove.  Remove all IPs.
        return server_line_re.sub('SERVERS=""', dhcp_config)

    # Create a list of the interface(s) the user specified.  The input might be a glob.
    ifnames = ifname_expand_glob(interface) if ifname_is_glob(interface) else [interface]  # A list of strings

    # Create a string containing the existing server interfaces in the configuration file.
    server_interfaces = server_line.group('interfaces')  # Might be the empty string

    # Test for an error condition.
    for ifname in ifnames:
        if ifname not in server_interfaces:
            print 'The DHCP relay SERVERS configuration does not include the interface {0}.'.format(interface)
            return None  # Error

        # Remove the given IPv6 interface from the configuration.
        server_interfaces = re.sub('-u (?:[:\da-f]+%)?{0}'.format(ifname), '', server_interfaces)

    # Replace runs of spaces with a single space.
    server_interfaces = re.sub(r"""\s{2,}""", ' ', server_interfaces)

    # Replace the contents of SERVERS in the configuration file.
    return server_line_re.sub('SERVERS="{0}"'.format(server_interfaces.strip()), dhcp_config)


def del_dhcp_relay6_interface(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from removing one or more IPv6 relay interfaces.

    If any of the given interfaces are already in the INTF_CMD string, return None to
    indicate an error.

    net del dhcp relay6 downstream [<interface>]
    """

    intf_cmd_line_re = re.compile(r'''INTF_CMD\s*=\s*"(?P<interfaces>[-l :\da-z%]*)"''', re.IGNORECASE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    if not interface:
        # The user did not specifiy an interface to remove.  Remove all interfaces.
        return intf_cmd_line_re.sub('INTF_CMD=""', dhcp_config)

    # Create a list of the interface(s) the user specified.  The input might be a glob.
    ifnames = ifname_expand_glob(interface)  # A list of strings

    # Create a string containing the existing interface commands in the configuration file.
    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-l 1::%vlan100"; might be the empty string.

    for ifname in ifnames:
        if re.search(r'''\b{0}\b'''.format(ifname), intf_cmd) is None:
            print 'The DHCP relay INTF_CMD configuration does not include the interface {0}.'.format(ifname)
            return None  # Error

        # Remove the interface from the configuration.
        intf_cmd = re.sub(r'''-l \b(?:[:\da-f]+%)?{0}\b'''.format(ifname), '', intf_cmd, re.IGNORECASE)

    # Replace runs of spaces with a single space.
    intf_cmd = re.sub(r"""\s{2,}""", ' ', intf_cmd)

    # Replace the contents of INTF_CMD in the configuration file.
    return intf_cmd_line_re.sub(r'''INTF_CMD="{0}"'''.format(intf_cmd.strip()), dhcp_config)


def dhcp_server_edit(args):
    """
    The guts for the following:

    netmisc add dhcp server <interface>
    netmisc del dhcp server [<interface>]
    netmisc add dhcp server6 <interface>
    netmisc del dhcp server6 [<interface>]

    IPv4 and IPv6 configurations live in different files.
    https://docs.cumulusnetworks.com/display/DOCS/Configuring+DHCP+Relays+and+Servers
    """
    command_ok = True

    # Does the user want to modify settings for IPv4 (server) or IPv6 (server6)?
    SERVER_FILE = ISC_DHCP_SERVER4 if args.get('server') else ISC_DHCP_SERVER6
    SCRATCHPAD = ISC_DHCP_SERVER_SCRATCHPAD4 if args.get('server') else ISC_DHCP_SERVER_SCRATCHPAD6
    DHCPD_CONFIG = DHCPD_CONFIG4 if args.get('server') else DHCPD_CONFIG6
    DHCPD_PID = DHCPD_PID4 if args.get('server') else DHCPD_PID6

    # Create a default configuration file if a configuration file does not already exist.
    if not exists(SCRATCHPAD):
        if not exists(SERVER_FILE):
            with open(SERVER_FILE, 'w') as f:
                f.write('DHCPD_CONF="-cf {0}"\n'.format(DHCPD_CONFIG))
                f.write('DHCPD_PID="-pf {0}"\n'.format(DHCPD_PID))
                f.write('OPTIONS=""\n')
                f.write('INTERFACES=""\n')
        copyfile(SERVER_FILE, SCRATCHPAD)

    # Read the configuration file.
    with open(SCRATCHPAD) as f:
        dhcp_config = f.read()

    # Get the interface(s) the user specified, if any.
    ifname = args.get('<interface>')
    if ifname is not None:  # Might be None for a "del" command
        ifnames = ifname_expand_glob(ifname) if ifname_is_glob(ifname) else [ifname]  # A list of strings

    # Create a string containing the existing interface commands in the configuration file.
    interfaces_re = re.compile(r'''^\s*#?\s*INTERFACES\s*=\s*"(?P<interfaces>[ \S]*?)"''', re.IGNORECASE | re.MULTILINE)
    interfaces_obj = interfaces_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not interfaces_obj:
        dhcp_config += '\nINTERFACES=""\n'
        interfaces_obj = interfaces_re.search(dhcp_config)

    interfaces = interfaces_obj.group('interfaces')  # A string like "swp1 swp2"; might be the empty string.

    if args.get('add'):     # net add dhcp server <interface>
        interfaces_to_add = []

        for ifname in ifnames:
            if ifname in interfaces:
                print 'DHCP server INTERFACES configuration already has {0}.'.format(ifname)
                command_ok = False
            else:
                interfaces_to_add.append(ifname)

        if interfaces_to_add:
            # Add the interface(s) to the configuration file text.  Don't include a lead blank if
            # the INTERFACES string does not already have values.
            if interfaces:
                dhcp_config = interfaces_re.sub('INTERFACES="{0} {1}"'.format(interfaces, ' '.join(interfaces_to_add)), dhcp_config)
            else:
                dhcp_config = interfaces_re.sub('INTERFACES="{0}"'.format(' '.join(interfaces_to_add)), dhcp_config)

    elif args.get('del'):  # net del dhcp server ...
        if ifname is not None:
            # net del dhcp server <interface>
            # Remove the given interface from the configuration file text.
            for ifname in ifnames:
                if ifname in interfaces:
                    interfaces = re.sub(ifname, '', interfaces, re.IGNORECASE)
                else:
                    print 'DHCP server INTERFACES configuration does not have {0}.'.format(ifname)
                    command_ok = False

            # Replace multi-spaces in the interfaces string with single spaces.
            interfaces = re.sub(r"""\s{2,}""", ' ', interfaces)

            # Update the INTERFACES line in the configuration file text.
            dhcp_config = interfaces_re.sub('INTERFACES="{0}"'.format(interfaces.strip()), dhcp_config)

        else:
            # net del dhcp server
            # The user did not specifiy an interface to remove.  Remove all interfaces.
            dhcp_config = interfaces_re.sub('INTERFACES=""', dhcp_config)

    if command_ok:
        # Write the modified configuration file to disk.
        with open(SCRATCHPAD, 'w') as f:
            f.write(dhcp_config)

    return command_ok


def hostname_valid(hostname):
    return re.match(r"""[-a-zA-Z\d]{1,63}""", hostname) is not None


def hostname_edit(new_hostname):
    command_ok = True

    if not hostname_valid(new_hostname):
        print "ERROR: %s is an invalid hostname, hostnames may only contain letters, numbers, or -" % new_hostname
        return

    # execlist
    with open(EXECLIST, 'w') as f:
        f.write('hostname %s\n' % new_hostname)
        f.write('hostnamectl set-hostname %s\n' % new_hostname)
        f.write('echo "You must re-login for your bash prompt to reflect the hostname change"\n')

    # /etc/hostname
    with open(ETC_HOSTNAME_SCRATCHPAD, 'w') as f:
        f.write("%s\n" % new_hostname)

    old_hostname = check_output(['/bin/hostname']).strip()

    # /etc/hosts
    with open(ETC_HOSTS) as f:
        etc_hosts_lines = [line.strip() for line in f]

    with open(ETC_HOSTS_SCRATCHPAD, 'w') as f:
        for line in etc_hosts_lines:
            re_old_hostname = re.match(r"""^(.*\s+)%s$""" % old_hostname, line)

            if re_old_hostname:
                f.write("{0}{1}\n".format(re_old_hostname.group(1), new_hostname))
            else:
                f.write(line + '\n')

    return command_ok


def build_doc_string():
    global __doc__

    __doc__ = __doc__.replace('    netmisc show version',"""    netmisc show version

    # interface show commands
    netmisc clear counters
    netmisc show counters [json]
    netmisc show interface (all|bonds|bondmems) [mac|json]
    netmisc show interface [<interface>] [json]
    netmisc show interface <interface> detail
    netmisc show interface pluggables [json]
    netmisc show lldp [json]
    netmisc show lldp [<interface>] [json]
    netmisc show system""")

    __doc__ = __doc__.replace('Show this screen and exit', """Show this screen and exit
    access     : Summary of physical ports with l2 or l3 config
    bondmems   : Summary of bond members
    bonds      : Summary of bonds
    bridges    : Summary of ports with bridge members
    counters   : show netstat counters
    interface  : Summary info of all interfaces
    l3         : Summary of ports with an IP
    lldp       : Link Layer Discovery Protocol
    system     : System information
    trunks     : Summary of trunk interfaces
    all        : Everything
    pluggables : SFPs
    json       : Print output in json
    mac        : Media Access Control
            """)

    # dhcp relay (IPv4)
    __doc__ = __doc__.replace('    netmisc show version', """    netmisc show version
    netmisc add dhcp relay server <ipv4>
    netmisc del dhcp relay server [<ipv4>]
    netmisc add dhcp relay interface <interface>
    netmisc del dhcp relay interface [<interface>]""")

    # dhcp relay6 (IPv6)
    __doc__ = __doc__.replace('    netmisc show version', """    netmisc show version
    netmisc add dhcp relay6 upstream <interface> [<ipv6>]
    netmisc del dhcp relay6 upstream [<interface>]
    netmisc add dhcp relay6 downstream <interface> [<ipv6>]
    netmisc del dhcp relay6 downstream [<interface>]""")

    # dhcp server (IPv4)
    __doc__ = __doc__.replace('    netmisc show version', """    netmisc show version
    netmisc add dhcp server <interface>
    netmisc del dhcp server [<interface>]""")

    # dhcp server6 (IPv6)
    __doc__ = __doc__.replace('    netmisc show version', """    netmisc show version
    netmisc add dhcp server6 <interface>
    netmisc del dhcp server6 [<interface>]""")

    # /etc/cumulus/ports.conf
    if exists(PORTS_CONF):
        __doc__ = __doc__.replace('    netmisc show version',"""    netmisc show version

    # /etc/cumulus/ports.conf
    netmisc add interface <interface> breakout (1x|2x|4x|/4|disabled)
            """)

        __doc__ = __doc__.replace('Show this screen and exit', """Show this screen and exit
    1x       : Do not split or gang the port
    2x       : Split the port into two interfaces
    4x       : Split the port into four interfaces
    /4       : Gang four ports into a single port
    disabled    : Disabled
    breakout    : Split an interface into multiple interface via a breakout cable
            """)


def get_pluggables(ifnames):
    data = OrderedDict()

    for ifname in ifnames:
        if re.match('swp\d+', ifname) or re.match('swp\d+s\d+', ifname):
            try:
                output = check_output(['/sbin/ethtool', '-m', ifname], stderr=STDOUT)
            except Exception:
                continue

            data[ifname] = {}

            for line in output.splitlines():
                # compress multiple spaces
                line = ' '.join(line.strip().split())

                if 'No such device' in line or 'Cannot get Module EEPROM data' in line:
                    break

                (key, value) = line.split(':', 1)
                key = key.strip()
                value = value.strip()

                if value:
                    if key == 'Identifier':
                        (number, description) = value.split(' ', 1)

                        if number.startswith('0x'):
                            number = int(number, 16)
                        elif number.isdigit():
                            number = int(number)

                        if 'no module' in description:
                            description = '(empty)'

                        data[ifname]['identifier'] = number
                        data[ifname]['identifierDescription'] = description
                    elif key == 'Vendor PN':
                        data[ifname]['vendorPartNumber'] = value
                    elif key == 'Vendor rev':
                        data[ifname]['vendorRev'] = value
                    elif key == 'Vendor name':
                        data[ifname]['vendorName'] = value
                    elif key == 'Vendor SN':
                        data[ifname]['vendorSerialNumber'] = value

    return data


def show_pluggables(cli):

    if cli.args.get('<interface>'):
        ifnames = [cli.args.get('<interface>'), ]
    else:
        ifnames = get_interfaces()

    data = get_pluggables(ifnames)

    if cli.args.get('json'):
        print json.dumps(data, indent=4)
    else:
        header = ('Interface', 'Identifier', 'Vendor Name', 'Vendor PN', 'Vendor SN', 'Vendor Rev')
        table = []
        for (ifname, value) in data.iteritems():
            table.append((ifname,
                          "0x%02x %s" % (value.get('identifier', 0), value.get('identifierDescription')),
                          value.get('vendorName'),
                          value.get('vendorPartNumber'),
                          value.get('vendorSerialNumber'),
                          value.get('vendorRev')))
        print '\n' + tabulate(tru(table), header, tablefmt="simple") + '\n\n'


def show_interface_detail(ifname):
    """
    ethtool swp1
    ethtool -m swp1
    ethtool -S swp1
    """
    results = []
    try:
        output = check_output(['/sbin/ethtool', ifname], stderr=STDOUT)
        results.extend(['Ethtool',
                        '-------'])
        results.extend(output.splitlines())
        results.append('')
    except Exception:
        pass

    try:
        output = check_output(['/sbin/ethtool', '-m', ifname], stderr=STDOUT)
        results.extend(['Ethtool Pluggables',
                        '-------------------'])
        results.extend(output.splitlines())
        results.append('')
    except Exception:
        pass

    try:
        output = check_output(['/sbin/ethtool', '-S', ifname], stderr=STDOUT)
        results.extend(['Ethtool Statistics',
                        '------------------'])
        results.extend(output.splitlines())
        results.append('')
    except Exception:
        pass

    print '\n'.join(results)


def get_ifname_to_master():
    """
    Example output:

54: swp52: <BROADCAST,MULTICAST> mtu 1500 qdisc pfifo_fast state DOWN mode DEFAULT group default qlen 1000\    link/ether 00:e0:ec:27:68:e6 brd ff:ff:ff:ff:ff:ff
55: swp53: <BROADCAST,MULTICAST> mtu 1500 qdisc pfifo_fast state DOWN mode DEFAULT group default qlen 1000\    link/ether 00:e0:ec:27:68:ea brd ff:ff:ff:ff:ff:ff
56: swp54: <BROADCAST,MULTICAST> mtu 1500 qdisc pfifo_fast state DOWN mode DEFAULT group default qlen 1000\    link/ether 00:e0:ec:27:68:ee brd ff:ff:ff:ff:ff:ff
370: vrf1001: <NOARP,MASTER,UP,LOWER_UP> mtu 65536 qdisc pfifo_fast state UP mode DEFAULT group default qlen 1000\    link/ether 4a:2d:c6:02:a9:71 brd ff:ff:ff:ff:ff:ff
371: swp4.1001@swp4: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue master vrf1001 state UP mode DEFAULT group default \    link/ether 00:e0:ec:27:68:ad brd ff:ff:ff:ff:ff:ff
372: swp3.1001@swp3: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue master vrf1001 state UP mode DEFAULT group default \    link/ether 00:e0:ec:27:68:ac brd ff:ff:ff:ff:ff:ff
    """
    result = {}
    output = check_output(['/bin/ip', '-o', 'link', 'show'])

    for line in output.splitlines():
        re_master = re.search('^\d+:\s+(\S+):.*master\s+(\S+)', line)

        if re_master:
            ifname = re_master.group(1)

            # For swp3.1057@swp3: keep the "swp3.1057'
            if '@' in ifname:
                ifname = ifname.split('@')[0]

            result[ifname] =  re_master.group(2)

    return result


def get_bgp_peers():
    """
    Return a set of all bgp peers
    """
    try:
        data = json.loads(check_output(['/usr/bin/vtysh', '-c', 'show ip bgp neighbor json']))
        return set(data.keys())
    except Exception:
        return set()


class ConfigWrapper(object):

    def __init__(self, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs

    def get_managed_files(self):
        return [ISC_DHCP_RELAY4, ISC_DHCP_RELAY6, ISC_DHCP_SERVER4, ISC_DHCP_SERVER6, PORTS_CONF, ETC_HOSTNAME, ETC_HOSTS]


    def get_service_cmd_for_file(self, filename):
        cmds = []

        if filename == ISC_DHCP_RELAY4:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcrelay.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcrelay.service'))

        elif filename == ISC_DHCP_RELAY6:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcrelay6.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcrelay6.service'))

        elif filename == ISC_DHCP_SERVER4:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcpd.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcpd.service'))

        elif filename == ISC_DHCP_SERVER6:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcpd6.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcpd6.service'))

        elif filename == PORTS_CONF:
            cmds.append(('/bin/systemctl', 'reset-failed', 'switchd.service'))
            cmds.append(('/bin/systemctl', 'restart', 'switchd.service'))

        elif filename == ETC_HOSTNAME:
            new_hostname = None

            with open(ETC_HOSTNAME) as f:
                for line in f.readlines():
                    line = line.strip()

                    if line:
                        new_hostname = line
                        break

            if new_hostname:
                cmds.append(('hostname', new_hostname))
                cmds.append(('hostnamectl', 'set-hostname', new_hostname))

        return cmds


    def abort_pending(self):
        if exists('/var/run/nclu/netmisc/'):
            for filename in os.listdir('/var/run/nclu/netmisc/'):
                full_filename = os.path.join('/var/run/nclu/netmisc/', filename)

                if os.path.isdir(full_filename):
                    shutil.rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def del_all(self, cli):
        """
        Remove all DHCP configuration file INTF_CMD, INTERFACES, OPTIONS, and SERVERS entries, if any.

        For now we are not going to alter ports.conf as part of "net del all".
        """
        # TODO - Move all this DHCP code into its own wrapper.

        # WARNING - The deleted values must be set to an empty string with double quotes because regular
        # expressions elsewhere in this file, in code related to add|del, specifically look for quotes.
        # Setting a key to nothing (i.e. SERVERS=<blank>) will break things.

        if exists(ISC_DHCP_RELAY4) or exists(ISC_DHCP_RELAY_SCRATCHPAD4):
            if not exists(ISC_DHCP_RELAY_SCRATCHPAD4):
                copyfile(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD4, 'SERVERS', '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD4, 'INTF_CMD', '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD4, 'OPTIONS', '""', True)

        if exists(ISC_DHCP_RELAY6) or exists(ISC_DHCP_RELAY_SCRATCHPAD6):
            if not exists(ISC_DHCP_RELAY_SCRATCHPAD6):
                copyfile(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD6, 'SERVERS', '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD6, 'INTF_CMD', '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD6, 'OPTIONS', '""', True)

        if exists(ISC_DHCP_SERVER4) or exists(ISC_DHCP_SERVER_SCRATCHPAD4):
            if not exists(ISC_DHCP_SERVER_SCRATCHPAD4):
                copyfile(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD4, 'INTERFACES', '""', True)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD4, 'OPTIONS', '""', True)

        if exists(ISC_DHCP_SERVER6) or exists(ISC_DHCP_SERVER_SCRATCHPAD6):
            if not exists(ISC_DHCP_SERVER_SCRATCHPAD6):
                copyfile(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD6, 'INTERFACES', '""', True)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD6, 'OPTIONS', '""', True)

    def commit_pending(self, verbose):
        files_modified = []
        commit_ok = True

        # These shouldn't be here, but better safe than sorry.
        for backup_file in (ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP,
                            ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP,
                            ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP,
                            ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP,
                            PORTS_CONF_SCRATCHPAD_BACKUP,
                            ETC_HOSTS_SCRATCHPAD_BACKUP,
                            ETC_HOSTNAME_SCRATCHPAD_BACKUP):
            if exists(backup_file):
                os.remove(backup_file)

        # TODO - Move all this DHCP code into its own wrapper.

        # DHCP IPv4 relay
        if exists(ISC_DHCP_RELAY_SCRATCHPAD4) and not files_match(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4):

            shutil.copyfile(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP)
            shutil.copyfile(ISC_DHCP_RELAY_SCRATCHPAD4, ISC_DHCP_RELAY4)
            files_modified.append(ISC_DHCP_RELAY4)

            # Service resets
            for cmd in (['/bin/systemctl', 'enable', 'dhcrelay.service'],
                        ['/bin/systemctl', 'reset-failed', 'dhcrelay.service'],
                        ['/bin/systemctl', 'restart', 'dhcrelay.service']):

                if self.show_linux_command:
                    print "exec: {0]".format(' '.join(cmd))

                try:
                    output = check_output(cmd, stderr=STDOUT)
                    if verbose:
                        print output
                except CalledProcessError as e:
                    print "\nWARNING: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)
                    commit_ok = False

                    try:
                        output = check_output(['/bin/systemctl', 'status', 'dhcrelay.service'], stderr=STDOUT)
                        print '\n/bin/systemctl status dhcrelay.service'
                        print '======================================'
                        print output
                    except CalledProcessError as e:
                        pass

        # DHCP IPv4 server
        if exists(ISC_DHCP_SERVER_SCRATCHPAD4) and not files_match(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4):

            shutil.copyfile(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP)
            shutil.copyfile(ISC_DHCP_SERVER_SCRATCHPAD4, ISC_DHCP_SERVER4)
            files_modified.append(ISC_DHCP_SERVER4)

            # Service resets
            for cmd in (['/bin/systemctl', 'enable', 'dhcpd.service'],
                        ['/bin/systemctl', 'reset-failed', 'dhcpd.service'],
                        ['/bin/systemctl', 'restart', 'dhcpd.service']):

                if self.show_linux_command:
                    print "exec: {0]".format(' '.join(cmd))

                try:
                    output = check_output(cmd, stderr=STDOUT)
                    if verbose:
                        print output
                except CalledProcessError as e:
                    print "\nWARNING: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)
                    commit_ok = False

                    try:
                        output = check_output(['/bin/systemctl', 'status', 'dhcpd.service'], stderr=STDOUT)
                        print '\n/bin/systemctl status dhcpd.service'
                        print '======================================'
                        print output
                    except CalledProcessError as e:
                        pass

        # DHCP IPv6 relay
        if exists(ISC_DHCP_RELAY_SCRATCHPAD6) and not files_match(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6):

            shutil.copyfile(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP)
            shutil.copyfile(ISC_DHCP_RELAY_SCRATCHPAD6, ISC_DHCP_RELAY6)
            files_modified.append(ISC_DHCP_RELAY6)

            # Service resets
            for cmd in (['/bin/systemctl', 'enable', 'dhcrelay6.service'],
                        ['/bin/systemctl', 'reset-failed', 'dhcrelay6.service'],
                        ['/bin/systemctl', 'restart', 'dhcrelay6.service']):

                if self.show_linux_command:
                    print "exec: {0}".format(' '.join(cmd))

                try:
                    output = check_output(cmd, stderr=STDOUT)
                    if verbose:
                        print output
                except CalledProcessError as e:
                    print "\nWARNING: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)
                    commit_ok = False

                    try:
                        output = check_output(['/bin/systemctl', 'status', 'dhcrelay6.service'], stderr=STDOUT)
                        print '\n/bin/systemctl status dhcrelay6.service'
                        print '======================================'
                        print output
                    except CalledProcessError as e:
                        pass

        # DHCP IPv6 server
        if exists(ISC_DHCP_SERVER_SCRATCHPAD6) and not files_match(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6):

            shutil.copyfile(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP)
            shutil.copyfile(ISC_DHCP_SERVER_SCRATCHPAD6, ISC_DHCP_SERVER6)
            files_modified.append(ISC_DHCP_SERVER6)

            # Service resets
            for cmd in (['/bin/systemctl', 'enable', 'dhcpd6.service'],
                        ['/bin/systemctl', 'reset-failed', 'dhcpd6.service'],
                        ['/bin/systemctl', 'restart', 'dhcpd6.service']):

                if self.show_linux_command:
                    print "exec: {0]".format(' '.join(cmd))

                try:
                    output = check_output(cmd, stderr=STDOUT)
                    if verbose:
                        print output
                except CalledProcessError as e:
                    print "\nWARNING: '{0}' failed due to\n{1}\n".format(' '.join(cmd), e)
                    commit_ok = False

                    try:
                        output = check_output(['/bin/systemctl', 'status', 'dhcpd6.service'], stderr=STDOUT)
                        print '\n/bin/systemctl status dhcpd6.service'
                        print '======================================'
                        print output
                    except CalledProcessError as e:
                        pass

        # hostname
        if exists(ETC_HOSTS_SCRATCHPAD) and not files_match(ETC_HOSTS, ETC_HOSTS_SCRATCHPAD):
            shutil.copyfile(ETC_HOSTS, ETC_HOSTS_SCRATCHPAD_BACKUP)
            shutil.copyfile(ETC_HOSTS_SCRATCHPAD, ETC_HOSTS)
            files_modified.append(ETC_HOSTS)

        if exists(ETC_HOSTNAME_SCRATCHPAD) and not files_match(ETC_HOSTNAME, ETC_HOSTNAME_SCRATCHPAD):
            shutil.copyfile(ETC_HOSTNAME, ETC_HOSTNAME_SCRATCHPAD_BACKUP)
            shutil.copyfile(ETC_HOSTNAME_SCRATCHPAD, ETC_HOSTNAME)
            files_modified.append(ETC_HOSTNAME)

        if exists(EXECLIST):
            with open(EXECLIST) as f:
                for line in f.readlines():
                    cmd = shlex.split(line.strip())

                    if self.show_linux_command:
                        print "exec: %s" % ' '.join(cmd)

                    try:
                        output = check_output(cmd, stderr=STDOUT)
                        if verbose:
                            print output
                    except CalledProcessError as e:
                        print "\nWARNING: '%s' failed due to\n%s\n" % (' '.join(cmd), e)
                        commit_ok = False

            os.remove(EXECLIST)

        # /etc/cumulus/ports.conf
        if exists(PORTS_CONF_SCRATCHPAD) and not files_match(PORTS_CONF, PORTS_CONF_SCRATCHPAD):
            shutil.copyfile(PORTS_CONF, PORTS_CONF_SCRATCHPAD_BACKUP)
            shutil.copyfile(PORTS_CONF_SCRATCHPAD, PORTS_CONF)
            files_modified.append(PORTS_CONF)

            try:
                output = check_output(['/bin/systemctl', 'reset-failed', 'switchd'])
                if verbose:
                    print output

                output = check_output(['/bin/systemctl', 'restart', 'switchd'])
                if verbose:
                    print output

            except Exception:
                print "ERROR: Failed to restart switchd\n%s" % output
                commit_ok = False

        if commit_ok:
            # The commit succeded.  Remove the scratchpad files.
            for filename in (ISC_DHCP_RELAY_SCRATCHPAD4,
                             ISC_DHCP_SERVER_SCRATCHPAD4,
                             ISC_DHCP_RELAY_SCRATCHPAD6,
                             ISC_DHCP_SERVER_SCRATCHPAD6,
                             PORTS_CONF_SCRATCHPAD,
                             ETC_HOSTS_SCRATCHPAD,
                             ETC_HOSTNAME_SCRATCHPAD):
                if exists(filename):
                    os.remove(filename)

        else:
            # The commit failed.  Revert everything.
            files_modified = []

            # DHCP reverts

            if exists(ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP):
                shutil.copyfile(ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP, ISC_DHCP_RELAY4)
                try:
                    check_output(['/bin/systemctl', 'reset-failed', 'dhcrelay.service'])
                    check_output(['/bin/systemctl', 'restart', 'dhcrelay.service'])
                except Exception:
                    print "ERROR: Failed to restart dhcrelay (post revert)\n%s" % output

            if exists(ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP):
                shutil.copyfile(ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP, ISC_DHCP_SERVER4)
                try:
                    check_output(['/bin/systemctl', 'reset-failed', 'dhcpd.service'])
                    check_output(['/bin/systemctl', 'restart', 'dhcpd.service'])
                except Exception:
                    print "ERROR: Failed to restart dhcpd (post revert)\n%s" % output

            if exists(ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP):
                shutil.copyfile(ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP, ISC_DHCP_RELAY6)
                try:
                    check_output(['/bin/systemctl', 'reset-failed', 'dhcrelay6.service'])
                    check_output(['/bin/systemctl', 'restart', 'dhcrelay6.service'])
                except Exception:
                    print "ERROR: Failed to restart dhcrelay6 (post revert)\n%s" % output

            if exists(ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP):
                shutil.copyfile(ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP, ISC_DHCP_SERVER6)
                try:
                    check_output(['/bin/systemctl', 'reset-failed', 'dhcpd6.service'])
                    check_output(['/bin/systemctl', 'restart', 'dhcpd6.service'])
                except Exception:
                    print "ERROR: Failed to restart dhcpd6 (post revert)\n%s" % output

            # ports.conf reverts
            if exists(PORTS_CONF_SCRATCHPAD_BACKUP):
                shutil.copyfile(PORTS_CONF_SCRATCHPAD_BACKUP, PORTS_CONF)
                try:
                    check_output(['/bin/systemctl', 'reset-failed', 'switchd'])
                    check_output(['/bin/systemctl', 'restart', 'switchd'])
                except Exception:
                    print "ERROR: Failed to restart switchd (post revert)\n%s" % output

            if exists(ETC_HOSTS_SCRATCHPAD_BACKUP):
                shutil.copyfile(ETC_HOSTS_SCRATCHPAD_BACKUP, ETC_HOSTS)

            if exists(ETC_HOSTNAME_SCRATCHPAD_BACKUP):
                shutil.copyfile(ETC_HOSTNAME_SCRATCHPAD_BACKUP, ETC_HOSTNAME)

        for filename in (ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP,
                         PORTS_CONF_SCRATCHPAD_BACKUP,
                         ETC_HOSTS_SCRATCHPAD_BACKUP,
                         ETC_HOSTNAME_SCRATCHPAD_BACKUP):
            if exists(filename):
                os.remove(filename)

        return (commit_ok, files_modified)


    def get_configured_hostname(self):
        """
        Return the hostname as a string, or return None.
        """
        if exists(ETC_HOSTNAME):
            with open(ETC_HOSTNAME) as f:
                for line in f:
                    line = line.strip()

                    if line:
                        return line
        return None


    def get_dhcp_configuration(self):
        dhcp_relay_servers = []
        dhcp_relay_interfaces = []
        dhcp_servers = []

        dhcp_relay6_servers = []
        dhcp_relay6_interfaces = []
        dhcp_servers6 = []

        relay_key_value_pair_re = re.compile(r'''^\s*(SERVERS|INTF_CMD)\s*=\s*"([^"]*)"''', re.IGNORECASE | re.MULTILINE)
        server_key_value_pair_re = re.compile(r'''^\s*INTERFACES\s*=\s*"([^"]*)"''', re.IGNORECASE | re.MULTILINE)

        # IPv4 and IPv6 relay
        for dhcp_file in [ISC_DHCP_RELAY4, ISC_DHCP_RELAY6]:
            if not exists(dhcp_file):
                continue

            with open(dhcp_file) as f:
                config_text = f.read()

            key_value_pairs = relay_key_value_pair_re.findall(config_text)

            for key, value in key_value_pairs:
                if key.upper() == 'SERVERS':
                    if dhcp_file.endswith('6'):
                        # relay6 server
                        interfaces = value.replace('-u', '').split()

                        try:
                            interfaces_as_glob = ifnames_to_glob(interfaces)
                        except (RuntimeError, ValueError) as e:
                            # Could not globify the interfaces.
                            log.warning("(get_dhcp_configuration) '{0}' raised an exception in ifnames_to_glob: {1}.".format(' '.join(interfaces), e))
                            interfaces_as_glob = None

                        if interfaces_as_glob:
                            dhcp_relay6_servers.append(interfaces_as_glob)
                        else:
                            dhcp_relay6_servers.extend(interfaces)
                    else:
                        # relay server
                        dhcp_relay_servers.extend(value.split())

                elif key.upper() == 'INTF_CMD':
                    if dhcp_file.endswith('6'):
                        # relay6 interface
                        interfaces = value.replace('-l', '').strip().split()
                    else:
                        # relay interface
                        interfaces = value.replace('-i', '').strip().split()

                    try:
                        interfaces_as_glob = ifnames_to_glob(interfaces)
                    except (RuntimeError, ValueError) as e:
                        # Could not globify the interfaces.
                        log.warning("(get_dhcp_configuration) '{0}' raised an exception in ifnames_to_glob: {1}.".format(str(interfaces), e))
                        interfaces_as_glob = None

                    if dhcp_file.endswith('6'):
                        if interfaces_as_glob:
                            dhcp_relay6_interfaces.append(interfaces_as_glob)
                        else:
                            dhcp_relay6_interfaces.extend(interfaces)
                    else:
                        if interfaces_as_glob:
                            dhcp_relay_interfaces.append(interfaces_as_glob)
                        else:
                            dhcp_relay_interfaces.extend(interfaces)

        # IPv4 and IPv6 server
        for dhcp_file in [ISC_DHCP_SERVER4, ISC_DHCP_SERVER6]:
            if not exists(dhcp_file):
                continue

            with open(dhcp_file) as f:
                config_text = f.read()

            servers = server_key_value_pair_re.findall(config_text)

            for value in servers:
                if dhcp_file.endswith('6'):
                    # server6
                    dhcp_servers6.extend(value.split())
                else:
                    # server
                    dhcp_servers.extend(value.split())

        if dhcp_relay_servers:
            dhcp_relay_servers = sort_for_humans(dhcp_relay_servers)

        if dhcp_relay_interfaces:
            dhcp_relay_interfaces = sort_for_humans(dhcp_relay_interfaces)

        if dhcp_servers:
            dhcp_servers = sort_for_humans(dhcp_servers)

        if dhcp_relay6_servers:
            dhcp_relay6_servers = sort_for_humans(dhcp_relay6_servers)

        if dhcp_relay6_interfaces:
            dhcp_relay6_interfaces = sort_for_humans(dhcp_relay6_interfaces)

        if dhcp_servers6:
            dhcp_servers6 = sort_for_humans(dhcp_servers6)

        return (dhcp_relay_servers, dhcp_relay_interfaces, dhcp_servers, dhcp_relay6_servers, dhcp_relay6_interfaces, dhcp_servers6)


    def get_ports_conf(self):
        ports = []

        if exists(PORTS_CONF):
            with open(PORTS_CONF) as f:
                for line in f.readlines():
                    line = line.strip()
                    re_line = re.match(r"""^(\d+)\s*=\s*(\w+)$""", line)

                    if re_line:
                        port_number = int(re_line.group(1))
                        port_setting = re_line.group(2)

                        if '4x' in port_setting:
                            port_setting = '4x'
                        elif '2x' in port_setting:
                            port_setting = '2x'
                        elif '/4' in port_setting:
                            port_setting = '/4'
                        elif 'disabled' in port_setting or 'disable' in port_setting:
                            port_setting = 'disabled'
                        else:
                            port_setting = '1x'

                        ports.append((port_number, port_setting))

        return ports


    def show_pending(self, print_errors=True):
        diff_pending = PENDING_COMMIT_NOCHANGE

        if exists(ISC_DHCP_RELAY_SCRATCHPAD4):
            if print_diff(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE

        if exists(ISC_DHCP_RELAY_SCRATCHPAD6):
            if print_diff(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE

        if exists(ISC_DHCP_SERVER_SCRATCHPAD4):
            if print_diff(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE

        if exists(ISC_DHCP_SERVER_SCRATCHPAD6):
            if print_diff(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE

        if exists(ETC_HOSTS_SCRATCHPAD):
            if print_diff(ETC_HOSTS, ETC_HOSTS_SCRATCHPAD, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE

        if exists(ETC_HOSTNAME_SCRATCHPAD):
            if print_diff(ETC_HOSTNAME, ETC_HOSTNAME_SCRATCHPAD, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE

        if exists(PORTS_CONF_SCRATCHPAD):
            if print_diff(PORTS_CONF, PORTS_CONF_SCRATCHPAD, self.color_diffs):
                diff_pending = PENDING_COMMIT_CHANGE
                print """\033[91m\n
********************************************************************
********************************************************************
NOTE: 'net commit' will run 'systemctl restart switchd' to apply the
change to /etc/cumulus/ports.conf. This can take a few seconds.
********************************************************************
********************************************************************
\033[0m
"""

        return diff_pending


    def show_config_files(self, user_may_edit):

        for filename in (ETC_HOSTNAME,
                         ETC_HOSTS,
                         ISC_DHCP_RELAY4,
                         ISC_DHCP_RELAY6,
                         ISC_DHCP_SERVER4,
                         ISC_DHCP_SERVER6,
                         PORTS_CONF):
            if exists(filename) and os.path.getsize(filename) > 0:
                print filename
                print '=' * len(filename)

                with open(filename) as f:
                    print f.read()  # Print the contents of the file.
                print ''


    def show_config_commands(self, user_may_edit):
        commands = []
        hostname = self.get_configured_hostname()

        if hostname is not None:
            commands.append('net add hostname {0}'.format(hostname))

        (dhcp_relay_servers, dhcp_relay_interfaces, dhcp_servers, dhcp_relay6_servers, dhcp_relay6_interfaces, dhcp_servers6) = self.get_dhcp_configuration()

        # DHCP IPv4
        commands.extend(['net add dhcp relay server {0}'.format(server) for server in dhcp_relay_servers])
        commands.extend(['net add dhcp relay interface {0}'.format(interface) for interface in dhcp_relay_interfaces])
        commands.extend(['net add dhcp server {0}'.format(server) for server in dhcp_servers])

        # DHCP IPv6
        commands.extend(['net add dhcp relay6 server {0}'.format(server) for server in dhcp_relay6_servers])
        commands.extend(['net add dhcp relay6 interface {0}'.format(interface) for interface in dhcp_relay6_interfaces])
        commands.extend(['net add dhcp server6 {0}'.format(server) for server in dhcp_servers6])

        ports_by_setting = {}
        for (port_number, port_setting) in self.get_ports_conf():
            if port_setting not in ports_by_setting:
                ports_by_setting[port_setting] = []
            ports_by_setting[port_setting].append(port_number)

        for port_setting in sort_for_humans(ports_by_setting.keys()):
            port_numbers = ports_by_setting[port_setting]
            port_numbers_glob = numbers_to_glob(port_numbers)
            commands.append('net add interface swp{0} breakout {1}'.format(port_numbers_glob, port_setting))

        return (commands, [])


    def show_config_summary(self, user_may_edit, summary, show_dhcp=False, print_json=False):
        """
        net show configuration [dhcp] [json]

        Populate the 'summary' dictionary with DHCP information.  If the configuration for
        DHCP was not specifically requested, also provide information about the hostname.

        summary - A dictionary to hold results (call by reference)

        TODO - This should just return a value and not require passing a
            dictionary to hold results.
        """

        if not isinstance(summary, dict):
            raise TypeError("'summary' must be a dictionary.")

        if not show_dhcp:
            # net show configuration [json]
            hostname = self.get_configured_hostname()

            if hostname:
                summary["hostname %s" % hostname] = []

        (dhcp_relay_servers, dhcp_relay_interfaces, dhcp_servers, dhcp_relay6_servers, dhcp_relay6_interfaces, dhcp_servers6) = self.get_dhcp_configuration()

        # IPv4 relay
        if dhcp_relay_servers or dhcp_relay_interfaces:
            summary['dhcp relay'] = OrderedDict()

            if dhcp_relay_servers:
                summary['dhcp relay']['SERVERS'] = dhcp_relay_servers

            if dhcp_relay_interfaces:
                summary['dhcp relay']['INTF_CMD'] = dhcp_relay_interfaces

        # IPv4 server
        if dhcp_servers:
            summary['dhcp server'] = OrderedDict()
            summary['dhcp server']['INTERFACES'] = dhcp_servers

        # IPv6 relay
        if dhcp_relay6_servers or dhcp_relay6_interfaces:
            summary['dhcp relay6'] = OrderedDict()

            if dhcp_relay6_servers:
                summary['dhcp relay6']['SERVERS'] = dhcp_relay6_servers

            if dhcp_relay6_interfaces:
                summary['dhcp relay6']['INTF_CMD'] = dhcp_relay6_interfaces

        # IPv6 server
        if dhcp_servers6:
            summary['dhcp server6'] = OrderedDict()
            summary['dhcp server6']['INTERFACES'] = dhcp_servers6

        if print_json:
            print json.dumps(summary, indent=4)


    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user.
        """
        args = cli.args
        command_ok = True
        features = {}  # Used by some "show" commands to specify a subset of available information

        if args.get('show'):

            if args.get('hostname'):
                # net show hostname
                sys.stdout.write(check_output(('/bin/hostname',)))

            elif args.get('version'):
                print "NCLU_VERSION=%s" % NCLU_VERSION
                with open('/etc/lsb-release') as f:
                    for line in f.readlines():
                        sys.stdout.write(line)

            elif args.get('pluggables'):
                show_pluggables(cli)

            # net show vrf
            elif args.get('vrf'):
                sys.stdout.write(check_output(('/usr/bin/vrf', 'list')))

            else:
                vendor = vendor_check()
                show = None

                # Execute the appropriate "show" subcommand.
                if interface_related(cli):
                    # net show access|bridges|bonds|bondmems|mgmt|l2|l3|phy|trunks|interface
                    ifname = args.get('<interface>')  # A unicode string or None

                    # Work-around CM-15642, which cannot be resolved elegantly without changing the command syntax.
                    if ifname == 'json':
                        args['<interface>'] = None
                        args['json'] = True
                        ifname = None

                    # Show a specific interface "net show int swp1"
                    if ifname is not None:
                        ifnames = ifname_expand_glob(ifname) if ifname_is_glob(ifname) else [ifname]  # A list of strings

                        for ifacename in ifnames:
                            cli.args['<interface>'] = ifacename
                            show = CumulusShowInterfaces(cli) if vendor == 'cumulus' else ShowInterfaces(cli)

                            try:
                                print show.run()

                                if args.get('detail'):
                                    show_interface_detail(ifacename)

                            except Exception as e:
                                log.exception(e)
                                print "ERROR: show command failed for %s" % ifacename
                                print e

                        show = None

                    # "net show int [json]" --> Show a list of interfaces and their configuration data
                    else:
                        if args['json']:
                            # The user wants JSON output.  Use the legacy implementation from netshowlib,
                            # which might result in a significant time delay.
                            show = CumulusShowInterfaces(cli) if vendor == 'cumulus' else ShowInterfaces(cli)

                            # Use decode_dict to convert unicode to strings
                            with Capturing() as show_run_output:
                                data = json.loads(show.run(), object_hook=decode_dict)

                            # If the LLDP service is not running netshow will display an
                            #
                            #   "ERROR: The lldpd service is not running"
                            #
                            # message for every interface. To clean this up, grab all of
                            # the output via a 'with Capturing' and then only display the
                            # unique lines of output.
                            if show_run_output:
                                unique_output = []
                                for line in show_run_output:
                                    if line not in unique_output:
                                        unique_output.append(line)
                                print '\n'.join(unique_output)

                            # Add the 'master' for each interface to data
                            for (ifname, master) in get_ifname_to_master().items():
                                if ifname in data:
                                    data[ifname]['master'] = master

                           # Update port_category for any interface that uses BGP unnumbered
                            bgp_peers = get_bgp_peers()

                            for ifname in data.keys():
                                if ifname in bgp_peers:
                                    data[ifname]['port_category'] = 'BGP Unnumbered'

                            print json.dumps(data, indent=4)

                        else:
                            # The user wants a non-JSON output.  This implementation avoids the large time delay
                            # by directly calling "netshow interface" and augmenting that output.

                            if args['mac']:
                                # The command includes the MAC keyword.
                                tabulate_header = ('', 'Name', 'Master', 'MAC', 'Speed', 'MTU', 'Mode',
                                                   'Remote Host', 'Remote Port', 'Summary')
                            else:
                                tabulate_header = ('', 'Name', 'Master', 'Speed', 'MTU', 'Mode',
                                                   'Remote Host', 'Remote Port', 'Summary')
                            tabulate_data = []
                            lldp = lldp_cacheinfo()  # Dictionary with interface names as keys and dictionaries as values.

                            # Start screen scraping data from the output of "netshow interface", which completes quickly.
                            netshow_output = (CumulusShowInterfaces(cli) if (vendor == 'cumulus') else ShowInterfaces(cli)).run()

                            # Replace two or more spaces present between parentheses so that the column
                            # variables get populated correctly when we split the netshow output by two spaces below.
                            strings_in_parentheses = re.findall(r"""\(.+?\)""", netshow_output)
                            space_sub_re = re.compile(r"""( ){2,}""")
                            for spaced_str in strings_in_parentheses:
                                trimmed_str = space_sub_re.sub(' ', spaced_str)
                                if trimmed_str != spaced_str:
                                    netshow_output = netshow_output.replace(spaced_str, trimmed_str)

                            regex = r"""^((?:UP|DN|ADMDN|DORMANT).+?)(?=(?:^(?:UP|DN|ADMDN|DORMANT))|\Z)"""
                            entry_re = re.compile(regex, re.MULTILINE | re.DOTALL)
                            entries = entry_re.findall(netshow_output)

                            # Used in the below loop to trim the summary information.
                            summary_re = re.compile(r"""(.+?)(Tagged Members|802\.1q|Untagged VLAN|Bridge Membership)""")

                            use_netshow_output = False
                            # Augment the netshow output with additional columns, and add each modified entry to tabulate_data.
                            for entry in entries:
                                entry_split = [x.strip() for x in entry.split('  ') if x]

                                num_columns = 6 if args['mac'] else 5
                                if len(entry_split) == 5 and args['mac']:
                                    # MAC is empty but num_columns is 6.
                                    # If the command includes the mac keyword but there is no MAC, set it to a blank string.
                                    mac = ''
                                    num_columns = 5
                                if len(entry_split) == num_columns:
                                    # The summary column is blank.
                                    if num_columns == 6:
                                        try:
                                            linkstate, name, mac, speed, mtu, mode = entry_split
                                            # Make sure the MAC Address is in the correct format.
                                            assert re.match(r"""(([a-fA-F0-9]){2}:){5}([a-fA-F0-9]){2}""", mac) is not None
                                        except Exception as e:
                                            log.exception(e)
                                            use_netshow_output = True
                                            break
                                    else:
                                        try:
                                            linkstate, name, speed, mtu, mode = entry_split
                                        except Exception as e:
                                            log.exception(e)
                                            use_netshow_output = True
                                            break
                                    summary = ''
                                else:
                                    # The summary column is not blank.
                                    if num_columns == 6:
                                        try:
                                            linkstate, name, mac, speed, mtu, mode = entry_split[:6]
                                            # Make sure the MAC Address is in the correct format.
                                            assert re.match(r"""(([a-fA-F0-9]){2}:){5}([a-fA-F0-9]){2}""", mac) is not None
                                        except Exception as e:
                                            log.exception(e)
                                            use_netshow_output = True
                                            break
                                    else:
                                        try:
                                            linkstate, name, speed, mtu, mode = entry_split[:5]
                                        except Exception as e:
                                            log.exception(e)
                                            use_netshow_output = True
                                            break
                                    summary = ' '.join(entry_split[num_columns:])

                                    # Abridge the summary, if necessary.
                                    summary_match = summary_re.match(summary)
                                    if summary_match is not None:
                                        # Strip undesired information from the end of the string.
                                        summary = summary_match.group(1).strip()

                                    # Replace runs of blanks with single spaces.
                                    summary = re.sub(r"""\s{2,}""", ' ', summary)

                                # Make sure the MTU only consists of digits.
                                assert re.match(r"""\d+""", mtu) is not None

                                # Abridge the name, if necessary.  For example, "VLAN3000 (VXLAN GW)" should
                                # be shortened to "VLAN3000."
                                if ' ' in name:
                                    name = name.split()[0]

                                # Derive the bond master, if any.
                                master = read_symlink('/sys/class/net/{0}/master'.format(name))  # A string or None
                                master = master if master else "None"

                                # Derive the remote host and remote port.
                                if lldp is not None and lldp.get(name) is not None:
                                    remote_host = lldp[name][0].get('adj_hostname')
                                    remote_port = lldp[name][0].get('adj_port')

                                    # The keys (above) can be present but have a value of None, which would
                                    # raise TypeError in the next block.  Preempt this condition.
                                    if remote_host is None:
                                        remote_host = ''
                                    if remote_port is None:
                                        remote_port = ''

                                    if '.' in remote_host:
                                        remote_host = remote_host.split('.')[0]
                                else:
                                    remote_host = ''
                                    remote_port = ''

                                if args['mac']:
                                    to_append = (linkstate, name, master, mac, speed, mtu, mode, remote_host, remote_port, summary)
                                else:
                                    to_append = (linkstate, name, master, speed, mtu, mode, remote_host, remote_port, summary)
                                tabulate_data.append(to_append)
                            if use_netshow_output:
                                cmd = ' '.join(cli.argv_expanded)
                                log.error("Defaulting to netshow output. Error in command:", cmd)
                                print netshow_output
                            else:
                                print '\n' + tabulate(tru(tabulate_data), tabulate_header, tablefmt="simple") + '\n\n'

                    show = None

                elif cli.get('system'):
                    # net show system
                    if vendor == 'cumulus':
                        show = CumulusShowSystem(cli)
                    else:
                        show = ShowSystem(cli)

                elif cli.get('lldp'):
                    # net show lldp
                    show = CumulusShowNeighbors(cli) if vendor == 'cumulus' else ShowNeighbors(cli)
                    features['lldp'] = 'linux'

                elif cli.get('counters'):
                    # net show counters
                    cmd = ['/usr/cumulus/bin/cl-netstat']

                    if args['json'] is not None:
                        cmd.append('-j')

                    try:
                        output = check_output(cmd, stderr=STDOUT)
                        print output
                    except Exception as e:
                        print "ERROR: '%s' failed" % ' '.join(cmd)
                        print e

                if show is not None:
                    # Did the user specify interfaces?
                    ifname_cl = args['<interface>']  # A unicode string or None

                    if ifname_cl is None:
                        ifnames = None
                    else:
                        # The user specified one or more interfaces.
                        ifnames = ifname_expand_glob(ifname_cl) if ifname_is_glob(ifname_cl) else [ifname_cl]  # A list of strings

                    if ifnames is not None and isinstance(show, ShowNeighbors) and cli.get('json') is None:
                        # net show lldp <interface(s)> (not json)

                        # This regex is used to strip the header text from the output of lldpctl.
                        lldp_header_re = re.compile(r"""(?:^-{3,}\s*?\n)|(?:^LLDP neighbors:\s*?\n)""", re.MULTILINE)

                        # Display the output of lldpctl for each interface.
                        for ifacename in ifnames:
                            try:
                                print lldp_header_re.sub('', exec_command('/usr/sbin/lldpctl ' + ifacename, utf_decode=False), 4)
                            except Exception:
                                print "ERROR: show command failed for {0}.".format(ifacename)
                    else:
                        # Execute 'netshow' with appropriate arguments.
                        try:
                            if isinstance(show, ShowNeighbors):
                                print show.run(features, ifnames)  # ex. "net show lldp" (no specified interface)
                            elif (isinstance(show, ShowInterfaces) and (ifnames is None)) or isinstance(show, ShowSystem):
                                print show.run()                   # ex. "net show interface all", "net show system"
                            else:
                                print show.run(ifnames)
                        except Exception as e:
                            log.exception(e)
                            print "ERROR: show command failed."
                            print e

        elif args.get('clear'):

            if args.get('counters'):
                cmd = ['/usr/cumulus/bin/cl-netstat', '-c']
                try:
                    check_output(cmd, stderr=STDOUT)
                except Exception as e:
                    print "'%s' failed" % ' '.join(cmd)
                    print e

        # net add|del
        else:
            if args.get('dhcp'):
                if args.get('relay') or args.get('relay6'):
                    command_ok = dhcp_relay_edit(args)   # net add|del dhcp relay|relay6 ...
                elif args.get('server') or args.get('server6'):
                    command_ok = dhcp_server_edit(args)  # net add|del dhcp server|server6 ...

            elif args.get('hostname'):
                command_ok = hostname_edit(args.get('<text>'))

            elif args.get('breakout'):
                command_ok = ports_conf_edit(args)

        return command_ok
