# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    # System level commands
    run run [<ipv6>|all] license-install path <text>
    run run [<ipv6>|all] reboot
    run run [<ipv6>|all] nos-install path <text>
    run run [<ipv6>|all] daemon restart <text>
    run run [<ipv6>|all] cl-support

Options:
    run : Execute system level commands immediately.
    all : Execute on all endpoints in a CLOS.
    <ipv6> : IPV6 address of an endpoint in a CLOS.
    license-install : Install Cumulus OS license.
    path : Path on the local system or a URL where license file is present
    reboot : Reboot endpoint in a CLOS
    nos-install : Install a new Cumulus Linux binary image (Destructive!!!! Please backup important files)
    daemon: Cumulus Linux Daemon related operations
    restart : Systemd restart of a service
"""
from nclu import (
    ConfigWrapperBase,
    ConfigFile,
    PENDING_COMMIT_NOCHANGE,
    PENDING_COMMIT_CHANGE,
    PENDING_COMMIT_ERROR,
    ConfigFileSingleLine,
    print_diff
)

from collections import OrderedDict
from subprocess import check_output, CalledProcessError
from tabulate import tabulate
import json
import logging
import logging.handlers
import os
import re
import shutil
import glob
import pprint
import copy
import subprocess
import httplib
import json
import socket
import cumulus.chassis
import ssl

log = logging.getLogger('netd')


def tab_complete_run(args, ended_with_space, last_argv):
    return []


def parse_v6_hosts():
    chassis_obj = chassis()
    v6hosts = ''
 
    if chassis_obj:
        v6hosts = chassis().GetV6Hosts().split('\n')
    hosts_dict = {}

    for host_entry in v6hosts:
        if 'bmc' not in host_entry and host_entry != '':
            host_entry_list = host_entry.split()
            hosts_dict[host_entry_list[1].split('-')[0]] = host_entry_list[0]
    return hosts_dict


def chassis():
    chassis_obj = None

    try:
        chassis_obj = cumulus.chassis.probe()
    except RuntimeError as e:
        print "Switch is not a chassis"
    return chassis_obj


class CliBase:
    hosts = {key: value for key, value in parse_v6_hosts().iteritems()}

    def hostname(self):
        return chassis().GetModuleName()

    def linklocalscope(self):
        return chassis().GetMgmtIf()

    def exec_cmd_on_localhost(self, cmd):
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            print e.output
            return False

        return True

    def exec_cmd_on_endpoint(self, post_request, ipv6addr=None):
        header = {"Content-Type": "application/json"}
        url = 'https://[' + ipv6addr + ']' + ':8080/nclu/v1/rpc'
        response = None

        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        context.verify_mode = ssl.CERT_REQUIRED
        context.load_verify_locations(cafile='/etc/ssl/certs/cumulus.pem')
        httpconn = httplib.HTTPSConnection(ipv6addr + '%' + self.linklocalscope(), 8080, timeout=300, context=context)

        try:
            httpconn.request("POST", url, json.dumps(post_request), headers=header)
            response = httpconn.getresponse()
        except Exception as e:
            print e.args[1]
            return False
        finally:
            if response:
                print response.read()
            if httpconn:
                httpconn.close()

        return True

    def exec_cmd_entire_chassis(self, post_request, cmd):
        ret_val = True
        hosts_status = {}
        header = {"Content-Type": "application/json"}

        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        context.verify_mode = ssl.CERT_REQUIRED
        context.load_verify_locations(cafile='/etc/ssl/certs/cumulus.pem')

        for host in self.hosts.keys():
            hosts_status[host] = {'status': '',
                                  'response_str': '',
                                  'conn_obj': None, }

        for host, addr in self.hosts.iteritems():
            if host != self.hostname():
                try:
                    hosts_status[host]['conn_obj'] = httplib.HTTPSConnection(addr + '%' + self.linklocalscope(),
                                                                            8080, timeout=300, context=context)
                    conn = hosts_status[host]['conn_obj']
                    conn.request("POST", 'https://[' + addr + ']' + ':8080/nclu/v1/rpc',
                                 json.dumps(post_request), headers=header)
                except Exception as e:
                    hosts_status[host]['conn_obj'] = None
                    hosts_status[host]['status'] = 'FAIL!'
                    ret_val = False

        for host, data in hosts_status.iteritems():
            print '*' * 80
            print '%s' % host
            response_str = ''

            if host != self.hostname():
                conn = data['conn_obj']
                if conn:
                    response = conn.getresponse()
                    if response.status != 200:
                        data['status'] = 'FAIL!'
                        ret_val = False
                    response_str = response.read()
                if data['status'] == 'FAIL!':
                    print 'FAIL!'
                elif response_str == '\n':
                    print 'SUCCESS!'
                else:
                    print '%s' % response_str
            else:
                # If none of the hosts have failed execute on
                # local host else mark local host as failed.
                if ret_val:
                    if self.exec_cmd_on_localhost(cmd):
                        print 'SUCCESS!'
            print '*' * 80

        return ret_val

    def exec_cmd(self, ipv6addr, is_all, post_request, cmd):

        # execute on local machine only
        if ipv6addr is None and not is_all:
            return self.exec_cmd_on_localhost(cmd)
        # execute on a single endpoint
        elif ipv6addr:
            return self.exec_cmd_on_endpoint(post_request, ipv6addr)
        # execute on all
        else:
            return self.exec_cmd_entire_chassis(post_request, cmd)


class LicenseInstall(CliBase):
    def __call__(self, cli_dict):
        license_path = cli_dict['path']
        ipv6addr = cli_dict['ipv6']
        is_all = cli_dict['is_all']
        post_request = {"cmd": "run license-install path " + license_path}
        license_install_cmd = ['/usr/cumulus/bin/cl-license', '-i', license_path]

        self.exec_cmd(ipv6addr, is_all, post_request, license_install_cmd)


class EndpointReboot(CliBase):
    def __call__(self, cli_dict):
        ipv6addr = cli_dict['ipv6']
        is_all = cli_dict['is_all']
        reboot_cmd = ['/sbin/reboot']
        post_request = {"cmd": "run reboot"}

        self.exec_cmd(ipv6addr, is_all, post_request, reboot_cmd)


class NosInstall(CliBase):
    def __call__(self, cli_dict):
        image_path = cli_dict['path']
        ipv6addr = cli_dict['ipv6']
        is_all = cli_dict['is_all']
        post_request = {"cmd": "run nos-install path " + image_path}
        onie_install_cmd = ['/usr/cumulus/bin/onie-install', '-a', '-f', '-i', image_path]

        self.exec_cmd(ipv6addr, is_all, post_request, onie_install_cmd)


class DaemonReload(CliBase):
    def __call__(self, cli_dict):
        daemon = cli_dict['daemon-name']
        ipv6addr = cli_dict['ipv6']
        is_all = cli_dict['is_all']
        post_request = {"cmd": "run daemon restart " + daemon}
        daemon_restart_cmd = ['/bin/systemctl', 'restart', daemon + '.service']

        self.exec_cmd(ipv6addr, is_all, post_request, daemon_restart_cmd)


class CLSupport(CliBase):
    def __call__(self, cli_dict):
        ipv6addr = cli_dict['ipv6']
        is_all = cli_dict['is_all']
        post_request = {"cmd": "run cl-support"}
        cl_support_cmd = ['/usr/cumulus/bin/cl-support']

        self.exec_cmd(ipv6addr, is_all, post_request, cl_support_cmd)


class AptMgmt(CliBase):
    def __call__(self, cli_dict):
        ipv6addr = cli_dict['ipv6']
        is_all = cli_dict['is_all']
        is_update = cli_dict['is_apt_update']
        is_upgrade = cli_dict['is_apt_upgrade']
        pkg_name = cli_dict['pkg_name']
        post_request = None
        apt_mgmt_cmd = None

        if is_update:
            post_request = {"cmd": "run apt-get update"}
            apt_mgmt_cmd = ['/usr/bin/apt-get', '-y', 'update']
        if is_upgrade:
            post_request = {"cmd": "run apt-get upgrade"}
            apt_mgmt_cmd = ['/usr/bin/apt-get', '-y', 'upgrade']
        if pkg_name:
            post_request = {"cmd": "run apt-get install " + pkg_name}
            apt_mgmt_cmd = ['/usr/bin/apt-get', '-y', 'install', pkg_name]

        self.exec_cmd(ipv6addr, is_all, post_request, apt_mgmt_cmd)


class CliFactory:
    config_array = {'path': {'arg': '<text>',
                             'cli_name': 'path',
                             'default': None},
                    'run': {'arg': '<ipv6>',
                            'cli_name': 'run',
                            'default': None},
                    'restart': {'arg': '<text>',
                                'cli_name': 'restart',
                                'default': None},
                    'install': {'arg': '<text>',
                                'cli_name': 'install',
                                'default': None},
                    }

    def __init__(self, cli):
        args = cli.args
        self.cli = cli
        self.cli_kwargs = {'path': '',
                           'ipv6': None,
                           'is_all': False,
                           'daemon-name': '',
                           'is_apt_update': False,
                           'is_apt_upgrade': False,
                           'pkg_name': '',
                           }

        for k in self.config_array.keys():
            argval = args.get(self.config_array[k]['arg'])
            cliname = args.get(self.config_array[k]['cli_name'])

            if cliname == 'path':
                self.cli_kwargs['path'] = argval
            if cliname == 'run':
                if argval is not None:
                    self.cli_kwargs['ipv6'] = argval
            if cliname == 'restart':
                self.cli_kwargs['daemon-name'] = argval
            if cliname == 'install':
                self.cli_kwargs['pkg_name'] = argval

        if args.get('all'):
            self.cli_kwargs['is_all'] = True
        if args.get('update'):
            self.cli_kwargs['is_apt_update'] = True
        if args.get('upgrade'):
            self.cli_kwargs['is_apt_upgrade'] = True

    def exec_cli(self):
        obj = None
        args = self.cli.args

        if args.get('license-install'):
            obj = LicenseInstall()
        if args.get('reboot'):
            obj = EndpointReboot()
        if args.get('nos-install'):
            obj = NosInstall()
        if args.get('daemon'):
            obj = DaemonReload()
        if args.get('cl-support'):
            obj = CLSupport()
        if args.get('apt-get'):
            obj = AptMgmt()

        # exec __call__ of the correct CLI
        if obj is not None:
            obj(self.cli_kwargs)


class ConfigWrapper(ConfigWrapperBase):
    def __init__(self, show_linux_command, color_diffs):
        super(ConfigWrapper, self).__init__('run', show_linux_command, color_diffs)

    def eval_command_line_args(self, cli):
        CliFactory(cli).exec_cli()

    def show_config_commands(self, user_may_edit):
        return ([], [])

    def show_config_summary(self, user_may_edit, summary):
        return ([], [])

    def del_all(self, cli):
        return ([], [])
