# Copyright (C) 2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    # snmp-server commands
    snmp-server (add|del) snmp-server listening-address localhost
    snmp-server (add|del) snmp-server listening-address all
    snmp-server (add|del) snmp-server listening-address <ip>
    snmp-server (add|del) snmp-server listening-address <ip> <ip>
    snmp-server (add|del) snmp-server listening-address <ip> <ip> <ip>
    snmp-server (add|del) snmp-server listening-address <ip> <ip> <ip> <ip>
    snmp-server (add|del) snmp-server listening-address <ip> <ip> <ip> <ip> <ip>
    snmp-server (add|del) snmp-server listening-address <ip> <ip> <ip> <ip> <ip> <ip>
    snmp-server del       snmp-server viewname <text>
    snmp-server (add|del) snmp-server viewname <text> (included|excluded) <text>
    snmp-server del       snmp-server readonly-community <text>
    snmp-server add       snmp-server readonly-community <text> access (any|localhost|<ipv4>|<ipv4/prefixlen>)
    snmp-server add       snmp-server readonly-community <text> access (any|localhost|<ipv4>|<ipv4/prefixlen>) oid <text>
    snmp-server add       snmp-server readonly-community <text> access (any|localhost|<ipv4>|<ipv4/prefixlen>) view <text>
    snmp-server del       snmp-server readonly-community-v6 <text>
    snmp-server add       snmp-server readonly-community-v6 <text> access (any|localhost|<ipv6>|<ipv6/prefixlen>)
    snmp-server add       snmp-server readonly-community-v6 <text> access (any|localhost|<ipv6>|<ipv6/prefixlen>) oid <text>
    snmp-server add       snmp-server readonly-community-v6 <text> access (any|localhost|<ipv6>|<ipv6/prefixlen>) view <text>
    snmp-server del       snmp-server username <text>
    snmp-server add       snmp-server username <text>     auth-none
    snmp-server add       snmp-server username <text>     auth-none oid <text>
    snmp-server add       snmp-server username <text>     auth-none view <text>
    snmp-server add       snmp-server username <text>     (auth-md5|auth-sha) <text>
    snmp-server add       snmp-server username <text>     (auth-md5|auth-sha) <text> oid <text>
    snmp-server add       snmp-server username <text>     (auth-md5|auth-sha) <text> view <text>
    snmp-server add       snmp-server username <text>     (auth-md5|auth-sha) <text> (encrypt-des|encrypt-aes) <text>
    snmp-server add       snmp-server username <text>     (auth-md5|auth-sha) <text> (encrypt-des|encrypt-aes) <text> oid <text>
    snmp-server add       snmp-server username <text>     (auth-md5|auth-sha) <text> (encrypt-des|encrypt-aes) <text> view <text>
    snmp-server add snmp-server system-location <wildcard>
    snmp-server del snmp-server system-location [<wildcard>]
    snmp-server add snmp-server system-contact  <wildcard>
    snmp-server del snmp-server system-contact  [<wildcard>]
    snmp-server add snmp-server system-name     <wildcard>
    snmp-server del snmp-server system-name     [<wildcard>]
    snmp-server add snmp-server trap-destination (localhost|<ip>) community-password <text>
    snmp-server add snmp-server trap-destination (localhost|<ip>) community-password <text> version (1|2c)
    snmp-server del snmp-server trap-destination (localhost|<ip>)
    snmp-server add snmp-server trap-link-up [check-frequency <5-300>]
    snmp-server del snmp-server trap-link-up
    snmp-server add snmp-server trap-link-down [check-frequency <5-300>]
    snmp-server del snmp-server trap-link-down
    snmp-server (add|del) snmp-server trap-snmp-auth-failures
    snmp-server add snmp-server trap-cpu-load-average one-minute <text>
    snmp-server add snmp-server trap-cpu-load-average one-minute <text> five-minute <text>
    snmp-server add snmp-server trap-cpu-load-average one-minute <text> five-minute <text> fifteen-minute <text>
    snmp-server del snmp-server trap-cpu-load-average
    snmp-server del snmp-server all
    snmp-server show snmp-server status
    snmp-server example snmp-server
    net show configuration snmp-server

Options:
    snmp-server : Configure the SNMP server
    listening-address : Configure the one or more (comma separated) listening IP address(s).
    trap-destination : Configure the destination IP address where traps will be sent.
    trap-snmp-auth-failures : Enable or disable Traps for SNMP Authentication failures.
    trap-cpu-load-average : Enable or disable Traps for CPU Load Average thresholds.
    one-minute : One Minute load average threshold to send SNMP Trap (e.g.  2.76)
    five-minute : Five Minute load average threshold to send SNMP Trap (e.g.  2.76)
    fifteen-minute : Fifteen Minute load average threshold to send SNMP Trap (e.g.  2.76)
    trap-link-up : Enable or disable SNMP Traps for interface link-up status.
    trap-link-down : Enable or disable SNMP Traps for interface link-down status.
    check-frequency : Sets the link up or link down checking frequency in seconds (default is 60s)
    localhost : Defines the loopback for listening-address or localhost sourced requests for access restriction.
    any : Allows requests from any source.
    2c : SNMP Version 2c
    included : Includes a OID MIB Tree or name in a viewname
    excluded : Excludes a OID MIB Tree or name in a viewname
    viewname : Define a view name to restrict MIB tree exposure.  This can be linked to a community string or SNMPv3 username
    view : Refers to a view name for restricting MIB tree exposure.
    oid : SNMP Object Identifer as either a dotted decimal string or object name
    readonly-community : Defines a read-only community string password for version 1 or 2c access for IPv4.
    readonly-community-v6 : Defines a read-only community string password for version 1 or 2c access for IPv6.
    community-password : A text string that will be used as a password for SNMP version 1 or 2c.
    access : Defines which host IP address or network subnet is allowed access to a community password.
    username : Defines an SNMPv3 username.
    system-location : Sets the SNMPv2-MIB system location field
    system-contact : Sets the SNMPv2-MIB system contact field
    system-name : Sets the SNMPv2-MIB system name field
    auth-none : No SNMPv3 authentication or encryption
    auth-sha : Sets SNMPv3 Secure Hash Algorithm for authentication
    auth-md5 : Sets SNMPv3 Message Digest 5 Algorithm for authentication
    encrypt-des : Sets SNMPv3 Data Encryption Algorithm for encryption
    encrypt-aes : Sets SNMPv3 Advanced Encryption Standard for encryption

"""

from nclu import (
    ConfigWrapperBase,
    ConfigFile,
    PENDING_COMMIT_NOCHANGE,
    PENDING_COMMIT_CHANGE,
    PENDING_COMMIT_ERROR,
    ConfigFileSingleLine,
    print_diff,
    tabulate_remove_unicode as tru
)

from collections import OrderedDict
from subprocess import check_output, CalledProcessError
from tabulate import tabulate
from ipaddr import IPNetwork
import json
import logging
import logging.handlers
import os
import re
import shutil
import glob
import pprint
import copy

log = logging.getLogger('netd')


class ConfigWrapper(ConfigWrapperBase):

    def __init__(self, show_linux_command, color_diffs):
        self.color_diffs = color_diffs
        self.show_linux_command = show_linux_command
        self.username_cache    = '/var/lib/snmp/snmpd.conf'
        self.snmp_scratch      = '/var/run/nclu/snmp/'
        self.scratch_config_file  = '/var/run/nclu/snmp/snmpd.conf'
        self.temp_config_file  = '/var/run/nclu/snmp/temp_snmpd.conf'
        self.config_file          = '/etc/snmp/snmpd.conf'

        '''
        config_array is indexed by the nclu configuration keywords.  It
        contains the following:
            config-name: a list of snmpd.conf keywords used for this keyword
            handler    : function used to convert from nclu to snmpd.conf settings.
            field-name : used when a single field can be selected from nclu
            regex_nclu_to_conf: array indexed by snmpd conf keyword that contains
                 list of tuples for (pattern, replacement) for writing
                 snmpd.conf.
            regex_conf_to_nclu: array indexed by snmpd conf keyword that
                 contains list of tuples for (pattern, replacement) for
                 converting snmpd.conf to nclu config
            hidden-entries: list of keywords used internally that will not be shown
        '''
        self.config_array = {'listening-address' : {'config-name'  : ['agentAddress'],
                                                    'handler'      : 'listening_address_handler',
                                                    'hidden-entries' : [],
                                                    'field-name'   : '<ip>',
                                                    'regex_conf_to_nclu' : {
                                                        'agentAddress' : [
                                                            # just show all the IP addresses given.  These have already been converted
                                                            (r'''(.*)''', '{}'),
                                                        ],
                                                    },
                                                    # these substitutions are for going from a config file to nclu keywords
                                                    'sub_conf_to_nclu' : {
                                                        'agentAddress' : [(r',', ' '), (r'udp:161', 'all'), (r'127.0.0.1', 'localhost')]},
                                                    'sub_nclu_to_conf' : {
                                                        'agentAddress' : [(r'all', 'udp:161'), (r'localhost', '127.0.0.1')]},
                                                   },
                             'viewname': {'config-name'  : ['view'],
                                          'handler'      : 'viewname_handler',
                                          'hidden-entries' : [],
                                          'field-name'   : '<text>',
                                          'regex_conf_to_nclu' : {
                                              'view' : [
                                                  (r'''(?P<viewname>\S+)\s+included\s+(?P<oidname>\S+)''', '{} included {}'),
                                                  (r'''(?P<viewname>\S+)\s+excluded\s+(?P<oidname>\S+)''', '{} excluded {}'),
                                              ],
                                          },
                                         },
                             'trap-snmp-auth-failures': {'config-name' : ['authtrapenable'],
                                                         'handler'      : 'trap_auth_handler',
                                                         'hidden-entries' : ['2'],
                                                         'regex_nclu_to_conf' : [
                                                             (r'''trap-snmp-auth-failures''', '1'),
                                                         ],
                                                         'regex_conf_to_nclu' : {
                                                             'authtrapenable' : [
                                                                 # only handle a 1 since other values will not show up
                                                                 # when showing this config, we show a space
                                                                 (r'''(1)''', ' '),
                                                             ],
                                                         },
                                                        },
                             'trap-destination': {'config-name'  : ['trapsink', 'trap2sink'],
                                                  'handler'      : 'trap_destination_handler',
                                                  'hidden-entries' : [],
                                                  'field-name'   : '<text>',
                                                  'regex_nclu_to_conf' : [
                                                          (r'''(?P<ipaddr>\S+)\s+community-password\s+(?P<password>\S+)\s+version\s+1''', 'trapsink {} {}'),
                                                          (r'''(?P<ipaddr>\S+)\s+community-password\s+(?P<password>\S+)\s+version\s+2c''', 'trap2sink {} {}'),
                                                          (r'''(?P<ipaddr>\S+)\s+community-password\s+(?P<password>\S+)''', 'trap2sink {} {}'),
                                                          (r'''(?P<ipaddr>\S+)''', 'trap2sink {}'),
                                                  ],
                                                  'regex_conf_to_nclu' : {
                                                      'trapsink' : [
                                                          (r'''(?P<ipaddr>\S+)\s+(?P<password>\S+)''', '{} community-password {} version 1'),
                                                      ],
                                                      'trap2sink' : [
                                                          (r'''(?P<ipaddr>\S+)\s+(?P<password>\S+)''', '{} community-password {} version 2c'),
                                                      ]
                                                  },
                                                  'sub_conf_to_nclu' : {
                                                      'trapsink' : [(r'127.0.0.1', 'localhost')],
                                                      'trap2sink' : [(r'127.0.0.1', 'localhost')],
                                                  },
                                                  'sub_nclu_to_conf' : [(r'localhost', '127.0.0.1')],
                                                 },
                             'trap-link-up': {'config-name'  : ['monitor'],
                                              'handler'      : 'trap_link_handler',
                                              'hidden-entries' : [],
                                              'field-name'   : '<5-300>',
                                              'regex_nclu_to_conf' : [
                                                  (r'''trap-link-up\s+check-frequency\s+(?P<frequency>\S+)''',
                                                   'CumulusLinkUP -S -r {} -o ifName -o ifIndex -o ifAdminStatus -o ifOperStatus ifOperStatus != 2'),
                                                  (r'''trap-link-up''',
                                                   'CumulusLinkUP -S -r 60 -o ifName -o ifIndex -o ifAdminStatus -o ifOperStatus ifOperStatus != 2'),
                                              ],
                                              'regex_conf_to_nclu' : {
                                                  'monitor' : [
                                                      (r'''CumulusLinkUP -S -r\s+(?P<frequency>\S+)\s+-o ifName -o ifIndex -o ifAdminStatus -o ifOperStatus ifOperStatus != 2''',
                                                       'check-frequency {}'),
                                                  ],
                                              },
                                             },
                             'trap-link-down': {'config-name'  : ['monitor'],
                                              'handler'      : 'trap_link_handler',
                                              'hidden-entries' : [],
                                              'field-name'   : '<5-300>',
                                              'regex_nclu_to_conf' : [
                                                  (r'''trap-link-down\s+check-frequency\s+(?P<frequency>\S+)''',
                                                   'CumulusLinkDOWN -S -r {} -o ifName -o ifIndex -o ifAdminStatus -o ifOperStatus ifOperStatus == 2'),
                                                  (r'''trap-link-down''',
                                                   'CumulusLinkDOWN -S -r 60 -o ifName -o ifIndex -o ifAdminStatus -o ifOperStatus ifOperStatus == 2'),
                                              ],
                                              'regex_conf_to_nclu' : {
                                                  'monitor' : [
                                                      (r'''CumulusLinkDOWN -S -r\s+(?P<frequency>\S+)\s+-o ifName -o ifIndex -o ifAdminStatus -o ifOperStatus ifOperStatus == 2''',
                                                       'check-frequency {}'),
                                                  ],
                                              },
                                             },
                             'trap-cpu-load-average': {'config-name'  : ['load'],
                                              'handler'      : 'trap_load_handler',
                                              'hidden-entries' : [],
                                              'field-name'   : '<text>',
                                              'regex_nclu_to_conf' : [
                                                  (r'''one-minute\s+(?P<onemin>\S+)\s+five-minute\s+(?P<fivemin>\S+)\s+fifteen-minute\s+(?P<fifteenmin>\S+)''',
                                                   '{} {} {}'),
                                                  (r'''one-minute\s+(?P<onemin>\S+)\s+five-minute\s+(?P<fivemin>\S+)''',
                                                   '{} {} 0'),
                                                  (r'''one-minute\s+(?P<onemin>\S+)''',
                                                   '{} 0 0')
                                              ],
                                              'regex_conf_to_nclu' : {
                                                  'load' : [
                                                      (r'''(?P<onemin>\S+)\s+(?P<fivemin>\S+)\s+(?P<fifteenmin>\S+)''',
                                                       'one-minute {} five-minute {} fifteen-minute {}'),
                                                  ],
                                              },
                                             },
                             'readonly-community': {'config-name'  : ['rocommunity'],
                                                    'handler'      : 'readonly_community_handler',
                                                    'hidden-entries' : [],
                                                    'field-name'   : '<text>',
                                                    'regex_nclu_to_conf' : {
                                                        'rocommunity' : [
                                                            (r'''(?P<password>\S+)\s+access\s+(?P<accessname>\S+)\s+view\s+(?P<viewname>\S+)''', '{} {} -V {}'),
                                                            (r'''(?P<password>\S+)\s+access\s+(?P<accessname>\S+)\s+oid\s+(?P<oidname>\S+)''', '{} {} {}'),
                                                            (r'''(?P<password>\S+)\s+access\s+(?P<accessname>\S+)''', '{} {}'),
                                                            (r'''(?P<password>\S+)''', '{}')
                                                        ],
                                                    },
                                                    'regex_conf_to_nclu' : {
                                                        'rocommunity' : [
                                                            (r'''(?P<roname>\S+)\s+(?P<accessname>\S+)\s+-V\s+(?P<viewname>\S+)''', '{} access {} view {}'),
                                                            (r'''(?P<roname>\S+)\s+(?P<accessname>\S+)\s+(?P<oidname>\S+)''', '{} access {} oid {}'),
                                                            (r'''(?P<roname>\S+)\s+(?P<accessname>\S+)''', '{} access {}'),
                                                        ],
                                                    },
                                                    'sub_conf_to_nclu' : {
                                                        'rocommunity' : [(r'default', 'any'), ('127.0.0.1', 'localhost')],
                                                    },
                                                    'sub_nclu_to_conf' : {
                                                        'rocommunity' : [(r'any', 'default'), (r'localhost', '127.0.0.1')],
                                                    },
                                                   },
                             'readonly-community-v6': {'config-name'  : ['rocommunity6'],
                                                       'handler'      : 'readonly_community_handler',
                                                       'hidden-entries' : [],
                                                       'field-name'   : '<text>',
                                                       'regex_nclu_to_conf' : {
                                                           'rocommunity6' : [
                                                               (r'''(?P<password>\S+)\s+access\s+(?P<accessname>\S+)\s+view\s+(?P<viewname>\S+)''', '{} {} -V {}'),
                                                               (r'''(?P<password>\S+)\s+access\s+(?P<accessname>\S+)\s+oid\s+(?P<oidname>\S+)''', '{} {} {}'),
                                                               (r'''(?P<password>\S+)\s+access\s+(?P<accessname>\S+)''', '{} {}'),
                                                               (r'''(?P<password>\S+)''', '{}')
                                                           ],
                                                       },
                                                       'regex_conf_to_nclu' : {
                                                           'rocommunity6' : [
                                                               (r'''(?P<roname>\S+)\s+(?P<accessname>\S+)\s+-V\s+(?P<viewname>\S+)''', '{} access {} view {}'),
                                                               (r'''(?P<roname>\S+)\s+(?P<accessname>\S+)\s+(?P<oidname>\S+)''', '{} access {} oid {}'),
                                                               (r'''(?P<roname>\S+)\s+(?P<accessname>\S+)''', '{} access {}'),
                                                           ],
                                                       },
                                                       'sub_conf_to_nclu' : {
                                                           'rocommunity6' : [(r'default', 'any'), (' +::1', ' localhost')]
                                                       },
                                                       'sub_nclu_to_conf' : { 'rocommunity6' : [(r'any', 'default'), (r' +localhost', ' ::1')]
                                                       },
                                                      },
                             'system-location': {'config-name' : ['sysLocation'],
                                                 'handler'      : 'sys_handler',
                                                 'hidden-entries' : [],
                                                 'field-name'   : '<wildcard>',
                                                 'regex_conf_to_nclu' : {
                                                     'sysLocation' : [
                                                         (r'''(.*)''', '{}'),
                                                     ],
                                                 },
                                                },
                             'system-contact': {'config-name' : ['sysContact'],
                                                'handler'      : 'sys_handler',
                                                'hidden-entries' : [],
                                                'field-name'   : '<wildcard>',
                                                 'regex_conf_to_nclu' : {
                                                     'sysContact' : [
                                                         (r'''(.*)''', '{}'),
                                                     ],
                                                 },
                                               },
                             'system-name': {'config-name' : ['sysName'],
                                             'handler'      : 'sys_handler',
                                             'hidden-entries' : [],
                                             'field-name'   : '<wildcard>',
                                             'regex_conf_to_nclu' : {
                                                 'sysName' : [
                                                     (r'''(.*)''', '{}'),
                                                 ],
                                             },
                                            },
                             'username': {'config-name'  : ['createUser', 'rouser'],
                                          'hidden-entries' : ['_snmptrapusernameX'],
                                          'handler'      : 'username_handler',
                                          'field-name'   : '<text>',
                                          'regex_nclu_to_conf' : {
                                              'createUser' : [
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+encrypt-aes\s+(?P<aes>\S+)\s+view\s+\S+''', '{} MD5 {} AES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+encrypt-aes\s+(?P<aes>\S+)\s+oid\s+\S+''', '{} MD5 {} AES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+encrypt-des\s+(?P<des>\S+)\s+view\s+\S+''', '{} MD5 {} DES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+encrypt-des\s+(?P<des>\S+)\s+oid\s+\S+''', '{} MD5 {} DES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+encrypt-aes\s+(?P<aes>\S+)\s+view\s+\S+''', '{} SHA {} AES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+encrypt-aes\s+(?P<aes>\S+)\s+oid\s+\S+''', '{} SHA {} AES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+encrypt-des\s+(?P<des>\S+)\s+view\s+\S+''', '{} SHA {} DES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+encrypt-des\s+(?P<des>\S+)\s+oid\s+\S+''', '{} SHA {} DES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+encrypt-aes\s+(?P<aes>\S+)''', '{} MD5 {} AES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+encrypt-des\s+(?P<des>\S+)''', '{} MD5 {} DES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+encrypt-aes\s+(?P<aes>\S+)''', '{} SHA {} AES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+encrypt-des\s+(?P<des>\S+)''', '{} SHA {} DES {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+view\s+\S+''', '{} MD5 {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)\s+oid\s+\S+''', '{} MD5 {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+view\s+\S+''', '{} SHA {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)\s+oid\s+\S+''', '{} SHA {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+(?P<md5>\S+)''', '{} MD5 {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+(?P<sha>\S+)''', '{} SHA {}'),
                                                  (r'''(?P<name>\S+)\s+auth-none\s+view\s+\S+''', '{}'),
                                                  (r'''(?P<name>\S+)\s+auth-none\s+oid\s+\S+''', '{}'),
                                                  (r'''(?P<name>\S+)\s+auth-none''', '{}'),
                                                  (r'''(?P<name>\S+)''', '{}'),
                                              ],
                                              'rouser' : [
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+encrypt-aes\s+\S+\s+view\s+(?P<view>\S+)''', '{} priv -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+encrypt-aes\s+\S+\s+oid\s+(?P<oid>\S+)''', '{} priv {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+encrypt-des\s+\S+\s+view\s+(?P<view>\S+)''', '{} priv -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+encrypt-des\s+\S+\s+oid\s+(?P<oid>\S+)''', '{} priv {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+encrypt-aes\s+\S+\s+view\s+(?P<view>\S+)''', '{} priv -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+encrypt-aes\s+\S+\s+oid\s+(?P<oid>\S+)''', '{} priv {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+encrypt-des\s+\S+\s+view\s+(?P<view>\S+)''', '{} priv -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+encrypt-des\s+\S+\s+oid\s+(?P<oid>\S+)''', '{} priv {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+encrypt-aes\s+\S+''', '{} priv'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+encrypt-des\s+\S+''', '{} priv'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+encrypt-aes\s+\S+''', '{} priv'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+encrypt-des\s+\S+''', '{} priv'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+view\s+(?P<view>\S+)''', '{} auth -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+\s+oid\s+(?P<oid>\S+)''', '{} auth {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+view\s+(?P<view>\S+)''', '{} auth -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+\s+oid\s+(?P<oid>\S+)''', '{} auth {}'),
                                                  (r'''(?P<name>\S+)\s+auth-md5\s+\S+''', '{} auth'),
                                                  (r'''(?P<name>\S+)\s+auth-sha\s+\S+''', '{} auth'),
                                                  (r'''(?P<name>\S+)\s+auth-none\s+view\s+(?P<view>\S+)''', '{} noauth -V {}'),
                                                  (r'''(?P<name>\S+)\s+auth-none\s+oid\s+(?P<oid>\S+)''', '{} noauth {}'),
                                                  (r'''(?P<name>\S+)\s+auth-none''', '{} noauth'),
                                                  (r'''(?P<name>\S+)''', '{}'),
                                              ]},
                                          'regex_conf_to_nclu' : {
                                              'createUser' : [
                                                  (r'''(?P<name>\S+)\s+MD5\s+(?P<md5>\S+)\s+AES\s+(?P<aes>\S+)''', '{} auth-md5 {} encrypt-aes {}'),
                                                  (r'''(?P<name>\S+)\s+MD5\s+(?P<md5>\S+)\s+DES\s+(?P<des>\S+)''', '{} auth-md5 {} encrypt-des {}'),
                                                  (r'''(?P<name>\S+)\s+SHA\s+(?P<sha>\S+)\s+AES\s+(?P<aes>\S+)''', '{} auth-sha {} encrypt-aes {}'),
                                                  (r'''(?P<name>\S+)\s+SHA\s+(?P<sha>\S+)\s+DES\s+(?P<des>\S+)''', '{} auth-sha {} encrypt-des {}'),
                                                  (r'''(?P<name>\S+)\s+MD5\s+(?P<md5>\S+)''', '{} auth-md5 {}'),
                                                  (r'''(?P<name>\S+)\s+SHA\s+(?P<sha>\S+)''', '{} auth-sha {}'),
                                                  (r'''(?P<name>\S+)''', '{} auth-none'),
                                              ],
                                              'rouser' : [
                                                  (r'''(?P<name>\S+)\s+noauth\s+-V\s+(?P<view>\S+)''', '{} view {}'),
                                                  (r'''(?P<name>\S+)\s+priv\s+-V\s+(?P<view>\S+)''', '{} view {}'),
                                                  (r'''(?P<name>\S+)\s+auth\s+-V\s+(?P<view>\S+)''', '{} view {}'),
                                                  (r'''(?P<name>\S+)\s+noauth\s+(?P<oid>\S+)''', '{} oid {}'),
                                                  (r'''(?P<name>\S+)\s+priv\s+(?P<oid>\S+)''', '{} oid {}'),
                                                  (r'''(?P<name>\S+)\s+auth\s+(?P<oid>\S+)''', '{} oid {}'),
                                              ]},
                                         },
                            }

        self.default_config_array = {'agentAddress': ['127.0.0.1'],
                                     'agentxperms': ['777 777 snmp snmp'],
                                     'agentxsocket': ['/var/agentx/master'],
                                     'createUser': ['_snmptrapusernameX'],
                                     'iquerySecName': ['_snmptrapusernameX'],
                                     'master': ['agentx'],
                                     'monitor' : ['-r 60 -o laNames -o laErrMessage "laTable" laErrorFlag != 0'],
                                     'pass': ['-p 10 1.3.6.1.2.1.1.1 /usr/share/snmp/sysDescr_pass.py'],
                                     'pass_persist': [
                                         '1.3.6.1.2.1.31.1.1.1.18 /usr/share/snmp/snmpifAlias_pp.py',
                                         '1.3.6.1.4.1.40310.1 /usr/share/snmp/resq_pp.py',
                                         '1.3.6.1.4.1.40310.2 /usr/share/snmp/cl_drop_cntrs_pp.py',
                                         '1.3.6.1.2.1.17 /usr/share/snmp/bridge_pp.py',
                                         '1.3.6.1.2.1.47 /usr/share/snmp/entity_pp.py',
                                         '1.3.6.1.2.1.99 /usr/share/snmp/entity_sensor_pp.py',
                                         '1.2.840.10006.300.43 /usr/share/snmp/ieee8023_lag_pp.py'
                                     ],
                                     'rouser': ['_snmptrapusernameX'],
                                     'sysObjectID': ['1.3.6.1.4.1.40310'],
                                     'sysServices': ['72']}

    def sys_handler(self, running_array, conf_word):
        '''
        sys_handler() is a simple method that converts nclu
        config for system-location, contact, and name.
        '''
        cli_dict = self.config_array.get(conf_word, {})
        config_names = cli_dict.get('config-name', [])
        for config_name in config_names:
            current_list = running_array.get(config_name, [])
            add_del = self.args.get('add') or self.args.get('del')
            field_name = cli_dict.get('field-name', None)
            entered_value = self.args.get(field_name, [])
            if add_del == 'add' and field_name:
                # and add will just overwrite the value
                running_array[config_name] = [entered_value]
            elif add_del == 'del' and running_array.has_key(config_name):
                del running_array[config_name]

        return(running_array)

    def trap_destination_handler(self, running_array, conf_word):
        '''
        The trap destination handler creates the trapsink or trap2sink in
        snmpd.conf. If it contains a version parameter, it will get get
        set accordingly: version 1 creates trapsink, version 2c creates
        trap2sink config.  If no version is specified, version 2c will be
        used.
        '''
        # this should be a list of 1 item
        cli_dict = self.config_array.get(conf_word, {})
        add_del = self.args.get('add') or self.args.get('del')

        if self.argv_expanded and len(self.argv_expanded) > 4 and \
           self.argv_expanded[3] == 'trap-destination':
            # we can only deal with strings for joining
            entered_line = [str(i) for i in self.argv_expanded[4:]]
            entered_line = ' '.join(entered_line)
            entered_distination = self.argv_expanded[4]
        else:
            print('Error: Invalid trap-destination command entered')
            return(running_array)

        sublist = cli_dict.get('sub_nclu_to_conf', [])
        regexlist = cli_dict.get('regex_nclu_to_conf', [])
        # entered line will include either the trapsink or trap2sink
        # keyword.  This is different then other conversions
        entered_line = self.converter(entered_line, sublist, regexlist)
        trap_word = entered_line.split()[0]
        # now we can remove the leading trapsink or trap2sink for a match
        entered_line = ' '.join(entered_line.split()[1:])
        if add_del == 'add':
            # now fix up syntax when converting from nclu to config
            # grab the current config list to fix up
            current_trap_list = running_array.get(trap_word, [])
            if entered_line not in current_trap_list:
                current_trap_list.append(entered_line)
                running_array[trap_word] = current_trap_list
            else:
                print('Warning: trap-destination already exists: %s' % \
                      (entered_line))
        elif add_del == 'del':
            # for deleting we only key off the destination
            # when deleting, we could have only the ip address
            destination = entered_line.split()[0]
            notfound = True
            for word in ['trapsink', 'trap2sink']:
                current_trap_list = running_array.get(word, [])
                delete_list = []
                for line in current_trap_list:
                    if destination in line:
                        delete_list.append(line)
                        notfound = False

                for line in delete_list:
                    current_trap_list.remove(line)
                running_array[word] = current_trap_list
            if notfound:
                print('Warning: trap destination %s not found.' % \
                      destination)

        return(running_array)

    def trap_auth_handler(self, running_array, conf_word):
        '''
        The trap auth SNMP Failure handler creates the authtrapenable command
        If this is set to 1, it will be enabled.  To disable, we
        remove it from the config.
        '''
        # this should be a list of 1 item
        cli_dict = self.config_array.get(conf_word, {})
        add_del = self.args.get('add') or self.args.get('del')
        config_name = cli_dict.get('config-name', [])[0]

        trap_word = self.argv_expanded[3]

        if add_del == 'add':
            # now fix up syntax when converting from nclu to config
            sublist = cli_dict.get('sub_nclu_to_conf', [])
            regexlist = cli_dict.get('regex_nclu_to_conf', [])
            converted_line = self.converter(trap_word, sublist, regexlist)
            # grab the current config list to fix up
            current_list = running_array.get(config_name, [])
            if not current_list:
                # we can only have one instance of authtrapenable
                # so it is either disabled (2) or enabled (1)
                current_list.append(converted_line)
                running_array[config_name] = current_list
            else:
                print('Warning: a command exists in the config,')
                print('possibly not added by NCLU.')
                print('Please run delete first.')
                return(running_array)

        elif add_del == 'del':
            current_list = running_array.get(config_name, [])
            sublist = cli_dict.get('sub_nclu_to_conf', [])
            regexlist = cli_dict.get('regex_nclu_to_conf', [])
            converted_line = self.converter(trap_word, sublist, regexlist)
            current_list = running_array.get(config_name, [])
            # for deletes we remove any lines with authtrapenable in it.
            del_word = converted_line.split()[0]
            del_list = []
            for line in current_list:
                # we will delete even disabled (2) entries
                # there should only be one anyway
                del_list.append(line)
            for line in del_list:
                current_list.remove(line)

            running_array[config_name] = current_list
            if not del_list:
                print('Warning: %s command not found' % trap_word)

        return(running_array)

    def trap_link_handler(self, running_array, conf_word):
        '''
        The trap link handler creates the monitor command for link uo or
        link down in snmpd.conf. If it contains a monitor command with
        all the parameters needed to identifty the interface.
        '''
        # this should be a list of 1 item
        cli_dict = self.config_array.get(conf_word, {})
        add_del = self.args.get('add') or self.args.get('del')
        config_name = cli_dict.get('config-name', [])[0]

        if self.argv_expanded:
            # we can only deal with strings for joining
            # the entered line includes the trap-link-up or down word
            entered_line = [str(i) for i in self.argv_expanded[3:]]
            entered_line = ' '.join(entered_line)
        else:
            print('Error: invalid command')
            return(running_array)

        if add_del == 'add':
            # now fix up syntax when converting from nclu to config
            sublist = cli_dict.get('sub_nclu_to_conf', [])
            regexlist = cli_dict.get('regex_nclu_to_conf', [])
            converted_line = self.converter(entered_line, sublist, regexlist)
            # grab the current monitor config list to fix up
            current_monitor_list = running_array.get(config_name, [])
            # just set it since there can only be one
            del_list = []
            for line in current_monitor_list:
                if converted_line.split()[0] == line.split()[0]:
                    # we have a CumulusLinkDown or CumulusLinkUp match
                    # just replace it
                    del_list.append(line)

            # Fist remove any existing linkup or linkdown lines
            for d in del_list:
                current_monitor_list.remove(d)
            # and add what is new
            running_array[config_name].append(converted_line)

        elif add_del == 'del':
            current_monitor_list = running_array.get(config_name, [])
            sublist = cli_dict.get('sub_nclu_to_conf', [])
            regexlist = cli_dict.get('regex_nclu_to_conf', [])
            converted_line = self.converter(entered_line, sublist, regexlist)
            # For deletes we remove any monitor lines with the CumulusLinkUP
            # or CumulusLinkDown word in them.  We leave other monitor commands
            # alone.
            del_word = converted_line.split()[0]
            del_list = []
            for line in current_monitor_list:
                if del_word in line:
                    del_list.append(line)
            for line in del_list:
                current_monitor_list.remove(line)

            running_array[config_name] = current_monitor_list
            if not del_list:
                print('Warning: %s command not found' % entered_line)

        return(running_array)

    def trap_load_handler(self, running_array, conf_word):
        '''
        The trap load handler creates the load command for link uo or
        link down in snmpd.conf. If it contains a monitor command with
        all the parameters needed to identifty the interface.
        '''
        # this should be a list of 1 item
        cli_dict = self.config_array.get(conf_word, {})
        add_del = self.args.get('add') or self.args.get('del')
        config_name = cli_dict.get('config-name', [])[0]

        if self.argv_expanded:
            # we can only deal with strings for joining
            # the entered line does not include the load word
            entered_line = [str(i) for i in self.argv_expanded[4:]]
            entered_line = ' '.join(entered_line)
        else:
            print('Error: invalid command')
            return(running_array)

        myfloats = [str(i) for i in self.argv_expanded[4:]]
        for i in range(1, len(myfloats), 2):
            # numbers are only in positions 1 3 and 5
            x = myfloats[i]
            y = x.replace('.','')
            if y.isdigit() and int(y) >= 0:
                # int or float is good
                pass
            else:
                print('Error: Only positive integers or floating point numbers allowed for cpu load values')
                return(running_array)
        # we can have only one load line in the config
        # the monitor command for this has to always exists in the conf
        if add_del == 'add':
            # now fix up syntax when converting from nclu to config
            sublist = cli_dict.get('sub_nclu_to_conf', [])
            regexlist = cli_dict.get('regex_nclu_to_conf', [])
            converted_line = self.converter(entered_line, sublist, regexlist)
            # grab the current config list to fix up
            running_array[config_name] = [converted_line]
        elif add_del == 'del':
            current_load_list = running_array.get(config_name, [])
            # For deletes we remove all load lines.
            current_load_list = running_array.get(config_name, [])
            if not current_load_list:
                print('Warning: %s command not found' % config_name)
            else:
                running_array[config_name] = []

        return(running_array)

    def viewname_handler(self, running_array, conf_word):
        '''
        The view handler grabs the <text> list and takes the view name
        as the fist entry and the OID as the second entry.
        We can have multiple lines with the same view name, e.g.
        view foo1  included  .1
        view foo1  exclused  .1.3.6
        view foo2  included  .1.3.6.1.2
        view foo3  included  .1

        The entered value is a list of the view name and the OID
        '''
        # grab views in currently running config
        cli_dict = self.config_array.get(conf_word, {})
        config_names = cli_dict.get('config-name', [])
        for config_name in config_names:
            current_list = running_array.get(config_name, [])
            add_del = self.args.get('add') or self.args.get('del')
            field_name = cli_dict.get('field-name', None)
            entered_value = self.args.get(field_name, [])
            # make sure we have a list to work with
            if not isinstance(entered_value, list):
                entered_value = [entered_value]
                #print('Error: expected list for entered value, got %s' % \
            #          entered_value)
            # Next, add or delete the entered value
            # entered values are viewname and oid
            # and the view entry in the running array is just a list
            viewname = entered_value[0]
            current_view = running_array.get(config_name, [])
            temp_view = copy.deepcopy(current_view)
            if add_del == 'add':
                oid = entered_value[1]
                viewtype = self.args.get('included') or self.args.get('excluded')
                viewline = '%s %s %s' % (viewname, viewtype, oid)
                if viewline not in current_view:
                    current_view.append(viewline)
                else:
                    print('Warning: %s %s already in view %s' % \
                          (oid, viewtype, viewname))
            else:
                # for deleting, we only look at view name
                for line in temp_view:
                    if viewname in line:
                        current_view.remove(line)

            running_array[config_name] = current_view
        return(running_array)

    def listening_address_handler(self, running_array, conf_word):
        '''
        The listening address handler may need to split the existing
        value and add or remove one part. In the snmpd.conf file,
        agentAddress has only one line but can contain comma
        separated values.  The entered value could
        be a single IP address or a list of IP addresses.
        (udp:1.1.1.1:161,2.2.2.2,udp6:5.5.5.5:161)
        '''
        # this should be a list of 1 item
        cli_dict = self.config_array.get(conf_word, {})
        config_names = cli_dict.get('config-name', [])
        for config_name in config_names:
            current_val = running_array.get(config_name, [])
            listeners = current_val
            if len(current_val) > 0 and ',' in current_val[0]:
                listeners = current_val[0].split(',')
                # make sure we have a list to work with
            add_del = self.args.get('add') or self.args.get('del')
            field_name = cli_dict.get('field-name', None)
            # translate words like all to snmpd.conf words like udp:161
            keyword = self.get_keyword(conf_word)
            # grab the keywords or the entered value
            entered_values = self.args.get(field_name, None) or keyword
            if not isinstance(entered_values, list):
                entered_values = [entered_values]
            # convert keywords before adding or deleting
            for ip in entered_values:
                if (ip == 'udp:161' or ip == '::') and add_del == 'add':
                    listeners = [ip]
                    print('Warning: configuring all removes other listening addresses')
                    break
                else:
                    # listeners should be list of one or more ip addresses
                    if add_del == 'add' and ('udp:161' in listeners or \
                                             '::' in listeners):
                        print('Warning: Removing all to add address %s.' % ip)
                        listeners.append(ip)
                        if 'udp:161' in listeners:
                            listeners.remove('udp:161')
                        if '::' in listeners:
                            listeners.remove('::')
                    elif add_del == 'add' and ip in listeners:
                        print('Warning: Cannot add %s. It is already a listener-address' %  ip)
                    elif add_del == 'add' and ip not in listeners:
                        listeners.append(ip)
                    elif add_del == 'del' and ip not in listeners:
                        print('Warning: Cannot delete %s. It is not a listener-address' %  ip)
                    elif add_del == 'del' and ip in listeners:
                        listeners.remove(ip)
                        # rest of code expects a list of comma separated IP addresses
            if len(listeners) == 0:
                print('Warning: No listening addresses configured.  Restoring localhost')
                listeners = ['127.0.0.1']

            running_array[config_name] = [str(','.join(listeners))]

        return(running_array)

    def readonly_community_handler(self, running_array, conf_word):
        '''
        The rocommunity handler creates the rocommunity config in snmpd.conf
        and must include an access field.  It can also contain an optional
        oid or view field at the end.
        It should be able to handle readonly-community or
        readonly-community-v6.  Note that only the first community password
        line is active even though snmpd does not complain if there are more
        then one rocommunity lines with the same password.

        '''
        # this should be a list of 1 item
        cli_dict = self.config_array.get(conf_word, {})
        config_names = cli_dict.get('config-name', [])
        if not config_names:
            print('Error: invalid config-name defined for rocommunity')
            return(running_array)

        # make sure we have a list to work with
        add_del = self.args.get('add') or self.args.get('del')
        if self.argv_expanded and len(self.argv_expanded) > 4:
            entered_line = ' '.join(self.argv_expanded[4:])
        else:
            print('Error: Invalid readonly community command entered')
            return(running_array)

        if self.args.get('<ipv4/prefixlen>') or self.args.get('<ipv6/prefixlen>'):
            fullip = IPNetwork(self.args.get('<ipv6/prefixlen>') or self.args.get('<ipv4/prefixlen>'))
            if fullip and (fullip.network != fullip.ip):
                print('Error: network/mask mismatch')
                return(running_array)

        # now fix up syntax when converting from config file to nclu
        for config_name in config_names:
            sublist = cli_dict.get('sub_nclu_to_conf', []).get(config_name, [])
            regexlist = cli_dict.get('regex_nclu_to_conf', []).get(config_name, [])
            entered_line = self.converter(entered_line, sublist, regexlist)
            entered_word = entered_line.split()[0]
            current_ro_list = running_array.get(config_name, [])
            current_ro_words = [i.split()[0] for i in current_ro_list]
            if add_del == 'add':
                if entered_word not in current_ro_words:
                    current_ro_list.append(entered_line)
                else:
                    print('Warning: readonly community exists: %s' % \
                          (entered_word))
            else:
                removed = False
                removed_list = []
                for line in current_ro_list:
                    if entered_word == line.split()[0]:
                        removed_list.append(line)
                        removed = True

                if not removed:
                    print('Warning: readonly community %s does not exist.' % \
                          (entered_line))

                for line in removed_list:
                    current_ro_list.remove(line)

            running_array[config_name] = current_ro_list
        return(running_array)

    def username_handler(self, running_array, conf_word):
        '''
        The username handler creates both createUser and
        rouser commands in snmpd.conf.
        '''
        cli_dict = self.config_array.get(conf_word, {})
        # this should be a list of 1 item
        wordlist = cli_dict.get('config-name', [])
        if not wordlist:
            print('Error: invalid config-name defined for username')
            return(running_array)

        # need to handle both createUser and rouser for the username
        for myword in wordlist:
            current_val = running_array.get(myword, [])

            # make sure we have a list to work with
            add_del = self.args.get('add') or self.args.get('del')
            username = None

            if self.argv_expanded and len(self.argv_expanded) > 4 and \
               self.argv_expanded[3] == 'username':
                entered_line = ' '.join(self.argv_expanded[4:])
                username = self.argv_expanded[4]
            else:
                print('Error: Invalid username command entered')
                return(running_array)

            # min length of passwords must be 8 characters
            if len(self.argv_expanded) > 6 and \
               ('auth-md5' in self.argv_expanded or 'auth-sha' in self.argv_expanded):
                if len(self.argv_expanded[6]) < 8:
                    print('Error: authentication passphrase must be greater then 8 characters')
                    return(running_array)

            # min length of passwords must be 8 characters
            if len(self.argv_expanded) > 8 and \
               ('encrypt-aes' in self.argv_expanded or 'encrypt-des' in self.argv_expanded):
                if len(self.argv_expanded[8]) < 8:
                    print('Error: encryption passphrase must be greater then 8 characters')
                    return(running_array)

            # now fix up syntax when converting from config file to nclu
            sublist = cli_dict.get('sub_nclu_to_conf', {}).get(myword, [])
            regexlist = cli_dict.get('regex_nclu_to_conf', {}).get(myword, [])
            entered_line = self.converter(entered_line, sublist, regexlist)
            if add_del == 'add':
                if [i for i in current_val if username in i] == []:
                    # only adds user if it was not there
                    current_val.append(entered_line)
                else:
                    print('Warning: username %s exists: delete it before re-adding it.' % \
                          (username))
                    return(running_array)
            else:
                # for deleting, we only care about the useranme
                if [i for i in current_val if username in i]:
                    # if the username is in the current commands, collect
                    # all the others.
                    current_val = [i for i in current_val if username not in i]
                else:
                    print('Warning: username %s does not exist: ' % (username))
                    return(running_array)

            running_array[myword] = current_val

        return(running_array)

    def get_keyword(self, conf_word):
        # the actual arg value if they did not use default keyword
        # we convert from the nclu keywords to the snmpd.conf words required
        cli_dict = self.config_array.get(conf_word, {})
        conf_name = cli_dict.get('config-name', [])
        trans_array = cli_dict.get('sub_nclu_to_conf', {})
        if not cli_dict or len(conf_name)==0 or not trans_array.get(conf_name[0], []):
            return None
        # we should only have 1 translation list
        trans_list = trans_array.get(conf_name[0], [])
        # check cli-keywords
        translated_word = None
        for word,newword in trans_list:
            # was the word entered
            if self.args.get(word):
                translated_word = newword
                break
        # return the translation so we can use it
        return(translated_word)

    def get_config(self, config_file=None):
        '''
        get_config() parses the snmpd.conf file and returns an array indexed
        by the first config word.  For multiline values, the entry is added to
        a list.  For single line values, it is a one element list. Note that
        some values in the config are not shown or changeable in NCLU.
        These unshown entries are just carried along and not changed.
        For Example, an snmpd.conf snippet like:

            master agentx
            rocommunity public default
            rocommunity mycom default -V myview

        results in:

            config_array = {'rocommunity' : ['public default',
                                             'mycom default -V myview'],
                            'master' : ['agentx'],
                           }
        '''
        config_array = {}
        if os.path.exists(config_file):
            config_fd = open(config_file, 'r')
            config_list = config_fd.read().strip().split('\n')
            config_list = [l for l in config_list \
                           if not l.startswith('#')]
            for line in config_list:
                l = line.split()
                if len(l) > 1:
                    if config_array.get(l[0], None):
                        config_array[l[0]].append(' '.join(l[1:]))
                    else:
                        config_array[l[0]] = [' '.join(l[1:])]
            config_fd.close()
        return config_array

    def clear_username_cache(self):
        '''
        clear_username_cache() reads the current cache file, usually
        in /var/lib/snmp/snmpd.conf, and then loops over the existing
        SNMPv3 usernames.  Lines with these usernames are then removed
        from the cache file as a new file is written out.  This forces
        snmpd to rehash the usernames contained in /etc/snmp/snmpd.conf
        to make sure they are correct.
        '''
        if not os.path.exists(self.username_cache):
            # we have no cache to clear so bale out.
            return
        fd = open(self.username_cache, 'r')
        oldcache = fd.readlines()
        fd.close()

        oldconfig = self.get_config(self.config_file)
        newconfig = self.get_config(self.scratch_config_file)
        delete_set = set([])
        oldusers = set([u.split()[0] for u in oldconfig.get('createUser', [])])
        newusers = set([u.split()[0] for u in newconfig.get('createUser', [])])
        # we need to delete any users that were removed
        delete_set = delete_set | (oldusers - newusers)
        # we need to add in the users that were added
        delete_set = delete_set | (newusers - oldusers)
        # for the users that did not change, we need to check their passwords
        stableusers = oldusers.intersection(newusers)
        oldconfigs = oldconfig.get('createUser', [])
        newconfigs = newconfig.get('createUser', [])
        for userconfig in oldconfigs:
            if userconfig.split()[0] in stableusers and userconfig in newconfigs:
                # the config did not change
                pass
            else:
                # passwords must have changed, delete it
                delete_set.add(userconfig.split()[0])

        if not delete_set:
            return

        # before writing out the new file, make sure to stop snmpd if it is
        # running
        try:
            enabled = check_output(['/bin/systemctl', 'is-active', 'snmpd'])
            # if we get here, we may be enabled
            if 'active' in enabled:
                # need to stop before modifying file
                # the rest of commit_pending should just restart it
                enabled = check_output(['/bin/systemctl', 'stop', 'snmpd'])
        except:
            # if snmpd is not active, we return subprocess.CalledProcessError
            # since it is not active, just pass
            pass

        newfd = open(self.username_cache, 'w')
        for line in oldcache:
            notfound = True
            for user in delete_set:
                # make sure to add quotes to catch the exact username
                if ('"%s"' % user) in line:
                    notfound = False
                    break

            if notfound:
                newfd.write(line)

        newfd.close()
        return

    def save_scratch_config(self, config_array={}):
        '''
        This method simply looks over the config_array and
        writes out the values in the scratch space.  This will
        eventually get written to the running config with a
        net commit.
        '''
        # just overwrite it
        if not config_array:
            config_array = self.default_config_array.copy()
        config_fd = open(self.scratch_config_file, 'w')
        os.chmod(self.scratch_config_file, 0o600)
        config_fd.write('#### Auto-generated config file: do not edit. ####\n')
        entries = config_array.keys()
        entries.sort()
        for entry in entries:
            values = config_array.get(entry, [])
            values.sort()
            for val in values:
                if val: config_fd.write('%s %s\n' % (entry, val))
        config_fd.close()
        return

    def del_all(self, cli):
        # our job here is to wipe the config clean in the scratch area.
        # We are not stopping the snmpd daemon.  We need a better
        # NCLU mechanism to make the stopping of all daemons clear before
        # we start stopping daemons.  netd calls us directly and expects no return
        self.save_scratch_config({})
        return True

    def abort_pending(self):
        # need to delete the /var/run/nclu/snmp configs
        for filename in glob.glob('%s*' % self.snmp_scratch):
            os.remove(filename)
        return

    def commit_pending(self, verbose):
        '''
        This first checks for various missing or incorrect information:
        '''
        commit_ok = True
        modified = []

        # we are doing our own copy and systemctl calls
        if not os.path.exists(self.scratch_config_file):
            # there is no scratch config, bail
            for filename in glob.glob('%s*' % self.snmp_scratch):
                os.remove(filename)
            return (commit_ok, modified)

        elif os.path.exists(self.scratch_config_file) and \
                os.path.exists(self.config_file) and \
                not print_diff(self.config_file,
                               self.scratch_config_file,
                               self.color_diffs):
            # Nothing has changed, but we could get
            # here if net del all is called and
            # creates an identical config in scratch.
            # Since every wrapper commit will get called, we have
            # to recheck for changed configs before proceeding.
            # Otherwise, we start or restart snmpd for no reason.
            for filename in glob.glob('%s*' % self.snmp_scratch):
                os.remove(filename)
            return (commit_ok, modified)

        elif os.path.exists(self.scratch_config_file) and \
                os.path.exists(self.config_file) and \
                print_diff(self.config_file,
                           self.scratch_config_file,
                           self.color_diffs):
            # we have a config and it is different from the one running
            # so we must keep going
            pass
        else:
            # all other cases should bail out
            for filename in glob.glob('%s*' % self.snmp_scratch):
                os.remove(filename)
            return (commit_ok, modified)

        command = 'restart'
        # at this point, we know the config has changed so we will
        # either restart snmpd or just stop.
        # for SNMPv3, we have to clear any user we control from the
        # cache file since it will be recreated with a restart.
        self.clear_username_cache()
        newconfig = self.get_config(self.scratch_config_file)
        if (len(newconfig.get('rocommunity', [])) == 0 and
            len(newconfig.get('rocommunity6', [])) == 0 and \
            newconfig.get('createUser', []) == ['_snmptrapusernameX']) or \
            len(newconfig.get('agentAddress', [])) == 0:
            # snmpd will not work if either there are no listening-addresses
            # or both rocommunity and createUser are not configured
            command = 'stop'
            print("\nWarning: missing readonly community or username")
            print("         snmpd will not be running.")

        # copy the config to the right place.
        shutil.copyfile(self.config_file, self.temp_config_file)
        shutil.copyfile(self.scratch_config_file, self.config_file)
        os.chmod(self.config_file, 0o600)
        modified.append(self.config_file)
        # if usernames were added or deleted we have to clear the
        # snmpv3 username cache file to force passphrases to be
        # rehashed after the restart.
        try:
            try:
                enabled = check_output(['/bin/systemctl', 'is-enabled', 'snmpd'])
            except:
                enabled = check_output(['/bin/systemctl', 'enable', 'snmpd'])
            try:
                enabled = check_output(['/bin/systemctl', 'is-active', 'snmpd'])
                # if it is running restart or stop
                enabled = check_output(['/bin/systemctl', command, 'snmpd'])
            except:
                # it is not running, better just restart it
                if command != 'stop':
                    enabled = check_output(['/bin/systemctl', 'restart', 'snmpd'])
        except CalledProcessError as e:
            print('Error: attept to %s snmpd failed: %s'%(command, e))
            commit_ok = False

        try:
            enabled = check_output(['/bin/systemctl', 'is-active', 'snmpd'])
        except:
            if command != 'stop':
                print("\nWarning: snmpd is not running")
                print("     Check journalctl -u snmpd for error messages")

        # remove rest of files since we just finished the commit
        if commit_ok:
            for filename in glob.glob('%s*' % self.snmp_scratch):
                os.remove(filename)
        elif os.path.exists(self.temp_config_file):
                shutil.copyfile(self.temp_config_file, self.config_file)
                modified = []

        #restart the snmpd daemon with the new interface configs
        return [commit_ok, modified]

    def show_pending(self, print_errors=True):
        # if any of the radius attributes changed, we currently need to restart
        # otherwise, a simple reload will handle interface adds and deletes.
        diff_pending = print_diff(self.config_file,
                                  self.scratch_config_file,
                                  self.color_diffs)
        if not diff_pending:
            return PENDING_COMMIT_NOCHANGE

        newconfig = self.get_config(self.scratch_config_file)
        if len(newconfig.get('agentAddress', '')) == 0:
            # snmpd will not work if either there are no listening-addresses
            # or both rocommunity and createUser are not configured
            print("\nERROR: Missing listening-address address")
            return PENDING_COMMIT_ERROR
        if  len(newconfig.get('rocommunity', '')) == 0 and \
            len(newconfig.get('createUser', '')) == 0:
            print("\nERROR: Either  listening-address address")
            return PENDING_COMMIT_ERROR

        return PENDING_COMMIT_CHANGE

    def do_show(self):
        '''
        This handles net show snmp-server status
        '''
        try:
            myshow = check_output(['/bin/systemctl', 'show', 'snmpd'])
        except CalledProcessError as ex:
            myshow = ex.output

        current = self.get_config(self.config_file)
        if current.get('rocommunity') or current.get('rocommunity6'):
            rocommunity = 'Configured'
        else:
            rocommunity = 'Not Configured'

        if len(current.get('createUser', '')) == 1 and \
           current.get('createUser', []) == ['_snmptrapusernameX']:
            v3users = 'Not Configured'
        elif len(current.get('createUser', '')) > 1:
            v3users = 'Configured'
        else:
            v3users = 'Not Configured'

        myshow = myshow.split('\n')
        output = {l.split('=')[0] : l.split('=')[1] for l in myshow if l}
        listeners = current.get('agentAddress')[0]
        if listeners == 'udp:161':
            listeners = 'all'
        elif '127.0.0.1' in listeners:
            listeners = listeners.replace('127.0.0.1', 'localhost')

        listeners = listeners.replace(',',' ')
        # grab the output for the ExecMainPID, if failed, MainPID==0
        try:
            if output.get('ActiveState') == 'failed':
                pid = '_PID=%s' % output.get('ExecMainPID')
                myerrors = check_output(['/bin/journalctl', pid]).split('\n')
            else:
                myerrors = []
        except CalledProcessError as ex:
            myerrors = []

        headers = ["Attribute", "Value"]
        table = []
        table.append(['Current Status', '%s (%s)' % (output.get('ActiveState', None),
                                                     output.get('SubState', None))])
        table.append(['Reload Status', output.get('UnitFileState')])
        table.append(['Listening IP Addresses', listeners])
        table.append(['Main snmpd PID', output.get('MainPID')])
        table.append(['Version 1 and 2c Community String', rocommunity])
        table.append(['Version 3 Usernames', v3users])
        for i in range(len(myerrors)):
            if i == 0:
                table.append(['Last Logs (with Errors)', myerrors[i]])
            else:
                table.append(['', myerrors[i]])

        print('')
        print(output.get('Description'))
        print(tabulate(tru(table), tablefmt="simple"))
        return True

    def show_config_files(self, user_may_edit):
        if os.path.exists(self.config_file):
            print "%s" % self.config_file
            print "======================="

            # If the user has edit privs then do not bother hiding passwords
            if user_may_edit:
                with open(self.config_file, 'r') as fh:
                    print ''.join(fh.readlines())
            else:
                output = []
                with open(self.config_file, 'r') as fh:
                    for line in fh.readlines():
                        output.append(self.hider(line))
                print ''.join(output)
            print ''

        return

    def hider(self, line):
        '''
        This hider function works in both directions: hiding attributes from
        the config and also from the nclu commands.
        '''
        line = re.sub(r'''community (\S+)''', 'community PASSWORD REMOVED', line, flags=re.I)
        line = re.sub(r'''createUser (\S+)''', 'createUser USERNAME PASSWORD REMOVED', line, flags=re.I)
        line = re.sub(r'''iquerySecName (\S+)''', 'iquerySecName USERNAME PASSWORD REMOVED', line, flags=re.I)
        line = re.sub(r'''trap-destination (\S+) community-password (\S+) version (\S+)''',
                      'trap-destination DESTINATION community-password PASSWORD version REMOVED', line, flags=re.I)
        line = re.sub(r'''trapsink (\S+) (\S+)''', 'trapsink DESTINATION PASSWORD REMOVED', line, flags=re.I)
        line = re.sub(r'''trap2sink (\S+) (\S+)''', 'trap2sink DESTINATION PASSWORD REMOVED', line, flags=re.I)
        line = re.sub(r'''rouser (\S+)''', 'rouser USERNAME PASSWORD REMOVED', line, flags=re.I)
        line = re.sub(r'''createuser (\S+) noauth (\S+)''', 'createUser REMOVED AUTH REMOVED', line, flags=re.I)
        line = re.sub(r'''createuser (\S+) (md5|des) (\S+)''', 'createUser REMOVED AUTH REMOVED', line, flags=re.I)
        line = re.sub(r'''createuser (\S+) (md5|sha) (\S+) (aes|des) (\S+)''', 'createUser REMOVED AUTH REMOVED ENCRYPT REMOVED', line, flags=re.I)
        line = re.sub(r'''username (\S+) restrict-access (\S+)''', 'username REMOVED restrict-access REMOVED', line, flags=re.I)
        line = re.sub(r'''username (\S+) auth-none (\S+)''', 'username REMOVED AUTH REMOVED', line, flags=re.I)
        line = re.sub(r'''username (\S+) auth-(md5|sha) (\S+)''', 'createUser REMOVED AUTH REMOVED', line, flags=re.I)
        line = re.sub(r'''username (\S+) auth-(md5|sha) (\S+) encrypt-(aes|des) (\S+)''', 'createUser REMOVED AUTH REMOVED', line, flags=re.I)
        return(line)

    def get_managed_files(self):
        files = []
        files.append(self.config_file)
        return files

    def get_combined_usernames(self, showarray):
        '''
        get_combined_usernames() is given an array indexed by usernames.
        Each entry is a list of converted values taken from createUser and rouser
        values but converted via regex to nclu cli keywords.
        '''
        if not showarray.get('username'):
            return showarray
        userlist = showarray.get('username', [])
        cli_dict = self.config_array.get('username', None)
        hidden = cli_dict.get('hidden-entries', [])
        userarray = {}
        for userstring in userlist:
            uservals = userstring.split(' ')
            name = uservals[0]
            if name in hidden:
                continue
            # only collect the auth keywords we care about. the encrypt
            # keywords may also exist an they will be added as well.
            if 'auth-none' in uservals or 'auth-md5' in uservals or 'auth-sha' in uservals:
                userarray[uservals[0]] = ' '.join(uservals[1:])
        # now attach the oid or view
        for userstring in userlist:
            uservals = userstring.split()
            if 'oid' in uservals or 'view' in uservals:
                name = uservals[0]
                if name in hidden:
                    continue
                if userarray.get(name):
                    # add on an oid or a view name if it exists
                    userarray[name] = '%s %s' % (userarray[name],
                                                 ' '.join(uservals[1:]))
        # need to create a list of entries with username attributes
        userlist = []
        entries = userarray.keys()
        entries.sort()
        for entry in entries:
           userlist.append('%s %s' % (entry, userarray[entry]))
        showarray['username'] = userlist

        return(showarray)

    def get_showarray(self):
        '''
        get_showarray() loops over all the nclu snmp-server keywords
        and for each keyword, loops over the snmpd.conf keywords for it.
        It collects the configs from snmpd.conf and correlates them
        to nclu keywords so we can show them. For a specific username,
        it collects the 'createUser' and 'rouser' configs and creates
        a single 'username' entry to show.
        '''
        running_array = self.get_config(self.config_file)
        # now show the supported commands
        entries = self.config_array.keys()
        entries.sort()
        showarray = {}
        for entry in entries:
            cli_dict = self.config_array[entry]
            newindexlist = cli_dict.get('config-name', [])
            for newindex in newindexlist:
                hidden = cli_dict.get('hidden-entries', [])
                vals = running_array.get(newindex, [])
                if vals:
                    for val in running_array.get(newindex, []):
                        if val in hidden:
                            # some things we do not want to show
                            # if you cannot configure it, you will not see it
                            # except when you look in the conf file
                            continue
                        sublist = cli_dict.get('sub_conf_to_nclu', {}).get(newindex, [])
                        regexlist = cli_dict.get('regex_conf_to_nclu', {}).get(newindex, [])
                        # make sure to convert the snmpd.config into usable nclu commands.
                        conv_val = self.converter(val, sublist, regexlist)
                        if not conv_val:
                            # we do not show what we cannot configure
                            continue
                        if showarray.get(entry):
                            # we are combining the createUser attributes with
                            # the rouser attributes in a list.
                            showarray.get(entry).append(conv_val)
                        else:
                            showarray[entry] = [conv_val]

        # now convert these two entries into one username array
        showarray = self.get_combined_usernames(showarray)
        return(showarray)

    def show_config_commands(self, user_may_edit):
        supported = []
        unsupported = []
        showarray = self.get_showarray()
        showkeys = showarray.keys()
        showkeys.sort()
        for entry in showkeys:
            for val in showarray[entry]:
                # for each item in list
                full_line = '%s %s' % (entry, val)
                if not user_may_edit:
                    supported.append('net add snmp-server %s' % (self.hider(full_line)))
                else:
                    supported.append('net add snmp-server %s' % (full_line))

        return (supported, unsupported)

    def show_config_summary(self, user_may_edit, summary):
        showarray = self.get_showarray()
        summary['snmp-server'] = OrderedDict()
        showkeys = showarray.keys()
        showkeys.sort()
        for entry in showkeys:
            for val in showarray[entry]:
                # for each item in list
                full_line = '%s %s' % (entry, val)
                if not user_may_edit:
                    summary['snmp-server'][self.hider(full_line)] = []
                else:
                    summary['snmp-server'][full_line] = []

        return

    def converter(self, val, sublist, regexlist):
        '''
        This should translate the config line into nclu options users are
        familar with.  Like udp:161 to all  and 127.0.0.1 to localhost.
        We also have to replace -V with view and insert OID for rocommunity.
        '''
        # handle substitutions first
        for pattern,replacement in sublist:
            val = re.sub(pattern, replacement, val)
        # now fix up syntax when converting from config file to nclu
        conv_val = ''
        for pattern,replacement in regexlist:
            m = re.match(pattern, val)
            if m:
                conv_val = replacement.format(*m.groups())
                break

        return(conv_val)

    def do_config(self):
        if (os.path.exists(self.scratch_config_file)):
            running_array = self.get_config(self.scratch_config_file)
        else:
            # we always make a copy of the config if it does not exist
            shutil.copyfile(self.config_file, self.scratch_config_file)
            os.chmod(self.scratch_config_file, 0o600)
            running_array = self.get_config(self.scratch_config_file)

        initial_array = copy.deepcopy(running_array)
        # running_array should have valid values for snmpd.conf, no special
        # words.
        for conf_word in self.config_array.keys():
            # before doing anything, check that the keyword was entered
            # or if we have no handler function to call, just continue
            # some keywords (view) could occur elsewhere in other commands
            if self.argv_expanded and len(self.argv_expanded) > 3:
                command = self.argv_expanded[3]
            else:
                command = ''
            if conf_word != command or not self.args.get(conf_word, None) or \
               not self.config_array.get(conf_word, None).get('handler', None):
                # we continue if command entered is not the current loop index or
                # the commanded was not entered
                # or there is no array to work with
                continue
            cli_dict = self.config_array.get(conf_word, {})
            handler = getattr(self, cli_dict.get('handler', None), None)
            if handler:
                running_array = handler(running_array, conf_word)

        changed = False
        if self.args.get('add'):
            keylist = running_array.keys()
        else:
            keylist = initial_array.keys()

        for i in keylist:
            if running_array.get(i, None) != initial_array.get(i, None):
                changed = True
        if changed:
            self.save_scratch_config(running_array)
        else:
            print("Configuration has not changed")
            return False
        return True

    def eval_command_line_args(self, cli):
        self.args = dict(((i,j) for (i,j) in cli.args.items() if j))
        self.argv_expanded = cli.argv_expanded
        if self.args.get('show'):
            rc = self.do_show()
        elif self.args.get('del') and self.argv_expanded[3] == 'all':
            rc = self.del_all(cli)
        else:
            rc = self.do_config()

        return rc

