# Copyright (C) 2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

import logging
import os
import re

from collections import OrderedDict
from nclu import ConfigFile, ConfigWrapperBase
from subprocess import call

log = logging.getLogger('netd')


def build_doc_string():
    global __doc__

    __doc__ = """
Usage:
    syslog (add|del) syslog host (ipv4 <ipv4>|ipv6 <ipv6>) port [udp|tcp] <number>
    net show configuration syslog

Options:
    host    : A remote host running a log server

"""


class ConfigWrapper(ConfigWrapperBase):
    """
    Use this wrapper to specify remote system log parameters.
    """
    # All log messages should go to a remote server, which implies the following destination (file name).
    REMOTE_LOG_OLD = '/etc/rsyslog.d/19-remotesyslog.conf'
    REMOTE_LOG = '/etc/rsyslog.d/11-remotesyslog.conf'


    def __init__(self, show_linux_command, color_diffs):
        super(ConfigWrapper, self).__init__('syslog', show_linux_command, color_diffs)

        # Create the remote syslog configuration file, if necessary.
        if not os.path.exists(ConfigWrapper.REMOTE_LOG):

            if os.path.exists(ConfigWrapper.REMOTE_LOG_OLD):
                os.rename(ConfigWrapper.REMOTE_LOG_OLD, ConfigWrapper.REMOTE_LOG)
                log.info("rename %s to %s and restart rsyslog" % (ConfigWrapper.REMOTE_LOG_OLD, ConfigWrapper.REMOTE_LOG))

                try:
                    call(('systemctl', 'reset-failed', 'rsyslog.service'))
                    call(('systemctl', 'restart', 'rsyslog.service'))
                except Exception as e:
                    log.exception(e)
            else:
                with open(ConfigWrapper.REMOTE_LOG, 'w') as f:
                    f.write('# This file was automatically generated by NCLU.\n')

        self.files[ConfigWrapper.REMOTE_LOG] = ConfigFile(self, ConfigWrapper.REMOTE_LOG,
                                                          [('systemctl', 'reset-failed', 'rsyslog.service'),
                                                           ('systemctl', 'restart', 'rsyslog.service')])

    def eval_command_line_args(self, cli):
        super(ConfigWrapper, self).eval_command_line_args(cli)  # Creates the scratchpad, if necessary.

        if cli.args.get('add') and cli.args.get('syslog') and cli.args.get('host'):
            self.add_remote_log_server(cli)
        elif cli.args.get('del') and cli.args.get('syslog') and cli.args.get('host'):
            self.del_remote_log_server(cli)
        else:
            raise RuntimeError("Unrecognized syslog command.")

        return True  # Returning True indicates that the command was successfully processed.  This has consequences downstream.

    def add_remote_log_server(self, cli):
        """
        syslog add syslog host (ipv4 <ip4>|ipv6 <ip6>) port [udp|tcp] <number>

        Start logging for UDP (default) or TCP on a given host.  If the user specifies
        a different port or protocol for an IP already in the configuration, the existing
        line should be replaced with the new information.

        http://stackoverflow.com/questions/186829/how-do-ports-work-with-ipv6#186848
        """
        ip = cli.args.get('<ipv4>') if (cli.args.get('ipv4') is not None) else cli.args.get('<ipv6>')
        port = cli.args.get('<number>')

        if cli.args.get('tcp') is None:
            # UDP (default)
            if cli.args.get('ipv4') is not None:
                output = '*.*   @{0}:{1}   # UDP\n'.format(ip, port)    # UDP and IPv4
            else:
                output = '*.*   @[{0}]:{1}   # UDP\n'.format(ip, port)  # UDP and IPv6
        else:
            # TCP
            if cli.args.get('ipv4') is not None:
                output = '*.*   @@{0}:{1}  # TCP\n'.format(ip, port)    # TCP and IPv4
            else:
                output = '*.*   @@[{0}]:{1}  # TCP\n'.format(ip, port)  # TCP and IPv6

        # Check for existing configurations for the given IP.
        with open(self.files[ConfigWrapper.REMOTE_LOG].scratchpad) as f:
            config_file = f.read()
        existing_re = re.compile(r'''^\s*\*\.\*\s+@@?\[?{0}\]?:\d+.*?$'''.format(ip), re.IGNORECASE | re.MULTILINE)
        rewrite_needed = existing_re.search(config_file) is not None

        if rewrite_needed:
            config_file = existing_re.sub('', config_file)  # Remove existing configurations for the given IP.
            with open(self.files[ConfigWrapper.REMOTE_LOG].scratchpad, 'w') as f:
                f.write(config_file)
                f.write(output)
        else:
            with open(self.files[ConfigWrapper.REMOTE_LOG].scratchpad, 'a') as f:
                f.write(output)

    def del_remote_log_server(self, cli):
        """
        net del syslog host (ipv4 <ipv4>|ipv6 <ipv6>) port [udp|tcp] <number>

        Stop logging for UDP and TCP on a given host.
        """
        ip = cli.args.get('<ipv4>') if (cli.args.get('ipv4') is not None) else cli.args.get('<ipv6>')
        port = cli.args.get('<number>')

        # Input the configuration scratchpad file.
        with open(self.files[ConfigWrapper.REMOTE_LOG].scratchpad) as f:
            config_file = f.read()

        if cli.args.get('tcp') is None:
            # UDP (default)
            if cli.args.get('ipv4') is not None:
                substitution = r'''^\s*\*\.\*\s+@{0}:{1}(?:\s+#\s*UDP)?\s*$'''       # UDP and IPv4
            else:
                substitution = r'''^\s*\*\.\*\s+@\[{0}\]:{1}(?:\s+#\s*UDP)?\s*$'''   # UDP and IPv6
        else:
            # TCP
            if cli.args.get('ipv4') is not None:
                substitution = r'''^\s*\*\.\*\s+@@{0}:{1}(?:\s+#\s*TCP)?\s*$'''      # TCP and IPv4
            else:
                substitution = r'''^\s*\*\.\*\s+@@\[{0}\]:{1}(?:\s+#\s*TCP)?\s*$'''  # TCP and IPv6

        config_file = re.sub(substitution.format(ip, port), '', config_file, 0, re.IGNORECASE | re.MULTILINE)

        # Output the modified configuration file.
        with open(self.files[ConfigWrapper.REMOTE_LOG].scratchpad, 'w') as f:
            f.write(config_file)

    def del_all(self, cli):
        """
        net del all
        """
        with open(self.files[ConfigWrapper.REMOTE_LOG].scratchpad, 'w'):
            pass  # Truncate the contents of the remote log scratchpad file.

    def show_config_commands(self, user_may_edit):
        if not os.path.exists(ConfigWrapper.REMOTE_LOG):
            return ([], [])

        # Read relevant data from the configuration file.
        with open(ConfigWrapper.REMOTE_LOG) as f:
            configs = re.findall(r'''^\s*\*\.\*\s+(?P<protocol>@@?)(?P<ip>[.:a-f\d]+):(?P<port>\d+)''', f.read(), re.IGNORECASE | re.MULTILINE)

        commands = []

        for protocol, ip, port in configs:
            protocol = 'udp' if protocol == '@' else 'tcp'
            commands.append('net add syslog host {0} port {1} {2}'.format(ip, protocol, port))

        return (commands, [])

    def show_config_summary(self, user_may_edit, summary):
        """
        net show configuration syslog
        """
        if not os.path.exists(ConfigWrapper.REMOTE_LOG):
            return  # Nothing to do.

        with open(ConfigWrapper.REMOTE_LOG) as f:
            config_file = f.read()

        ipv4 = re.findall(r'''^\s*\*\.\*\s+(@@?)([.\d]+:\d+)''', config_file, re.IGNORECASE | re.MULTILINE)
        ipv6 = re.findall(r'''^\s*\*\.\*\s+(@@?)(\[[:\da-f]+\]:\d+)''', config_file, re.IGNORECASE | re.MULTILINE)

        if not (ipv4 or ipv6):
            return

        summary['syslog'] = OrderedDict()

        if ipv4:
            summary['syslog']['ipv4'] = OrderedDict()
            for protocol, ip_port in ipv4:
                if protocol == '@':
                    # UDP
                    if summary['syslog']['ipv4'].get('udp') is None:
                        summary['syslog']['ipv4']['udp'] = []
                    summary['syslog']['ipv4']['udp'].append(ip_port)
                else:
                    # TCP
                    if summary['syslog']['ipv4'].get('tcp') is None:
                        summary['syslog']['ipv4']['tcp'] = []
                    summary['syslog']['ipv4']['tcp'].append(ip_port)

        if ipv6:
            summary['syslog']['ipv6'] = OrderedDict()
            for protocol, ip_port in ipv6:
                if protocol == '@':
                    # UDP
                    if summary['syslog']['ipv6'].get('udp') is None:
                        summary['syslog']['ipv6']['udp'] = []
                    summary['syslog']['ipv6']['udp'].append(ip_port)
                else:
                    # TCP
                    if summary['syslog']['ipv6'].get('tcp') is None:
                        summary['syslog']['ipv6']['tcp'] = []
                    summary['syslog']['ipv6']['tcp'].append(ip_port)
